/* ======================================================================
 *
 *  CCCCC          drawing.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: (unknown, sometime before June 20, 1998)
 * CC   CC         Last Modified: February 25, 2024
 *  CCCCC
 *
 * Code file for a sample VR application to render a simple world.
 *   A lot of this code came from my VR programming training examples.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <stdio.h>		/* needed for printf & all other standard I/O */
#include <string.h>		/* needed for strcmp() */
#include <stdlib.h>		/* needed for abort() */
#include <GL/gl.h>

#if defined(FREEVR) /* { */
#  include "freevr.h"
#  include "vr_visren.h"	/* needed for RenderInfo to get eye info */
#  include "vr_visren.glx.h"	/* needed to get the value of ENABLE_STENCIL_STEREO_TEST (remove when that test is complete) */

#  define	HEAD_SENSOR		0
#  define	WAND_SENSOR		1
#  define	WIREDHEAD_SENSOR	2
#  define	WIREDWAND_SENSOR	3
#else /* } { */
   void vrPrintf() { }		/* needed for vr_basicgfx.glx.h */
#  define	vrDbgDo(n)	1
#  define	vrPrintf	printf
#  define	vrErrPrintf	printf
#  define	vrFprintf	fprintf
#endif /* } */

#include "vr_basicgfx.glx.h"

#undef	LIMITED_TRACE
#undef	SECOND_USER_PERSPECTIVE_TEST

/* draw_world(): draw a simple world typically used by the "travel" */
/* draw_server_world(): draw nearly empty world typically used by "inputs" */
void world1(vrRenderInfo *rendinfo);		/* draws the standard "travel" world */
void world2(vrRenderInfo *rendinfo);		/* draws a world used for some configuration testing */
void gray_world(vrRenderInfo *rendinfo);	/* draws a very simple world entirely made of gray-scale objects */
void draw_axes();				/* draw colored axes at the origin: red=x, green=y, blue=z */
float draw_text(vrRenderInfo *rendinfo, float y);		/* puts into text info on left and front CAVE walls */
float draw_inputs_text(vrRenderInfo *rendinfo, float y);	/* add text info to CAVE front screen about all the input types */
void draw_lines(vrRenderInfo *rendinfo);	/* draw lines connected to tracked head & wand  */
void draw_config_lines(vrRenderInfo *rendinfo);	/* adds to draw_lines grey crosshairs on each screen */
void draw_inputs(vrRenderInfo *rendinfo);	/* draw representations of all the inputs (buttons, valuators, etc. */
void draw_indev_axes(vrRenderInfo *rendinfo);	/* draw axes for the origin of each input device */
void draw_6sensor_axes(vrRenderInfo *rendinfo);	/* draw axes at the position of each 6-sensor */
void draw_vrui_outline(vrRenderInfo *rendinfo);	/* draws a fixed outline of a chaperone as created by Vrui */

/*********************************************************************/
/* init_gfx(): initialize general graphics properties (eg. lighting) */
/*********************************************************************/
void init_gfx(vrRenderInfo *rendinfo)
{
static	int	num_init = 0;

#ifdef UNPICKY_COMPILER
	GLfloat	light_position[] = { 1.0, 1.0, 9.0, 1.0 };
	GLfloat	light_diffuse[] = { 0.8, 0.8, 0.8, 1.0 };
	GLfloat	light_ambient[] = { 0.3, 0.3, 0.3, 1.0 };
#endif

#if defined(FREEVR) && 0
	vrPrintf("App: " BOLD_TEXT "Entering init_gfx().  When am I?\n" NORM_TEXT);
	vrPrintf("App: init graphics for context %p, persp %p, window %p (%d), eye %p\n",
		rendinfo->context, rendinfo->persp, rendinfo->window, rendinfo->window->num, rendinfo->eye);
#endif
	/* If we init too many times, there is probably a bug, so abort and dump core */
	num_init++;
	if (num_init == 10) abort();


	glShadeModel(GL_SMOOTH);
#if 1 /* this causes overly thick lines on SGI/IRIX */
	glEnable(GL_LINE_SMOOTH);
#endif
	glLineWidth(2.0);

	glDrawBuffer(GL_FRONT_AND_BACK);
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawBuffer(GL_BACK);

#if 0
	/* TODO: when using a lighting model, objects without a */
	/*   material will appear as shades of gray.            */
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
#endif
}


/********************************************************************/
/* draw_world(): draw a simple world typically used by the "travel" */
/*   program that is often used to test the basic functionality of  */
/*   the library, including the ability to use the *user-travel*    */
/*   functionality.  It is also often used to test the configuration*/
/*   of a standard CAVE layout, as many of the objects are placed   */
/*   at key locations for a 10x10 CAVE.                             */
/*                                                                  */
/* Primarily calls:                                                 */
/*   * world1() | world2() | gray_world() -- choice of simple worlds*/
/*   * draw_text() --                                               */
/*   * draw_lines() -- draw lines connected to tracked head & wand  */
/*   * draw_inputs() -- draw representations of all the inputs      */
void draw_world(vrRenderInfo *rendinfo)
{

#if 0 /* include this when testing the stats feature -- especially for multiple bins */
vrSleep(12000);
vrRenderCategory(rendinfo, 2);
vrSleep(8000);
#endif
#if 0
	int		val;
	do {
		val = glGetError();
		vrPrintf(RED_TEXT "glGetError returned %d\n" NORM_TEXT, val);
	} while (val != 0);
#endif

#if defined(FREEVR) && 0
	vrPrintf("App: render graphics for context %p, persp %p, window %p (%d), eye %p (%d)\n",
		rendinfo->context, rendinfo->persp,
		rendinfo->window, rendinfo->window->num,
		rendinfo->eye, rendinfo->eye->num);
#endif
#if 0 /* To print information about which process/thread we are rendering to */
	vrFprintProcessInfo(stdout, rendinfo->window->proc, oneline);
#endif
	vrTrace("draw_world", "beginning");

#if !defined(ENABLE_STENCIL_STEREO_TEST) || 0 /* Don't do this when rendering to checkerboard stencil */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#endif

	/* Draw colored axes at the origin: red=x, green=y, blue=z */
	draw_axes();

#if defined(FREEVR) /* { */
	/* Draw colored axes at the head: red=x, green=y, blue=z */
	if ((rendinfo->window->mount == VRWINDOW_SIMULATOR) && (rendinfo->context->input->num_6sensors > 0)) {
		glPushMatrix();
			vrRenderTransform6sensor(rendinfo, 0);
			draw_axes();
		glPopMatrix();
	}
#endif /* } */

vrTrace("draw_world", "yo1");
	/* Render a simple world */
#if 1
	world1(rendinfo);	/* world populated by various objects in the RW, plus a pyramid in the VW */
#elif 1
	world2(rendinfo);	/* a world with a cube in each of the 6 directions, plus a pyramid on the floor -- all in the RW */
#else
	gray_world(rendinfo);	/* just a RW gray cube and VW gray pyramid */
#endif
vrTrace("draw_world", "yo2");

	/* Add some other information */
	draw_text(rendinfo, 8.5);
vrTrace("draw_world", "yo3");
	draw_lines(rendinfo);
vrTrace("draw_world", "yo4");
	draw_inputs(rendinfo);
vrTrace("draw_world", "yo5");

#if 0
	glColor3ub(100,255,100);		/* light green */
	vrShapeGLFloor();
#endif

	vrTrace("draw_world", "ending");
}


/***************************************************************************/
/* draw_server_world(): draw nearly empty world typically used by "inputs" */
/*   (nee valtest) program that is often used as a simulator input or as an*/
/*   input analyzer.  The world contains only text about the inputs along  */
/*   with representations of each type of primary input (buttons, valuators*/
/*   & 6sensors) in a darker green hue.                                    */
/*                                                                         */
/* Primarily calls:                                                        */
/*   * draw_inputs_text() -- add text info to CAVE front screen about all the input types */
/*   * draw_inputs() -- draw representations of all the inputs             */
void draw_server_world(vrRenderInfo *rendinfo)
{
	vrPoint		wand1, wand2;

	vrTrace("draw_server_world", "beginning");

#if !defined(ENABLE_STENCIL_STEREO_TEST) || 0 /* Don't do this when rendering to checkerboard stencil */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#endif

#if 0
	/* Draw colored axes at the origin: red=x, green=y, blue=z */
	draw_axes();
#endif

#if defined(FREEVR) /* { */   && 0 /* this is now covered by drawing all 6-sensor axes */ /* 09/17/21: so delete me? */
	/* Draw colored axes at the head: red=x, green=y, blue=z */
	if ((rendinfo->window->mount == VRWINDOW_SIMULATOR) && (rendinfo->context->input->num_6sensors > 0)) {
		glPushMatrix();
			vrRenderTransform6sensor(rendinfo, 0);
			draw_axes();
		glPopMatrix();
	}
#endif /* } */

	/* Add some useful information for a server display */
#if 1 /* it's a bit too much right now */
	draw_lines(rendinfo);
	draw_inputs_text(rendinfo, 5.5);
#endif
	draw_text(rendinfo, 8.5);
	draw_inputs(rendinfo);

	if (rendinfo->window->simulator_mask & SIMRENDER_INDEV_AXES) {
		draw_indev_axes(rendinfo);
	}
	if (rendinfo->window->simulator_mask & SIMRENDER_SENSOR_AXES) {
		draw_6sensor_axes(rendinfo);
	}

#if 0 /* set to '1" to include the experimental Vrui chaparone shape */
	/* Experiment: draw the Vrui chaparone shape */
	glPushMatrix();
#  if 1 /* temporarily add this to tweak wand-avatar representations */
	glTranslatef(0.0, 4.02, 0.0);
	glTranslatef(-5.0, 0.0, 9.5);
	glRotatef(-60.0, 0.0, 1.0, 0.0);
	//glTranslatef(10.0, 0.0, 9.0);
#  endif
	draw_vrui_outline(rendinfo);	/* NOTE: this is adjusted based on the Vrui input device t2rw coordinates */
	glPopMatrix();
#endif

	/* Experiement: draw a line from controller to controller */
	/* Draw a cyan line pointing from the user to center of gray pyramid */
	vrPointGetRWFrom6sensor(&wand1, 1);
	vrPointGetRWFrom6sensor(&wand2, 2);

	//vrPointGetRWFromVWUserPoint(&drawto, 0, &drawto);
	glColor3ub(50, 240, 240);		/* cyan */
	glBegin(GL_LINE_STRIP);
		glVertex3dv(wand1.v);
		glVertex3dv(wand2.v);
	glEnd();


	vrTrace("draw_server_world", "ending");
}


/***********************************************************************/
/* world1(): this routine draws the standard "travel" world used for   */
/*   most of the basic FreeVR testing, including library functionality */
/*   and CAVE configuration.                                           */
void world1(vrRenderInfo *rendinfo)
{
#ifndef LIMITED_TRACE
	vrTrace("draw_world", "the more complicated test world");
#endif
	/* NOTE: start in RW coordinate space */

	/* A 1' rose cube behind the cave */
	glPushMatrix();
		glColor3ub(240, 138, 133);		/* rose */
		glTranslatef(0.0, 5.0, 7.0);
		glScalef(0.5, 0.5, 0.5);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 4' blue cube in front of the cave */
	glPushMatrix();
		glColor3ub(0, 0, 255);			/* blue */
		glTranslatef(0.0, 3.0, -7.0);
		glScalef(2.0, 2.0, 2.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 4' white cube left of the cave */
	glPushMatrix();
		glColor3ub(255, 255, 255);		/* white */
		glTranslatef(-7.0, 3.0, 0.0);
		glScalef(2.0, 2.0, 2.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A pink pyramid on the floor of the cave near the front wall */
	glPushMatrix();
		glColor3ub(255, 100, 100);		/* pink */
		glTranslatef(0.0, 1.0, -4.0);
		vrShapeGLPyramidPlusOutline();
	glPopMatrix();

	/* A 2' light purple cube at the front left corner of the cave */
	glPushMatrix();
		glColor3ub(100, 100, 255);		/* light purple */
		glTranslatef(-6.0, 7.0, -6.0);
		glScalef(1.0, 1.0, 1.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 2' green cube below the cave */
	glPushMatrix();
		glColor3ub(0, 255, 0);			/* green */
		glTranslatef(5.0, 1.0, 0.0);
		glScalef(1.0, 1.0, 1.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 2' dark purple cube below the center of the cave */
	glPushMatrix();
		glColor3ub(100, 0, 255);		/* dark purple */
		glTranslatef(0.0, -1.0, 0.0);
		glScalef(1.0, 1.0, 1.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 10' tall red pillar inside the right front corner of the cave */
	glPushMatrix();
		glColor3ub(240, 50, 50);		/* red */
		glTranslatef(4.5, 5.0, -4.5);
		glScalef(0.5, 5.0, 0.5);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 10' deep cyan pillar inside the right front corner of the cave */
	glPushMatrix();
		glColor3ub(50, 240, 240);		/* cyan */
		glTranslatef(4.5, -5.0, -4.5);
		glScalef(0.5, 5.0, 0.5);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 4' tall yellow pillar near the right wall of the cave */
	glPushMatrix();
		glColor3ub(240, 240, 50);		/* yellow */
		glTranslatef(-2.5, 2.0, 2.0);
		glScalef(0.5, 2.0, 0.5);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

#  if 0 /* temporarily turned off so we can see the front screen */
	/* A 4' long orange pillar out of the right wall of the cave */
	glPushMatrix();
		glColor3ub(240, 130, 30);		/* orange */
		glTranslatef(-3.0, 2.0, -2.5);
		glScalef(2.0, 0.5, 0.5);
		vrShapeGLCubePlusOutline();
	glPopMatrix();
#  endif

	glPushMatrix();
#if defined(FREEVR) /* { */
		vrRenderTransformUserTravel(rendinfo);	/* now in VW coordinate space */
#endif /* } */

		/* An 80% gray pyramid near eye-height */
		glPushMatrix();
			glColor3ub(204, 204, 204);	/* 80% gray */
			glTranslatef(-2.3, 6.2, -0.5);
			glScalef(1.5, 1.5, 1.5);
			glRotatef(90.0, 1.0, 0.0, 0.0);
			vrShapeGLPyramidPlusOutline();
		glPopMatrix();
	glPopMatrix();			/* back to RW coordinate space */

#if defined(FREEVR) && defined(SECOND_USER_PERSPECTIVE_TEST) /* (FreeVR-only feature) { */
	/* This is a new one!  [01/27/2009]  An object that is rendered from  */
	/*   the perspective of a second user (although for the moment really */
	/*   just some other tracking device.                                 */
	glPushMatrix();
#if 0 /* Choose an option */
		/* Render from the perspective of sensor-1 (aka the wand) -- skip if push fails */
		/* Hmmm, so this will mean that if the push fails, this bit of the world won't  */
		/*   get rendered -- interesting to consider whether this is the appropriate    */
		/*   behavior.                                                                  */
		if (vrRenderPushPerspFrom6sensor(rendinfo, WAND_SENSOR)) {
			glPushMatrix();
				glColor3ub(240, 130, 30);	/* orange */
				glTranslatef(0.0, 5.0, 0.0);	/* ~center of a CAVE */
				glScalef(0.5, 0.5, 0.5);
				vrShapeGLPyramidPlusOutline();
			glPopMatrix();

			vrRenderPopPersp(rendinfo);
		}
#elif 0
		/* This version puts the orange pyramid at the location of the WAND, */
		/*   but viewed from the 2nd user POV.                               */
		if (vrRenderPushPerspFrom6sensor(rendinfo, WIREDHEAD_SENSOR)) {
			glPushMatrix();
				glColor3ub(240, 130, 30);	/* orange */
				vrRenderTransform6sensor(rendinfo, WAND_SENSOR);	/* located at the WAND */
				glScalef(0.5, 0.5, 0.5);
				vrShapeGLPyramidPlusOutline();
			glPopMatrix();

			vrRenderPopPersp(rendinfo);
		}
#elif 1
		/* This version puts the orange pyramid at the location of the WAND, */
		/*   and renders it from both users' points of view.                 */
		glPushMatrix();
			glColor3ub(240, 130, 30);	/* orange */
			vrRenderTransform6sensor(rendinfo, WAND_SENSOR);	/* located at the WAND */
			glScalef(0.5, 0.5, 0.5);
			vrShapeGLPyramidPlusOutline();
		glPopMatrix();
		if (vrRenderPushPerspFrom6sensor(rendinfo, WIREDHEAD_SENSOR)) {
			glPushMatrix();
				glColor3ub(240, 130, 30);	/* orange */
				vrRenderTransform6sensor(rendinfo, WAND_SENSOR);	/* located at the WAND */
				glScalef(0.5, 0.5, 0.5);
				vrShapeGLPyramidPlusOutline();
			glPopMatrix();

			vrRenderPopPersp(rendinfo);
		}
#endif

	glPopMatrix();
#endif /* }  (FreeVR-only code) */

#ifndef LIMITED_TRACE
	vrTrace("draw_world", "end of the more complicated test world");
#endif
}


/***********************************************************************/
/* world2(): this routine draws the modified "travel" world used for   */
/*   some configuration testing.  In this world, 6 3x3x3' cubes are    */
/*   rendered, 4, 2' on the other side of each vertical screen as      */
/*   though resting on an extended floor, one (green) about 7' down,   */
/*   one (dark purple) one above the user about 7' above the floor.    */
/*   Also, there is a pink pyramid on the floor.                       */
void world2(vrRenderInfo *rendinfo)
{

	/* NOTE: start in RW coordinate space */

	/* A 6' blue cube left of the cave */
	glPushMatrix();
		glColor3ub(0, 0, 255);			/* blue */
		glTranslatef(-10.0, 3.0, 0.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 6' dark purple cube above the user */
	glPushMatrix();
		glColor3ub(100, 0, 255);		/* dark purple */
		glTranslatef(0.0, 10.0, 0.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 6' red cube right of the cave */
	glPushMatrix();
		glColor3ub(255, 0, 0);			/* red */
		glTranslatef(10.0, 3.0, 0.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 6' green cube below the cave */
	glPushMatrix();
		glColor3ub(0, 255, 0);			/* green */
		glTranslatef(0.0, -10.0, 0.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 6' light purple cube behind the cave */
	glPushMatrix();
		glColor3ub(100, 100, 255);		/* light purple */
		glTranslatef(0.0, 3.0, 10.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A 6' white cube in front of the cave */
	glPushMatrix();
		glColor3ub(255, 255, 255);		/* white */
		glTranslatef(0.0, 3.0, -10.0);
		glScalef(3.0, 3.0, 3.0);
		vrShapeGLCubePlusOutline();
	glPopMatrix();

	/* A pyramid sticking through the floor of the cave near the right wall */
	glPushMatrix();
		glColor3ub(255, 100, 100);		/* pink */
		glTranslatef(4.0, 0.0, 0.0);
		vrShapeGLPyramidPlusOutline();
	glPopMatrix();
}


/********************************************************************/
/* gray_world(): this routine draws a very simple world that is     */
/*   entirely made of gray-scale objects.  In fact it has only      */ 
/*   two objects, a 50% gray cube resting on the floor, and a white */
/*   floating pyramid that moves with the user-travel input.        */
void gray_world(vrRenderInfo *rendinfo)
{

	/* NOTE: start in RW coordinate space */

	/* A 2' 50% gray cube just above the floor of the cave */
	glPushMatrix();
		glColor3ub(128, 128, 128);		/* 50% gray */
		glTranslatef(-2.5, 2.0, -0.5);
		glScalef(1.0, 1.0, 1.0);
		vrShapeGLCubePlusOutline();
		glPopMatrix();

		glPushMatrix();
#if defined(FREEVR) /* { */
		vrRenderTransformUserTravel(rendinfo);	/* now in VW coordinate space */
#endif /* } */

		/* An 80% gray pyramid near eye-height */
		glPushMatrix();
			glColor3ub(204, 204, 204);	/* 80% gray */
			glTranslatef(-2.3, 6.2, -0.5);
			glScalef(1.5, 1.5, 1.5);
			glRotatef(90.0, 1.0, 0.0, 0.0);
#if 0
			vrShapeGLPyramidPlusOutline();
#else
			vrShapeGLPyramid();
#endif
		glPopMatrix();

	glPopMatrix();			/* back to RW coordinate space */
}


/************************************************************************/
/* draw_axes(): draw colored axes at the origin: red=x, green=y, blue=z */
void draw_axes()
{
	glBegin(GL_LINES);
		/* Positive X-axis in red */
		glColor3ub(172, 64, 64);		/* red */
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(2.0, 0.0, 0.0);

		/* Positive Y-axis in green */
		glColor3ub(64, 172, 64);		/* green */
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 2.0, 0.0);

		/* Positive Z-axis in blue */
		glColor3ub(64, 64, 172);		/* blue */
		glVertex3f(0.0, 0.0, 0.0);
		glVertex3f(0.0, 0.0, 2.0);
	glEnd();
}

/**********************************************************************/
/* draw_inputs(): this routine renders representations of the button, */
/*   valuator and N-way and N-sensor inputs.  Buttons are rendered as */
/*   dark green boxes that pop up when the given buttin is pressed.   */
/*   Valuators are represented as dark green pyramids that grow and   */
/*   shrink along with the valuators input, pointing up for positive  */
/*   values and down for negative.  N-way switches are also dark green*/
/*   boxes, but N-sensors are dark-blue boxes.  The later two are     */
/*   rarely used/tested.  Note that 6-sensors are not handled by this,*/
/*   as they are typically rendered in simulator view, or in physical */
/*   use map directly to what the user is already holding/wearing.    */
void draw_inputs(vrRenderInfo *rendinfo)
{
	int		count;
	vrInputInfo	*inputinfo;	/* Pointer to the overall input data structure */

#ifndef LIMITED_TRACE
	vrTrace("draw_inputs", "doing the input-vis rendering");
#endif
	inputinfo = rendinfo->context->input;

	/*******************************************************/
	/** Print some boxes representing 2-switch input info **/
	/*******************************************************/
	for (count = 0; count < inputinfo->num_2ways; count++) {
		glPushMatrix();
		glTranslatef(-4.5 + (0.35 * count), 0.0, -4.0);
		glScalef(0.1, 1.2 * vrGet2switchValueNoLastUpdate(count) + 0.1, 0.1);
		glColor3ub(84, 134, 72);		/* dark green */
		vrShapeGLCubePlusOutline();
		glPopMatrix();
	}

	/*******************************************************/
	/** Print some boxes representing N-switch input info **/
	/*******************************************************/
	for (count = 0; count < inputinfo->num_Nways; count++) {
		glPushMatrix();
		glTranslatef(-4.5 + (0.35 * count), 0.0, -3.0);
		glScalef(0.1, 0.4 * vrGetNswitchValueNoLastUpdate(count) + 0.1, 0.1);
		glColor3ub(84, 134, 72);		/* dark green */
		vrShapeGLCubePlusOutline();
		glPopMatrix();
	}

	/**********************************************************/
	/** Print some pyramids representing valuator input info **/
	/**********************************************************/
	for (count = 0; count < inputinfo->num_valuators; count++) {
		glPushMatrix();
		glTranslatef(-4.5 + (0.35 * count), 0.0, -2.0);
		glScalef(0.1, 4.0 * vrGetValuatorValueNoLastUpdate(count), 0.1);
		glColor3ub(84, 134, 72);		/* dark green */
		vrShapeGLPyramidPlusOutline();
		glPopMatrix();
	}

	/*******************************************************/
	/** Now display boxes from N-sensor[0] (if available) **/
	/*******************************************************/
	if (inputinfo->num_Nsensors > 0) {
		for (count = 0; count < 22; count++) {
			glPushMatrix();
			glTranslatef(-4.5 + (0.35 * count), -1.0, -3.5);
			glScalef(0.1, 4.0 * vrGetNsensorValueNoLastUpdate(0, count), 0.1);
			glColor3ub(84, 72, 134);	/* dark/navy blue */
			vrShapeGLCube();
			glPopMatrix();
		}
	}

#ifndef LIMITED_TRACE
	vrTrace("draw_inputs", "ending");
#endif
}


/**********************************************************************/
/* draw_lines(): this routine draws lines that emanate from the head  */
/*   (1st tracked object) and the wand (2nd tracked object).  In      */
/*   particular, from the head:                                       */
/*      - a red line connects the head to the back-low-center of CAVE */
/*      - a green line points out from the front                      */
/*      - a blue line points to the center of the gray pyramid        */
/*      - a magenta lines points toward the front of the virtual world*/
/*   and from the wand:                                               */
/*      - a yellow line points out the front of the "wand"            */
/*      - a cyan line points up from the "wand"                       */
void draw_lines(vrRenderInfo *rendinfo)
{
	vrMatrix	head;
	vrPoint		headloc;
	vrPoint		headlocvw;
	vrPoint		proploc;
	vrVector	vec;
	vrPoint		drawto;
	int		count;
	vrInputInfo	*inputinfo;	/* Pointer to the overall input data structure */

	inputinfo = rendinfo->context->input;

#ifndef LIMITED_TRACE
	vrTrace("draw_lines", "doing the lines rendering");
#endif

	/* Ensure that there is a matrix that maps the real world to the head */
	/* TODO: [01/22/21] Negate this condition and return immediately if NULL, and then reindent */
	if (vrMatrixGetRWFromUserHead(&head, 0) != NULL) {
		vrPointGetRWLocationFromMatrix(&headloc, &head);
		vrPointGetVWFromUserMatrix(&headlocvw, 0, &head);

		/* First store, then adjust some graphics attributes for vector rendering */
		glPushAttrib(GL_COLOR_BUFFER_BIT | GL_LINE_BIT);
		glLineWidth(5.0);
		glDisable(GL_BLEND);

		/* Only draw these lines in simulator views & in systems with at least one 6sensor */
		if (rendinfo->window->mount == VRWINDOW_SIMULATOR && inputinfo->num_6sensors >= 1) {
			/* Draw a red line connecting the user to the back-low-center of the CAVE */
			drawto.v[VR_X] = 0.0;
			drawto.v[VR_Y] = 0.0;
			drawto.v[VR_Z] = 5.0;
			glColor4ub(255, 0, 0, 255);	/* red */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

			/* Draw a green line pointing out the front of the user */
			vrVectorGetRWFrom6sensorDir(&vec, 0, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &headloc, &vec, 2.0);
			glColor4ub(0, 255, 0, 255);	/* green */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

			/* Draw a blue line pointing from the user to center of gray pyramid */
			drawto.v[VR_X] = -2.3;
			drawto.v[VR_Y] =  6.2;
			drawto.v[VR_Z] = -0.5;
			vrPointGetRWFromVWUserPoint(&drawto, 0, &drawto);
			glColor4ub(0, 0, 255, 255);	/* blue */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

#if 1 /* MP_NONE (single-thread) FreeVR messes this up.  TODO: Why? */
			glPushMatrix();
			vrRenderTransformUserTravel(rendinfo);	/* now in VW coordinate space */
			/* Draw a magenta line pointing out of the user toward the front of the VW*/
			/* NOTE: this should be the same as the green line coming from the user's */
			/*   nose, but in this case we're doing all the calculations in VW space. */
			vrVectorGetVWFromUser6sensorDir(&vec, 0, 0, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &headlocvw, &vec, 1.8);
			glColor4ub(255, 0, 255, 255);	/* magenta */
			glLineWidth(16.0);
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headlocvw.v);
				glVertex3dv(drawto.v);
			glEnd();
			glLineWidth(5.0);
			glPopMatrix();
#endif
		}

		if (inputinfo->num_6sensors >= 2) {
			/* Draw a yellow line pointing out the front of the "wand" */
			vrPointGetRWFrom6sensor(&proploc, WAND_SENSOR);
			vrVectorGetRWFrom6sensorDir(&vec, WAND_SENSOR, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &proploc, &vec, 2.0);
			glColor4ub(255, 255, 0, 255);		/* yellow */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(proploc.v);
				glVertex3dv(drawto.v);
			glEnd();

			/* Draw a cyan line pointing up from the "wand" */
			vrVectorGetRWFrom6sensorDir(&vec, WAND_SENSOR, VRDIR_UP);
			vrPointAddScaledVector(&drawto, &proploc, &vec, 1.5);
			glColor4ub(0, 255, 255, 255);		/* cyan */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(proploc.v);
				glVertex3dv(drawto.v);
			glEnd();

		}

		glPopAttrib();

#if 0
		snprintf(buffer, sizeof(buffer), "drawto: (%6.2f %6.2f %6.2f)", drawto.v[VR_X], drawto.v[VR_Y], drawto.v[VR_Z]);
		glRasterPos3f(-6.5,  13.0, -5.01);
		vrRenderText(rendinfo, buffer);
#endif
	}

#ifndef LIMITED_TRACE
	vrTrace("draw_lines", "ending");
#endif
}


/*********************************************************************/
/* draw_config_lines(): this routine draws some of the same lines as */
/*   rendered by draw_lines (red line from head, and two lines from  */
/*   wand), but then adds gray cross hairs renderings on each of the */
/*   standard 4-sided CAVE screens (front, left, right & floor).     */
/* TODO: [01/22/21] probably this should just be reduced to rendering */
/*   the cross hairs, and then can (or not) be combined with the      */
/*   draw_lines() function above.  Presently this function is not used*/
void draw_config_lines(vrRenderInfo *rendinfo)
{
	vrMatrix	head;
	vrPoint		headloc;
	vrPoint		headlocvw;
	vrPoint		proploc;
	vrVector	vec;
	vrPoint		drawto;
	int		count;
	vrInputInfo	*inputinfo;	/* Pointer to the overall input data structure */

	inputinfo = rendinfo->context->input;

#ifndef LIMITED_TRACE
	vrTrace("draw_config_lines", "doing the lines rendering");
#endif

	/* Ensure that there is a matrix that maps the real world to the head */
	/* TODO: [01/22/21] Negate this condition and return immediately if NULL, and then reindent */
	if (vrMatrixGetRWFromUserHead(&head, 0) != NULL) {
		vrPointGetRWLocationFromMatrix(&headloc, &head);
		vrPointGetVWFromUserMatrix(&headlocvw, 0, &head);

		/* First store, then adjust some graphics attributes for vector rendering */
		glPushAttrib(GL_COLOR_BUFFER_BIT | GL_LINE_BIT);
		glLineWidth(5.0);
		glDisable(GL_BLEND);

		/* Only draw these lines in simulator views & in systems with at least one 6sensor */
		if (rendinfo->window->mount == VRWINDOW_SIMULATOR && inputinfo->num_6sensors >= 1) {
			/* Draw a red line connecting the user to the front-screen of the CAVE */
			drawto.v[VR_X] = headloc.v[VR_X];
			drawto.v[VR_Y] = headloc.v[VR_Y];
			drawto.v[VR_Z] = 0.0;	/* changed from -5 to 0.0 to be in the center of the CAVE where the IQ-station is */
			glColor4ub(255, 0, 0, 255);		/* red */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

#if 0
			/* Draw a green line pointing out the front of the user */
			vrVectorGetRWFrom6sensorDir(&vec, 0, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &headloc, &vec, 2.0);
			glColor4ub(0, 255, 0, 255);		/* green */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

			/* Draw a blue line pointing from the user to center of gray pyramid */
			drawto.v[VR_X] = -2.3;
			drawto.v[VR_Y] =  6.2;
			drawto.v[VR_Z] = -0.5;
			vrPointGetRWFromVWUserPoint(&drawto, 0, &drawto);
			glColor4ub(0, 0, 255, 255);		/* blue */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headloc.v);
				glVertex3dv(drawto.v);
			glEnd();

			glPushMatrix();
			vrRenderTransformUserTravel(rendinfo);	/* now in VW coordinate space */
			/* Draw a magenta line pointing out of the user toward the front of the VW */
			/* NOTE: this should be the same as the green line coming from the user's  */
			/*   nose, but in this case we're doing all the calculations in VW space.  */
			vrVectorGetVWFromUser6sensorDir(&vec, 0, 0, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &headlocvw, &vec, 1.8);
			glColor4ub(255, 0, 255, 255);		/* magenta */
			glLineWidth(16.0);
			glBegin(GL_LINE_STRIP);
				glVertex3dv(headlocvw.v);
				glVertex3dv(drawto.v);
			glEnd();
			glLineWidth(5.0);
			glPopMatrix();
#endif
		}

		if (inputinfo->num_6sensors >= 2) {
			/* Draw a yellow line pointing out the front of the "wand" */
			vrPointGetRWFrom6sensor(&proploc, WAND_SENSOR);
			vrVectorGetRWFrom6sensorDir(&vec, WAND_SENSOR, VRDIR_FORE);
			vrPointAddScaledVector(&drawto, &proploc, &vec, 2.0);
			glColor4ub(255, 255, 0, 255);		/* yellow */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(proploc.v);
				glVertex3dv(drawto.v);
			glEnd();

			/* Draw a cyan line pointing up from the "wand" */
			vrVectorGetRWFrom6sensorDir(&vec, WAND_SENSOR, VRDIR_UP);
			vrPointAddScaledVector(&drawto, &proploc, &vec, 1.5);
			glColor4ub(0, 255, 255, 255);		/* cyan */
			glBegin(GL_LINE_STRIP);
				glVertex3dv(proploc.v);
				glVertex3dv(drawto.v);
			glEnd();

		}

		/* Draw cross-hair guides on some of the CAVE-walls */
		glColor3ub((GLbyte)55, (GLbyte)55, (GLbyte)55);	/* gray */
		glLineWidth(2.0);

		/* Front wall */
		glBegin(GL_LINE_STRIP);
			glVertex3d(-5.0, 5.0, -5.0);
			glVertex3d( 5.0, 5.0, -5.0);
			glVertex3d( 5.0, 0.0, -5.0);	/* filler point to allow one strip */
			glVertex3d( 0.0, 0.0, -5.0);
			glVertex3d( 0.0,10.0, -5.0);
		glEnd();

		/* Floor surface */
		glBegin(GL_LINE_STRIP);
			glVertex3d(-5.0, 0.0,  0.0);
			glVertex3d( 5.0, 0.0,  0.0);
			glVertex3d( 5.0, 0.0,  5.0);	/* filler point to allow one strip */
			glVertex3d( 0.0, 0.0,  5.0);
			glVertex3d( 0.0, 0.0, -5.0);
		glEnd();

		/* Left wall */
		glBegin(GL_LINE_STRIP);
			glVertex3d(-5.0, 5.0, -5.0);
			glVertex3d(-5.0, 5.0,  5.0);
			glVertex3d(-5.0, 0.0,  5.0);	/* filler point to allow one strip */
			glVertex3d(-5.0, 0.0,  0.0);
			glVertex3d(-5.0,10.0,  0.0);
		glEnd();

		/* Right wall */
		glBegin(GL_LINE_STRIP);
			glVertex3d( 5.0, 5.0, -5.0);
			glVertex3d( 5.0, 5.0,  5.0);
			glVertex3d( 5.0, 0.0,  5.0);	/* filler point to allow one strip */
			glVertex3d( 5.0, 0.0,  0.0);
			glVertex3d( 5.0,10.0,  0.0);
		glEnd();

		glPopAttrib();
	}

#ifndef LIMITED_TRACE
	vrTrace("draw_config_lines", "ending");
#endif
}


/*********************************************************************/
/* draw_text(): this routine "prints" some text information onto the */
/*   left and front CAVE walls.  All the text is colored medium gray.*/
/*   The information presented includes:                             */
/*     - shared memory usage                                         */
/*     - frame rates (rendering and input)                           */
/*     - time values (simulation, running, paused)                   */
/*     - input values (buttons, valuators, N-switches, 6-sensors)    */
/*     - location of the head & eyes                                 */
/*     - left/right eye rendering indicator                          */
/*     - relative coordinate calculations                            */
/* NOTE: based on debuging needs, some of these values may be        */
/*   excluded from particular compilations using #if 0.              */
float draw_text(vrRenderInfo *rendinfo, float y_start)
{
static	GLbyte		text_color[3] = { (GLbyte)155, (GLbyte)155, (GLbyte)155 };	/* medium gray */
	int		count;
	char		buffer[2048];
	vrTime		beginRenderTime = vrCurrentWallTime();/* Time when we begin the rendering process */
	vrTime		renderTime;	/* amount of time spent rendering */
	vrMatrix	head;
	vrProcessInfo	*input_proc;
	vrInputInfo	*inputinfo;	/* Pointer to the overall input data structure */
	float		text_y = y_start;/* Current y value at which to display next text message */
	float		text_height;	/* Height of a line of text in the current window */

	inputinfo = rendinfo->context->input;

	text_height = (float)((vrGlxPrivateInfo *)rendinfo->window->aux_data)->fontHeight / (float)(rendinfo->window->geometry.height);
	text_height *= 30.0;		/* A fudge-factor that seems to work */

#ifndef LIMITED_TRACE
	vrTrace("draw_text", "doing the text rendering");
#endif

	/* First store, then adjust some graphics attributes for text rendering */
	/*      disregard z-buffer -- always display the text on top */
	glPushAttrib(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

	/* Set the text color */
	glColor3ub(text_color[0], text_color[1], text_color[2]);


	/* Render on left wall */
	if (inputinfo->num_6sensors >= 1) {
		input_proc = inputinfo->sensor6[0]->my_device->proc;

		snprintf(buffer, sizeof(buffer), "Shared memory usage:%7ld (%ld freed)", vrShmemUsage(), vrShmemFreed());
		glRasterPos3f(-5.0, text_y, 2.75);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;

		snprintf(buffer, sizeof(buffer), "input info:%9ld %6.2f(1) %6.2f(10)",
			input_proc->frame_count, input_proc->fps1, input_proc->fps10);
		glRasterPos3f(-5.0, text_y, 2.75);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;

		snprintf(buffer, sizeof(buffer), "render info:%8ld %6.2f(1) %6.2f(10)",
			rendinfo->frame_count, vrThisProc->fps1, vrThisProc->fps10);
		glRasterPos3f(-5.0, text_y, 2.75);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	/* Render data on front wall */
	text_y = 9.0;
	snprintf(buffer, sizeof(buffer), "sim running time: %8.4f -- %8.4f (paused %8.4f)", (double)(rendinfo->frame_stime), vrCurrentSimTime(), vrContext->paused_time);
	glRasterPos3f(-4.5, text_y, -5.01);
	vrRenderText(rendinfo, buffer);
	text_y -= text_height;

	glRasterPos3f(-4.5, text_y, -5.01);
	snprintf(buffer, sizeof(buffer), "buttons:");
	vrRenderText(rendinfo, buffer);
	for (count = 0; count < inputinfo->num_2ways; count++) {
		snprintf(buffer, sizeof(buffer), " %d", vrGet2switchValueNoLastUpdate(count));
		vrRenderText(rendinfo, buffer);
	}
	text_y -= text_height;

	if (inputinfo->num_valuators > 0) {
		glRasterPos3f(-4.5, text_y, -5.01);
		snprintf(buffer, sizeof(buffer), "valuators:");
		vrRenderText(rendinfo, buffer);
		for (count = 0; count < inputinfo->num_valuators; count++) {
			snprintf(buffer, sizeof(buffer), " %6.3f", vrGetValuatorValueNoLastUpdate(count));
			vrRenderText(rendinfo, buffer);
		}
		text_y -= text_height;
	}

	if (inputinfo->num_Nways > 0) {
		glRasterPos3f(-4.5, text_y, -5.01);
		snprintf(buffer, sizeof(buffer), "nswitches:");
		vrRenderText(rendinfo, buffer);
		for (count = 0; count < inputinfo->num_Nways; count++) {
			snprintf(buffer, sizeof(buffer), " %d", vrGetNswitchValueNoLastUpdate(count));
			vrRenderText(rendinfo, buffer);
		}
		text_y -= text_height;
	}

	if (vrMatrixGetRWFromUserHead(&head, 0) != NULL) {
		vrPoint	*leye = &(vrRenderCurrentEye(rendinfo)->loc);

		snprintf(buffer, sizeof(buffer), "head: %5.2f %5.2f %5.2f", VRMAT_ROWCOL(&head, VR_X, VR_W), VRMAT_ROWCOL(&head, VR_Y, VR_W), VRMAT_ROWCOL(&head, VR_Z, VR_W));
		glRasterPos3f(-5.0, text_y,  4.50);
		vrRenderText(rendinfo, buffer);
		glRasterPos3f(-4.5, text_y, -5.01);	/* ditto this to left wall */
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;

		if (rendinfo->eye->type == VREYE_LEFT) {
			snprintf(buffer, sizeof(buffer), "leye: %5.2f %5.2f %5.2f", leye->v[VR_X], leye->v[VR_Y], leye->v[VR_Z]);
			glRasterPos3f(-4.5, text_y, -5.01);
			vrRenderText(rendinfo, buffer);
			text_y -= text_height;
		}

		if (rendinfo->eye->type == VREYE_LEFT) {
			/* At/near the front wall */
			glRasterPos3f(-4.5, text_y, -5.01);
			vrRenderText(rendinfo, "Left Eye");

			/* At/near the left wall */
			glRasterPos3f(-4.49, text_y, 0.5);
			vrRenderText(rendinfo, "Left Eye");

			/* At/near the right wall */
			glRasterPos3f( 4.49, text_y, -0.5);
			vrRenderText(rendinfo, "Left Eye");

			/* At/near the floor wall */
			glRasterPos3f( -0.5,  0.01, 0.0);
			vrRenderText(rendinfo, "Left Eye");
		} else {
			/* At/near the front wall */
			glRasterPos3f(-4.0, text_y, -5.01);
			vrRenderText(rendinfo, "Right Eye");

			/* At/near the left wall */
			glRasterPos3f(-4.49, text_y, 0.0);
			vrRenderText(rendinfo, "Right Eye");

			/* At/near the right wall */
			glRasterPos3f( 4.49, text_y, 0.0);
			vrRenderText(rendinfo, "Right Eye");

			/* At/near the floor wall */
			glRasterPos3f( -0.0,  0.01, 0.0);
			vrRenderText(rendinfo, "Right Eye");
		}
		text_y -= 2.0 * text_height;
	} else {
		snprintf(buffer, sizeof(buffer), "no head configured.");
		glRasterPos3f(-4.5, text_y, -5.01);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	if (inputinfo->num_6sensors >= 1) {
		vrMatrix	tmp_matrix;

		vrMatrixGet6sensorValues(&tmp_matrix, 0);
		snprintf(buffer, sizeof(buffer), "%csensor6[0]: %5.2f %5.2f %5.2f",
			(vrGet6sensorOobValue(0) ? ' ' : '>'),
			VRMAT_ROWCOL(&tmp_matrix, VR_X, VR_W), VRMAT_ROWCOL(&tmp_matrix, VR_Y, VR_W), VRMAT_ROWCOL(&tmp_matrix, VR_Z, VR_W));
		glRasterPos3f(-8.5, text_y, -5.01);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	if (inputinfo->num_6sensors >= 2) {
#if 1 /* set to "1" to add the OOB flag and 6-sensor[1] location to the on screen information display */
		snprintf(buffer, sizeof(buffer), "%csensor6[1]: %5.2f %5.2f %5.2f",
			(inputinfo->sensor6[1]->oob ? ' ' : '>'),
			VRMAT_ROWCOL(inputinfo->sensor6[1]->position, VR_X, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[1]->position, VR_Y, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[1]->position, VR_Z, VR_W));
#else
	/* Temporarily change this for a test on how to calculate a point relative to another sensor's coordinate system */
		vrMatrix	head_mat;
		vrMatrix	head_invert;
		vrPoint		wand_point;
		vrPoint		relwand;

		vrMatrixGet6sensorValues(&head_mat, HEAD_SENSOR);
		vrPointGetRWFrom6sensor(&wand_point, WAND_SENSOR);
		vrMatrixInvert(&head_invert, &head_mat);
		vrPointTransformByMatrix(&relwand, &wand_point, &head_invert);

		snprintf(buffer, sizeof(buffer), "wand rel to head: %5.2f %5.2f %5.2f", relwand.v[VR_X], relwand.v[VR_Y], relwand.v[VR_Z]);
#endif

		glRasterPos3f(-8.5, text_y, -5.01);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	if (inputinfo->num_6sensors >= 3) {
		snprintf(buffer, sizeof(buffer), "%csensor6[2]: %5.2f %5.2f %5.2f",
			(inputinfo->sensor6[2]->oob ? ' ' : '>'),
			VRMAT_ROWCOL(inputinfo->sensor6[2]->position, VR_X, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[2]->position, VR_Y, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[2]->position, VR_Z, VR_W));

		glRasterPos3f(-8.5, text_y, -5.01);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	if (inputinfo->num_6sensors >= 4) {
		snprintf(buffer, sizeof(buffer), "%csensor6[3]: %5.2f %5.2f %5.2f",
			(inputinfo->sensor6[3]->oob ? ' ' : '>'),
			VRMAT_ROWCOL(inputinfo->sensor6[3]->position, VR_X, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[3]->position, VR_Y, VR_W),
			VRMAT_ROWCOL(inputinfo->sensor6[3]->position, VR_Z, VR_W));

		glRasterPos3f(-8.5,  -0.5, -5.01);
		vrRenderText(rendinfo, buffer);
		text_y -= text_height;
	}

	glPopAttrib();

	text_y -= 2.0 * text_height;

#if 0
	/* Just print how long it takes to render this routine (frame). */
	renderTime = vrCurrentWallTime() - beginRenderTime;
	snprintf(buffer, sizeof(buffer), "frame render time: %8.8f", (float)(renderTime));
	glRasterPos3f(-4.5, text_y, -5.01);
	vrRenderText(rendinfo, buffer);
	text_y -= text_height;
#endif

#ifndef LIMITED_TRACE
	vrTrace("draw_text", "ending");
#endif

	return text_y;	/* return y-location of next text -- to pass down the chain */
}


/*******************************************************************/
/* draw_inputs_text(): render textual information about the inputs */
/*   of the system suitable for analyzing how well the inputs are  */
/*   working.  It may also be helpful when manually simulating     */
/*   inputs for another system.                                    */
/* Typically used by the "inputs" (nee valtest) program.           */
float draw_inputs_text(vrRenderInfo *rendinfo, float y_start)
{
//static	GLbyte	text_color[3] = { (GLbyte)155, (GLbyte)255, (GLbyte)155 };	/* gray */
//static	GLbyte		text_color[3] = { (GLbyte)155, (GLbyte)255, (GLbyte)155 };	/* gray-green */
static	GLbyte		text_color[3] = { (GLbyte)255, (GLbyte)255, (GLbyte)155 };	/* ?? */
	int		count;
	char		buffer[2048];
	vrTime		beginRenderTime = vrCurrentWallTime();/* Time when we begin the rendering process */
	vrTime		renderTime;		/* Amount of time spent rendering */
	vrMatrix	head;
	vrProcessInfo	*input_proc;
	vrInputInfo	*inputinfo;		/* Pointer to the overall input data structure */
	float		text_y = y_start;	/* Current y value at which to display next text message */
	float		text_height;		/* Height of a line of text in the current window */

	inputinfo = rendinfo->context->input;

#ifndef LIMITED_TRACE
	vrTrace("draw_inputs_text", "doing the text rendering");
#endif

	text_height = (float)((vrGlxPrivateInfo *)rendinfo->window->aux_data)->fontHeight / (float)(rendinfo->window->geometry.height);
	text_height *= 30.0;		/* A fudge-factor that seems to work */

	/* First store, then adjust some graphics attributes for text rendering */
	/*      disregard z-buffer -- always display the text on top */
	glPushAttrib(GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

	/* Set the text color */
	glColor3ub(text_color[0], text_color[1], text_color[2]);


	/* Render on front wall */
	glRasterPos3f(-4.5, text_y-=text_height , -5.01);
	snprintf(buffer, sizeof(buffer), "system: %s", rendinfo->context->config->system_name);
	vrRenderText(rendinfo, buffer);

	/* Render text information about each of the buttons */
	glRasterPos3f(-4.5, text_y-=text_height, -5.01);
	snprintf(buffer, sizeof(buffer), "buttons:");
	vrRenderText(rendinfo, buffer);
	for (count = 0; count < inputinfo->num_2ways; count++) {
		snprintf(buffer, sizeof(buffer), " %d", vrGet2switchValueNoLastUpdate(count));
		vrRenderText(rendinfo, buffer);
	}

	/* Render text information about each of the valuators */
	if (inputinfo->num_valuators > 0) {
		glRasterPos3f(-4.5, text_y-=text_height, -5.01);
		snprintf(buffer, sizeof(buffer), "valuators:");
		vrRenderText(rendinfo, buffer);
		for (count = 0; count < inputinfo->num_valuators; count++) {
			snprintf(buffer, sizeof(buffer), " %6.3f", vrGetValuatorValueNoLastUpdate(count));
			vrRenderText(rendinfo, buffer);
		}
	}

	/* Render text information about each of the N-ways */
	if (inputinfo->num_Nways > 0) {
		glRasterPos3f(-4.5, text_y-=text_height, -5.01);
		snprintf(buffer, sizeof(buffer), "nswitches:");
		vrRenderText(rendinfo, buffer);
		for (count = 0; count < inputinfo->num_Nways; count++) {
			snprintf(buffer, sizeof(buffer), " %d", vrGetNswitchValueNoLastUpdate(count));
			vrRenderText(rendinfo, buffer);
		}
	}

#if 0
	if (vrMatrixGetRWFromUserHead(&head, 0) != NULL) {
		vrPoint	*leye = &(vrRenderCurrentEye(rendinfo)->loc);

		snprintf(buffer, sizeof(buffer), "head: %5.2f %5.2f %5.2f", VRMAT_ROWCOL(&head, VR_X, VR_W), VRMAT_ROWCOL(&head, VR_Y, VR_W), VRMAT_ROWCOL(&head, VR_Z, VR_W));
		glRasterPos3f(-5.0, text_y-=text_height,  4.50);
		vrRenderText(rendinfo, buffer);
	} else {
		snprintf(buffer, sizeof(buffer), "no head configured.");
		glRasterPos3f(-4.5, text_y-=text_height, -5.01);
		vrRenderText(rendinfo, buffer);
	}
#endif

	/* Render text information about each of the 6-sensors */
	if (inputinfo->num_6sensors > 0) {
		vrMatrix	tmp_matrix;
		vrEuler		r2e_euler;

		glRasterPos3f(-4.5, text_y-=text_height, -5.01);	/* front wall */
		snprintf(buffer, sizeof(buffer), "6-sensors:");
		vrRenderText(rendinfo, buffer);
		for (count = 0; count < inputinfo->num_6sensors; count++) {
			glRasterPos3f(-4.5, text_y-=text_height, -5.01);	/* front wall */

			vrMatrixGet6sensorValues(&tmp_matrix, count);
			snprintf(buffer, sizeof(buffer), "  %csensor6[%d]: %5.2f %5.2f %5.2f",
				(vrGet6sensorOobValue(count) ? '!' : ' '), count,
				VRMAT_ROWCOL(&tmp_matrix, VR_X, VR_W), VRMAT_ROWCOL(&tmp_matrix, VR_Y, VR_W), VRMAT_ROWCOL(&tmp_matrix, VR_Z, VR_W));

#if 1 /* Add some additional info for the new "config" program */
			/* Set the text color based on the state */
			if (vrGet6sensorActiveValue(count))
				glColor3ub((GLbyte)0, (GLbyte)230, (GLbyte)0);	/* green */
			if (vrGet6sensorOobValue(count))
				glColor3ub((GLbyte)0, (GLbyte)0, (GLbyte)230);	/* blue */
			if (vrGet6sensorErrorValue(count))
				glColor3ub((GLbyte)230, (GLbyte)0, (GLbyte)0);	/* red */
			if (vrGet6sensorDummyValue(count))
				glColor3ub((GLbyte)252, (GLbyte)206, (GLbyte)100);	/* yellow */

			vrRenderText(rendinfo, buffer);
			vrEulerSetFromMatrix(&r2e_euler, inputinfo->sensor6[count]->r2e_xform);
			snprintf(buffer, sizeof(buffer), "  r2e: % .2f % .2f % .2f  % .2f % .2f % .2f", r2e_euler.t[VR_X], r2e_euler.t[VR_Y], r2e_euler.t[VR_Z], r2e_euler.r[VR_AZIM], r2e_euler.r[VR_ELEV], r2e_euler.r[VR_ROLL]);
#endif

			vrRenderText(rendinfo, buffer);

			snprintf(buffer, sizeof(buffer), "  (%s)", inputinfo->sensor6[count]->my_device->name);
			vrRenderText(rendinfo, buffer);

			glColor3ub(text_color[0], text_color[1], text_color[2]);/* original color */
		}
	}

	/* Render text information about each of the Input devices (indevs) */
	snprintf(buffer, sizeof(buffer), "Input devices: %d", inputinfo->num_input_devices);
	glRasterPos3f(-4.5, text_y-=text_height, -5.01);
	vrRenderText(rendinfo, buffer);

	/* Loop through the devices list */
	for (count = 0; count < inputinfo->num_input_devices; count++) {
		vrEuler		t2rw_euler;

		/* Warn for non-operating devices */
		if (inputinfo->input_devices[count]->operating < 1)
			glColor3ub((GLbyte)230, (GLbyte)0, (GLbyte)0);	/* red */

		glRasterPos3f(-4.5, text_y-=text_height, -5.01);
		snprintf(buffer, sizeof(buffer), "  %15s", inputinfo->input_devices[count]->name);
		vrRenderText(rendinfo, buffer);
#if 1 /* More additional info for the new "config" program */
		vrEulerSetFromMatrix(&t2rw_euler, inputinfo->sensor6[0]->my_device->t2rw_xform);
		snprintf(buffer, sizeof(buffer), ":  t2rw: % .2f % .2f % .2f  % .2f % .2f % .2f", t2rw_euler.t[VR_X], t2rw_euler.t[VR_Y], t2rw_euler.t[VR_Z], t2rw_euler.r[VR_AZIM], t2rw_euler.r[VR_ELEV], t2rw_euler.r[VR_ROLL]);
		vrRenderText(rendinfo, buffer);

		glColor3ub(text_color[0], text_color[1], text_color[2]);/* original color */
#endif
	}


	/* Restore the graphics state */
	glPopAttrib();

	renderTime = vrCurrentWallTime() - beginRenderTime;

#ifndef LIMITED_TRACE
	vrTrace("draw_text", "ending");
#endif

	return text_y;	/* return y-location of next text -- to pass down the chain */
}


/***********************************************************************/
/* draw_indev_axes(): draw a set of axes for each input device, using  */
/*   the t2rw_xform matrix to position the axes.                       */
void draw_indev_axes(vrRenderInfo *rendinfo)
{
	int		count;

	for (count = 0; count < rendinfo->context->input->num_input_devices; count++) {
		glPushMatrix();
		glPushAttrib(GL_LINE_BIT);
			glMultMatrixd(rendinfo->context->input->input_devices[count]->t2rw_xform->v);
			glLineWidth(3.0);
			draw_axes();
			glColor4ub(0, 255, 255, 255);		/* cyan */
			glRasterPos3f(0.1, 0.1, 0.1);
			vrRenderText(rendinfo, rendinfo->context->input->input_devices[count]->name);
		glPopAttrib();
		glPopMatrix();
	}
}


/********************************************************************/
/* draw_6sensor_axes(): draw a set of axes for each 6-sensor input, */
/*   using the r2e_xform matrix to show the final transformation.   */
/*   (actually, the r2e_xform (and t2rw) are accounted for in the   */
/*   sensor6->position matrix.                                      */
void draw_6sensor_axes(vrRenderInfo *rendinfo)
{
	int		count;

	for (count = 0; count < rendinfo->context->input->num_6sensors; count++) {
		glPushMatrix();
			vrRenderTransform6sensor(rendinfo, count);	/* this uses the frame-frozen "visren_position" copy of the position matrix */
			draw_axes();
			glColor4ub(252, 206, 100, 255);		/* yellow */
			glRasterPos3f(0.1, 0.1, 0.1);
			vrRenderText(rendinfo, rendinfo->context->input->sensor6[count]->my_object->name);
		glPopMatrix();
	}
}


/**********************************************************************/
/* enhanced_simulator(): add a couple of the features implemented by  */
/*   functions here in "drawing.c" to how the normal simulator is     */
/*   rendered.                                                        */
/* NOTE: after trying this, there's a good chance I'll just add these */
/*   features to be part of the standard simulator renderer.          */
void enhanced_simulator(vrRenderInfo *renderinfo, int mask) {

	draw_indev_axes(renderinfo);
	draw_6sensor_axes(renderinfo);
	vrRenderDefaultSimulator(renderinfo, mask);	/* Now render the rest as usual */
}


/***********************************************************************/
/* draw_vrui_outline(): take the polygon data from a Vrui.cfg file (in */
/*   the "screenProtectorAreas" setting), and draw those polygons in   */
/*   the world.  This will show where the tracked devices are relative */
/*   to the outline of the physical surroundings.                      */
/*                                                                     */
/* NOTE: of course, this is presently all hard-coded, and the data is  */
/*   hand massaged (in Vim) to provided the pre-defined values for the */
/*   point array.  And in one case, Oliver's data had a 7-vertex poly  */
/*   whereas all the others are 4-vertex polys (quads).                */
void draw_vrui_outline(vrRenderInfo *rendinfo)
{
	int		count;
	int		num_polys;			/* the number of polygons in the chaparone*/
	vrInputDevice	*vrui_input_device = NULL;	/* pointer to the first(!) VruiDD input device */

	for (count = 0; count < rendinfo->context->input->num_input_devices; count++) {
		vrui_input_device = rendinfo->context->input->input_devices[count];
		if (!strcmp(vrui_input_device->name, "vrui")) {
			break;
		}
	}

#if 0
	if (vrui_input_device != NULL) {
		printf("Found a VruiDD input device (%p): %d\n", vrui_input_device, count);
	}
#endif

	/* The following list of points is from Vrui.cfg:Vive:screenProtectorAreas */
	double		vrui_screenProtectorAreas[][4][3] = {
			 {{-0.21384540544106409, -1.19389563798904419, -2.56573233237633325},
			  {-0.242738088773142902, -1.19389563798904419, -2.56004057391997319},
			  {-0.242738088773142902, 1.30610436201095581, -2.56004057391997319},
			  {-0.21384540544106409, 1.30610436201095581, -2.56573233237633325}},
			 {{-0.242738088773142902, -1.19389563798904419, -2.56004057391997319},
			  {0.178975250791100882, -1.19389563798904419, -3.24536170678980218},
			  {0.178975250791100882, 1.30610436201095581, -3.24536170678980218},
			  {-0.242738088773142902, 1.30610436201095581, -2.56004057391997319}},
			 {{0.178975250791100882, -1.19389563798904419, -3.24536170678980218},
			  {1.51284636259078975, -1.19389563798904419, -2.24873127937316886},
			  {1.51284636259078975, 1.30610436201095581, -2.24873127937316886},
			  {0.178975250791100882, 1.30610436201095581, -3.24536170678980218}},
			 {{1.51284636259078975, -1.19389563798904419, -2.24873127937316886},
			  {0.163068781522187317, -1.19389563798904419, -0.408411564474756072},
			  {0.163068781522187317, 1.30610436201095581, -0.408411564474756072},
			  {1.51284636259078975, 1.30610436201095581, -2.24873127937316886}},
			 {{0.163068781522187317, -1.19389563798904419, -0.408411564474756072},
			  {-1.76475577428936958, -1.19389563798904419, -1.5970991924405098},
			  {-1.76475577428936958, 1.30610436201095581, -1.5970991924405098},
			  {0.163068781522187317, 1.30610436201095581, -0.408411564474756072}},
			 {{-1.76475577428936958, -1.19389563798904419, -1.5970991924405098},
			  {-1.24006732666131225, -1.19389563798904419, -3.09615379391294532},
			  {-1.24006732666131225, 1.30610436201095581, -3.09615379391294532},
			  {-1.76475577428936958, 1.30610436201095581, -1.5970991924405098}},
			 {{-1.24006732666131225, -1.19389563798904419, -3.09615379391294532},
			  {-0.21384540544106409, -1.19389563798904419, -2.56573233237633325},
			  {-0.21384540544106409, 1.30610436201095581, -2.56573233237633325},
			  {-1.24006732666131225, 1.30610436201095581, -3.09615379391294532}},
			 {{-0.242738088773142902, -1.19389563798904419, -2.56004057391997319},
			  {0.178975250791100882, -1.19389563798904419, -3.24536170678980218},
			  {1.51284636259078975, -1.19389563798904419, -2.24873127937316886},
			  {0.163068781522187317, -1.19389563798904419, -0.408411564474756072}},
#if 1 /* the above 4 were part of a 7-vertex poly, I've made a 2nd 4-vertex poly by duplicating the middle */
			 {{0.163068781522187317, -1.19389563798904419, -0.408411564474756072},
			  {-1.76475577428936958, -1.19389563798904419, -1.5970991924405098},
			  {-1.24006732666131225, -1.19389563798904419, -3.09615379391294532},
			  {-0.21384540544106409, -1.19389563798904419, -2.56573233237633325}},
#endif
			 {{-0.271134438839825742, -0.445708267664948865, -2.55488618937405665},
			  {-0.0709788133700688662, -0.445708267664948865, -2.9241770426432292},
			  {-0.762180562202747103, -0.445708267664948865, -3.34766059655409576},
			  {-0.963270466578634132, -0.445708267664948865, -2.95237227490073773}},
			 {{-1.31442650881680567, -0.452237510545687227, -3.18679274212230323},
			  {-1.54189217090606689, -0.452237510545687227, -3.24629497528076172},
			  {-1.66247087717056274, -0.452237510545687227, -2.77774305343627947},
			  {-1.46581322496587574, -0.452237510545687227, -2.65160400217229686}},

			};

	num_polys = sizeof(vrui_screenProtectorAreas) / (sizeof(vrui_screenProtectorAreas[0][0][0]) * 12);

#if 0
	printf("num_polys = %d\n", num_polys);
	printf("first point = (%lf, %lf, %lf)\n",
		/* First of the first group of four */
		vrui_screenProtectorAreas[0][0][VR_X],
		vrui_screenProtectorAreas[0][0][VR_Y],
		vrui_screenProtectorAreas[0][0][VR_Z]);
	printf("second point = (%lf, %lf, %lf)\n",
		/* Second of the first group of four */
		vrui_screenProtectorAreas[1][0][VR_X],
		vrui_screenProtectorAreas[1][0][VR_Y],
		vrui_screenProtectorAreas[1][0][VR_Z]);
#endif
	glColor3ub((GLbyte)252, (GLbyte)206, (GLbyte)100);	/* yellow */
	glPushMatrix();
		if (vrui_input_device != NULL) {
#if 0 /* temporarily remove this to hand-tweak wand-avatar representations in calling function */
			/* The Vrui.cfg data is in the original Vrui CS, so if known, shift to that CS. */
			glMultMatrixd(vrui_input_device->t2rw_xform->v);/* shift to VruiDD input device CS */
#endif
			/* Ideally, the scale operation would be somehow retrievalble through standard */
			/*   function calls, or alternatively actually integrated into t2rw_xform, but */
			/*   neither is presently the case, so we have to hardcode the scaling factor. */
			glScalef(3.28, 3.28, 3.28);			/* scale from meters to feet */
		} else {
			/* If I have a configuration matrix, then it will do the scaling (or so I thought) */
			glScalef(3.28, 3.28, 3.28);			/* scale from meters to feet */
		}
		//glTranslatef(0.176255133002996445, 0.00610436201095554409, -1.75264239311218262);	// From Vrui.cfg displayCenter()

		for (count = 0; count < num_polys; count++) {
			glBegin(GL_LINE_STRIP);
				glVertex3dv(vrui_screenProtectorAreas[count][0]);
				glVertex3dv(vrui_screenProtectorAreas[count][1]);
				glVertex3dv(vrui_screenProtectorAreas[count][2]);
				glVertex3dv(vrui_screenProtectorAreas[count][3]);
				glVertex3dv(vrui_screenProtectorAreas[count][0]);
			glEnd();
		}
	glPopMatrix();
}

