/* ======================================================================
 *
 *  CCCCC          vr_input.vrpn.c
 * CC   CC         Author(s): Bill Sherman
 * CC              Created: May 11, 2013 (copied from vr_input.vrpn.cxx)
 * CC   CC         Last Modified: January 17, 2024 (minor cleanup 02/11/24 & tweak 03/04/24)
 *  CCCCC
 *
 * Code file for FreeVR inputs from the VRPN input device.
 *
 * Copyright 2024, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

NOTE: this is the C version of the VRPN FreeVR interface, and unlike the C++
	version, it does not require the use of the VRPN library.  Instead it
	directly interprets the incoming VRPN stream.  But otherwise the
	end-user experience should be exactly the same.

USAGE (FreeVR):
	Of course, to use VRPN as an input device, the VRPN server must
	be up and running prior to starting the FreeVR application.

	Here is an example of how to run VRPN:
		(from the "/vr/apps/vrpn/current/vrpn" directory)
			A% .../bin/vrpn_server -f vrpn.myconfig.cfg [-q]

	And to test a running VRPN server from another shell:
			B% .../bin/printvals <vrpn-server-machine>
		  or	B% .../bin/vrpn_print_messages <vrpn-server-machine>
		  or	B% .../bin/vrpn_print_devices DTrack@tcp:<vrpn-server-machine>
		  or	B% .../bin/vrpn_orientation DTrack@tcp:<vrpn-server-machine>
		  or	B% .../bin/vrpn_print_performance SpinAE@tcp:127.0.0.1 SpinAz@tcp:127.0.0.1

	NOTE: it used to be that two slashes were required after "tcp://<host>",
		but it seems that is now optional -- just "tcp:<host>"
		[Though ParaView's "VR Plugin" was requiring them until
		ParaView version 5.10.0.]


	Inputs are specified with the "input" option:
		input "<logical-name>" = "<freevr-type>(<device-name>:<VRPN-type>[<input-num>])";

		input "2switch[left]" = "2switch(Joystick0:button[0])";
	  :-(	input "2switch[hat-x]"= "2switch(Joystick0:analog[4], threshold)";
		input "valuator[x]"   = "valuator(Joystick0:analog[3])";
		input "valuator[y]"   = "valuator(Joystick0:analog[-4])";
		input "6sensor[head]" = "6sensor(Tracker0:tracker[0, r2e])";
		input "6sensor[wand]" = "6sensor(Tracker0:tracker[1, r2e])";
		input "6sensor[wand2]" = "6sensor(sim6[1])";

		input "2switch[yo]"   = "2switch(Tracker0:button[1])";

		future possibilities:

	  :-(	input "<logical-name>" = "Nswitch(<device>:switch[<number>])";
	  :-(	input "<logical-name>" = "Nsensor(<device>:glove[<number>])";


	Controls are specified with the "control" option:
		control "<control option>" = "2switch(<device-name>:<VRPN-type>[<num>])";

		control "print_devinfo" = "2switch(Joystick0:button[10])";
		control "sensor_next" = "2switch(Joystick0:button[11])";
	  ...

	Here are the available (2switch) control options for FreeVR:
		"system_pause_toggle" -- toggle the system pause flag
		"print_help" -- print info on how to use the input device
		"print_context" -- print the overall FreeVR context data structure (for debugging)
		"print_config" -- print the overall FreeVR config data structure (for debugging)
		"print_input" -- print the overall FreeVR input data structure (for debugging)
		"print_devinfo" -- print the internal VRPN data structure (for debugging)
		"print_sim6opts" -- print the internal data for simulated 6-sensors (for debugging)

		"sensor_next" -- jump to the next 6-sensor on the list
		"setsensor(<num>)" -- set the simulated sensor to a particular one
		"sensor_reset" -- reset the current 6-sensor
		"sensor_resetall" -- reset the all the 6-sensors
		"temp_valuator" -- temporarily disable 6-sensor for valuator values
		"toggle_valuator" -- toggle valuator values instead of 6-sensor
		"temp_valuator_only" -- temporarily use translation values for valuator
		"toggle_valuator_only" -- toggle whether translation is saved for valuator
		"toggle_relative" -- toggle whether movement is relative to sensor's position
		"toggle_space_limit" -- toggle whether 6-sensor can go outside working volume
		"toggle_return_to_zero" -- toggle whether return-to-zero operation is on
		"toggle_swap_transrot" -- toggle whether to swap translation and rotation inputs
		"toggle_swap_yz" -- toggle the swapping of Y and Z translational movements

	Here are the available (valuator) control options for FreeVR:
		"set_transx" -- set the translation value for the simulated 6-sensor
		"set_transy" -- set the translation value for the simulated 6-sensor
		"set_transz" -- set the translation value for the simulated 6-sensor
		"set_azim" -- set the azimuth rotation value for the simulated 6-sensor
		"set_elev" -- set the elevation rotation value for the simulated 6-sensor
		"set_roll" -- set the roll rotation value for the simulated 6-sensor

	Here are the FreeVR configuration argument options for the VRPN server:
		"host" (or "hostname") - string of the machine running the VRPN server.
		"port" - socket port of the VRPN server (3883 is the VRPN default).
	 	"valScale" - float value of the valuator sensitivity scale
		"scale" (or "trackScale") - float value by which to scale the
			tracking location data.
			[It *no longer* (defaults to 3.28084 to convert from
			meters to feet because 1.0 should be the default!).]
		"govern_time" - boolean flag that indicates whether the VRPN
			data stream timestamp should be used to regulate the
			usage of data to match.  This is off by default because
			for normal usage of real-time data the timestamp should
			be the same as when it arrives.  However when recorded
			VRPN data is used, then the stream may flow faster than
			when initially recorded, making it generally better to
			regulate the flow by setting this to "on".

	 -- These options are for using the VRPN device as a simulated 6-sensor:
		"restrict" - limit simulated 6-sensors to move within the working-volume
		"valuatorOverride" - produce valuator values instead of moving
			the simulated 6-sensor
		"returnToZero" - valuator offsets map to absolute or relative
			position of simulated 6-sensor
		"relativeRot" - rotation of simulated 6-sensor is based on
			sensor's or world's coord-sys
	  	"swaptransRot" -- swap the translational and rotational inputs
			on the simulated 6-sensor
		"workingVolume" - orthogonally defined volume specifying the
			valid range of simulated 6-sensor movement
		"transScale" - scaling factor used to tune the translational
			movement of a simulated 6-sensor
		"rotScale" - scaling factor used to tune the rotational
			movement of a simulated 6-sensor


	NOTE: VRPN inputs cannot be combined with an X-windows configuration
		that uses extension devices of X.  It's not clear why this is,
		it just is.  [subnote: probably this only pertains to the
		VRPN-library version (which, by default, is not compiled).]

USAGE (vrpntest):
	The official VRPN testing program that sends out a variety of types
	of information packets is "test_vrpn".  And, perhaps unfortunately,
	it is not automatically installed via the VRPN build's "make install"
	command.
		% vrpn-<version>/<Cmake-build-directory>/server_src/test_vrpn


HISTORY:
	15 March 2000 (Bill Sherman)
		I used (newly created) vr_input.skeleton.c to help create the
		  initial version.

	12 April 2001 (Bill Sherman) -- mid mod
		I began getting all the proper variables, etc. laid out.
		  At this point, it compiles, and calls the input creation
		  routines for the inputs -- though doesn't really create
		  them yet.  But at least it works with the mixed C++ dso and
		  normal C application (using travel_n32), and it does call
		  the input creation routines.

	30 April 2001 (Bill Sherman) -- mid mod
		I managed to get the Button and Analog inputs working.  First
		  problem solved was find the cause of the VRPN server hangs --
		  oddly enough an Xwindows function: XListInputDevices().  By
		  not calling this, things work pretty well.

	 3 May 2001 (Bill Sherman) -- mid mod
		I now have tracker sensor inputs working, and have added
		  new 'print_context' and 'print_config' control options
		  not only here, but to all input devices.

		The BIG NEWS is that after getting the basics to work, I
		  decided to try getting the dso to work with a normally
		  compiled C application.  And it now works!  Hooray for
		  Captain Spaulding!

	 4 May 2001 (Bill Sherman)
		I added the r2e transform option to the 6-sensor inputs, and
		  a scale factor argument for valuators.
		I also did some general cleanup, since this file is mostly
		  working now -- though still need to add Phantom, etc.

	------------------------------------------------------------------
	17 December 2012 (Bill Sherman)
		I made huge progress in reverse-engineering the VRPN protocol.
		  Details of this are in my "impl_diary" journal.

	January 2013 (Bill Sherman)
		I continued reverse-engineering the VRPN protocol while
		  developing the initial "vrpntest" program.

		On the 18th I found a #define for "vrpn_FILTER_MESSAGES", which
		  is the reason why the client sends device requests, and why
		  that is totally ignored -- because the library is never
		  compiled with that flag turned on!  (That's also the day
		  I found the "Format" file!)

	11 May 2013 (Bill Sherman)
		After an initial foray into creating from scratch a new VRPN
		  reader written in C and directly interpreting the VRPN protocol
		  based on my VruiDD input device, today I decided that I could
		  reuse most of what already was written for the C++ VRPN reader
		  based on the actual VRPN library.  Except now entirely
		  independent of the actual VRPN library, so no extra libraries
		  need to be linked with a FreeVR application.

	17 May 2013 (Bill Sherman)
		In comparing the original VRPN C++ code with the input-skeleton
		  code, I found that there was some updating that needed to be
		  done.  So I did it.

		Also began working on bringing in my non-library Vrpn
		  communication code.

	17-23 May 2013 (Bill Sherman)
		At this stage, enough has been done to initiate communications
		  with a VRPN server and exchange version numbers.  I am doing
		  things a little differently with this input device in that
		  I'm using the 5 standard device callbacks.  As this requires
		  a vrInputDevice type, I've created a shell version of
		  vrInputDevice that has the bare minimum information for the
		  test application.

	28-29 May 2013 (Bill Sherman)
		I copied the "_VrpnReadInput()" routine from the stand-alone
		  "vrpntest.c" program into here, and from that got the basic
		  communication working.  Still need to convert it into FreeVR
		  format.

	12-16 July 2013 (Bill Sherman)
		I added storage of the incoming data into the auxiliary data
		  structure.  Tracking "posquat" data, button data and
		  analog/valuator data are now stored in a VRPN_Device structure,
		  and an array of those structures is part of the auxiliary
		  structure.

		I added flags for each input to indicate whether it has been
		  received since the last input processing operation, or has been
		  received at all since the beginning of the program.

		I also now store the name of devices as reported by the VRPN
		  server, and I now store all the types of messages reported by
		  the server, not just tags for the 5 messages I'm especially
		  looking for.

		The "vrpntest" application now basically works.  [In line
		  rendering mode.]

	22-25 July 2013 (Bill Sherman)
		I revamped the FreeVR side of the device handler (ie. for the
		  no VRPN client library version).  First, just to get it to
		  compile, then to get it to print out all the incoming values.
		  Presently it now handles button presses as inputs and as
		  self-controls.

	30 July 2013 (Bill Sherman)
		Tracker (aka 6sensor) inputs have now been implemented!  Thus
		  we now have a fully working version of the VRPN input system.
		  I also added some robustness to device-name access (making sure
		  there's not a NULL pointer first).  [NOTE: though there still
		  is some robustness that needs to be added in other areas.]
		  I also added some new arguments including the surprisingly
		  missing "host" argument (which can now also be referred to as
		  "hostname"), plus "scale"/"transScale" and "valScale".
		[06/06/22: "transScale" renamed to "trackScale".]

	9 August 2013 (Bill Sherman)
		I improved the robustness of the FreeVR VRPN interface.
		  Specifically, I now do a better job of handling the case where
		  an unknown host is provided as the VRPN server.  Also I better
		  handle the situation where the FreeVR configuration requests a
		  VRPN device that doesn't exist (and thus resulting in a case
		  where aux->button_map_device == NULL, which obviously can't be
		  dereferenced).  So I now print warnings when this happens.

	2-5 September 2013 (Bill Sherman)
		Adopting a consistent usage of str<...>cmp() functions (using
		  logical negation).

		Also, implemented simulated 6-sensors for VRPN.  This is the
		  first time there is a device that can have both actual 6-sensor
		  inputs as well as simulated inputs.  Thus there is a hack-ish
		  method used to differentiate between them.  There already was
		  a VRPNdevice[] array that listed all the devices coming from
		  the vrpn_server, so device 0 (VRPNdevice[0]) is now reserved
		  as the internal "simulator" device and thus does not come from
		  the vrpn_server inputs.

		However, apart from the "name" and "newly_reported_posquat"
		  fields, it doesn't really use the contents of VRPNdevice, but
		  is more of a placeholder.  Because simulated 6-sensors are
		  (presently!) designed to just have one per device (with a
		  selection of active 6-sensors), and generally don't have button
		  or valuator values -- because those don't need to be simulated!

		I also added the ability to quit "vrpntest" when pressing
		  buttons 1 & 2 of any VRPN device -- was easier than watching
		  for exactly the first two overall buttons, and in reality will
		  work better too.

	15 September 2013 (Bill Sherman)
		Renaming "active_sensor" to "active_sim6sensor".
		Changed the "0x%p" format to the improved "%#p" format.

	19 September 2013 (Bill Sherman)
		Added "-p"/"-port" command line argument to "vrpntest".
		Wrote the man page.

	10 November 2013 (Bill Sherman)
		Adjusted the inclusion of "endian.h" for APPLE compilations.

	16-21 October 2015 (Bill Sherman)
		Began preparation work for adding a ScreenRender mode to
		  "vrpntest".

		Also (19th) fixed a bug in how the "vrpntest" arguments were
		  processed that resulted in a memory bounds issue and a
		  seg-fault.

		In fact, there were a handful of issues that still really
		  had to be resolved.

	21 October 2015 (Bill Sherman)
		Now began actual ScreenRender coding (_RenderValuesScreen)
		  for "vrpntest".

		As part of this, I added new fields to the "VRPNdevice"
		  structure to indicate when any input of a particular type
		  ("button", "analog", "posquat" or "velquat" -- which isn't
		  used) is received for a given VRPN device.  This then
		  allows me to not set aside screen space for data that
		  doesn't exist -- and saves me from having to loop through
		  all the buttons, etc. to see whether a device got any
		  button (etc.) information.

	21-22 October 2020 (Bill Sherman) [5 years to the day from last change]
		Bug fixes primarily to correct some memory issues that came
		  to light with a higher number of devices reported by the
		  vrpn_server.  Also made adjustments to the screen renderer
		  display for "vrpntest".  (More work needs to be done on the
		  screen renderer.)

	29 October 2020 (Bill Sherman) [5 days to the day from last change!]
		Made adjustments to the screen renderer.  It is still not
		  rendering the rotations of the position trackers.

	26-30 November 2020 (Bill Sherman)
		I did some cleanup work to match this file better with the
		  "vr_input.skeleton.c" file (along with "vr_input.vruidd.c").

	22 January 2021 (Bill Sherman)
		Minor rename of _VrpnPrintStruct() to _VrpnPrintPrivateInfo(),
		  and "print_struct" control command to "print_devinfo".

	25 January 2021 (Bill Sherman)
		Adding code to output Euler angles in the screen rendering,
		  converted from the input Quaternions.  That does require
		  some routines from vr_math.c, which means that file will need
		  to be "included".  It also means that "-lm" is required when
		  compiling "vrpntest".

	7 April 2021 (Bill Sherman)
		Adopted new ANSI_... #defines for ANSI control sequences that
		  are useful for doing screen rendering -- including setting the
		  top and bottom margin (which now works) to push all debugging
		  output away from the screen rendering.

		NOTE: initially I was going to change all the existing ANSI
		  #define macros as well, but I'm going to save that for a code
		  cleanup round when shifting to a new minor version number.

		I now have the screen-rendering mode displaying rotations both
		  in "AER" and "ijkW" (quaternion) format, which allows me to
		  better judge that things are working properly.

		Added the top/bottom margins for debugging output.  Note that to
		  place debugging output into the proper location on the screen,
		  surround the output-producing code with:
			vrPrintf(ANSI_CURSORLOC_RETURN);  // before debug output
			vrPrintf(ANSI_CURSORLOC_SAVE);    // after debug output

		[corresponds to FreeVR Version 0.6g]

	4 May 2021 (Bill Sherman)
		Did a clean-up of some of the old code.  There's still some
		  old code left, but clearing it a little at a time.

	5 May 2021 (Bill Sherman)
		Minor clean-up to remove sprintf() and make sizeof() safer.

		[corresponds to FreeVR Version 0.7b]

	31 August - 2 September 2021 (Bill Sherman)
		Just fixing some comments and minor things here and there as I
		  develop the VRPN output server.

		Also, I now handle the "vrpn_Button States" message which takes
		  all the button values at once.  (Generally used during startup.)

		[corresponds to FreeVR Version 0.7c-dev]

	21 September 2021 (Bill Sherman)
		I fixed a couple of small but important bugs: 1) the screen
		  rendering was scaling the numeric location values based on the
		  size of the terminal, so that was removed.  And 2) the default
		  translational scale was set to 3.28 for meters-to-feet scaling,
		  but a default should always be 1.0, and let the configuration
		  change it to other values!

		[corresponds to FreeVR Version 0.7c]

	27 October 2021 (Bill Sherman)
		Added the 6-sensor initial position ("ipos") code to place
		  them at a reasonable default location prior to making a
		  network connection.

		Also did some general clean-up and changes to more closely
		  match other position tracker input devices.

		[corresponds to FreeVR Version 0.7c]

	28 October 2021 (Bill Sherman)
		Improved the "vrpntest" screenrendering output when X/Y/Z go
		  off the top of the screen.

		[corresponds to FreeVR Version 0.7c]

	29 October 2021 (Bill Sherman)
		Added the ability to parse the VRPN ping and text_message
		  messages.

		[corresponds to FreeVR Version 0.7c]

	26 November 2021 (Bill Sherman)
		Created the ability to delay VRPN input parsing until the
		  proper amount of time has elapsed.  This allows inputs that
		  were "recorded" to be played back at real-time rather than
		  as-fast-as-parsable.
		NOTE: This was done as a separate "vrpnfile" input-device type
		  until it has been robustly handled.

		Added VRPN Text_Messages as an "input" that is reported by
		  "vrpntest".  [11/27/21: Added Text Message output to the
		  ScreenRender (-screen) output option -- and so only clear
		  a portion of the screen before re-rendering.]

		Added the "CSV" (comma-separated-values) optional rendering
		  style to "vrpntest".  [11/27/21: Added time to the CSV output,
		  and removed spaces.]

		[corresponds to FreeVR Version 0.7d]

	4-6 April 2022 (Bill Sherman)
		Made the protocol parsing more robust -- especially when
		  dealing with some of the mal-formed data that the VRPN
		  "test_vrpn" program produces at the beginning.  It seems
		  that the parsing will now be more robust, but there may
		  still be some cases where bad data causes issues.

		Tweaked the screen-rendering a little to have location values
		  that are better aligned with the text-rendering window.

		Added the parsing/watching of Pong messages.

		I figured out, and documented, what the 8th value of the
		  velocity and acceleration reports are (a time delta).  Also,
		  discovered, and documented, that the 4-byte padding in the
		  header is actually used as a message count!

		[corresponds to FreeVR Version 0.7d]

	6 June 2022 (Bill Sherman)
		I added a run-time option ("govern_time") to control whether
		  the incoming VRPN data's timestamp should be used to regulate
		  the rate at which input data is processed.  (This is primarily
		  for use with pre-recorded VRPN data which thus might be sent
		  at a rate faster than real-time.)

		I renamed the "transScale" that applied to the overall tracking
		  to now be "trackScale".  Note that there was another option
		  called "transScale" that affected the simulated 6-dof input,
		  so the two options would overwrite eachother.

		[corresponds to FreeVR Version 0.7d]

	17 January 2024 (Bill Sherman)
		I fixed a known, but semi-tolerable bug where the initial
		  button values had semi-random values rather than the actual
		  0/1 values (usually 0 on initialization).  This was causing
		  some applications to trigger actions on startup, so I
		  finally spent the time (2-3 hours!) debugging this.  Was
		  just a case of donig an 8-byte endian swap instead of 4 byte.

		[corresponds to FreeVR Version 0.7e]

	4 March 2024 (Bill Sherman)
		Fixed some very minor bugs in printf() formatting & one
		  erroneous conditional -- during release testing.

		[corresponds to FreeVR Version 0.7e (post)]

TODO:
	- DONE: Handle the "vrpn_Button States" message which reports all the
		buttons at once (I think).  I think this is done to send the
		initial values for all the buttons.

	- Create a configuration option that allows VRPN data to be read from a
		file rather than from a socket -- presently we do a work-around
		whereby a file is converted into a socket stream using "nc".

	- DONE: Figure out why some initial button values are random integers
		(at least for the screen rendering)
		[Fixed 01/17/2024]

	- Figure out why the last valuator seems to have its has-been-reported
		value set to true before it's received any data.  (A good
		test is with a Razer Hydra.)

	- Have "vrpntest" return a success/failure value when terminating
		(either because a limited-data option was given, or the
		application was terminated).  This should be based on whether
		a successful connection to a server was established.

	- DONE: Report VRPN Text_message messages in "vrpntest".
		[11/27/21: Also added this feature to the "-screen" option.]

	- DONE: Parse/Watch Pong messages (do similar to text_message)

	- For the unwatched types: output the payload as 4-byte integers,
		8-byte integers, doubles, and ASCII (in addition to the
		1-byte hex that I'm doing now).

	- DONE: For "vrpntest -screen" (screen-render) output VRPN-text messages
		in a special area -- perhaps also include unwatched messages
		there as well.

	- DONE: Add the CSV output to "vrpntest" -- this can be used to
		translate recorded VRPN data in a file, into a CSV dataset
		that can be analyzed.

	- MOSTLY: Add a "vrpn-file" playback feature (first doing as a separate
		inputdevice).

	- DONE: Add time regulation (from a vrpn-file playback) to "vrpntest"

	- Add a flag to enable/disable time-regulation for "vrpntest" (I
		don't want it when converting saved VRPN data into a CSV file
		for example).

	- Add the ability to scale the location values as a command-line-arg.
		(Useful for the "-screen" rendering option to keep the values
		in range.)

	- MAYBE: In "-screen" mode, there are times when the sensor component
		characters have negative values, and so then are mis-rendered.
		[04/06/22: I think I've fixed this already -- need to confirm]

	- Confirm whether the comment in quatlib/quat.h that claims that
		rotations are mapped such that:
			- Q_YAW (rotation about Z)
			- Q_PITCH (rotation about Y)
			- Q_ROLL (rotation about X)
		I'm hoping this is not true, because that doesn't match most
		common coordinate systems -- X & Y are swapped from the norm.

		[04/08/21: but that could explain the issue I'm having with
		the Razer Hydra.]

	- parse the log-level value in the "version" string (VRPN documentation
		calls this the "cookie").  Presently just expect a "0" value.

	- Add an ASCII-poly rendering mode to "vrpntest -screen"

	- Make handling of lack of available simulated 6-sensors robust!
		Basically what this means is adding a conditional to all
		of the self-control callbacks that make use of the
		"active_sim6sensor" value.

	- I'm concerned that the instance numbers are not being handled
		correctly.  Specifically, they are meant to indicate the
		number of the button/analog/tracker on the VRPN server side,
		but they may also be inadvertently being used as the overall
		button/analog/etc number on the FreeVR side, and that's not
		correct -- it needs to be handled on a device by device basis.

		So first thing to do is some testing.

	- DONE: I may need to handle simulated 6-sensors separate from trackers
		coming from the VRPN server.  (In fact, this may be a bigger
		problem even if I have two joystick devices on the server!)

	- DONE: robustly handle the unknown host circumstance

	- DONE: robustly handle the situation of (aux->button_map_device == NULL)

	- DONE: print a stern warning when VRPN devices listed in ".freevrrc"
		are not being sent from the VRPN server (which is the cause of
		the above problem).

	- Add the capability of having dual-purpose for valuators as both
		normal valuator input or as simulated 6-sensor controls.
		(See "ignore_trans" and "ignore_all" in the Magellan and
		Spacetec drivers.)

		Perhaps I would have a configuration line:
			"val[x]" = "valuator(sim6[transx|VR_X|X]);"
		which would mean that valuator arbitrarily named "val[x]" gets
		it's value from the VR_X component of the simulated 6-sensor
		values in "aux->sim_values[]" -- of course only when the
		temp_valuator flag is set (aka "ignore_trans" or "ignore_all").

		Actually, looking at the code, for the Magellan and Spacetec
		devices it's always just mapped to VR_X and VR_Y.  But this way
		would be more flexible.

	- find out whether "vrpn_Forwarder_Brain" is always the last device
		sent by the server, and if so, use that as a flag
		(see questions, below).
		[10/22/2020: I found that the vrpn_server may think it has
		other devices when it receives requests for non-existent (i.e.
		un-configured devices).  For some reason it then remembers
		these -- and they are listed after "vrpn_Forwarder_Brain".]

	- consider how to use the mapped_button/mapped_analog/mapped_tracker
		VRPNdevice fields (or get rid of them).  (I should at least be
		printing them somewhere.)

	- DONE: Add a "-nodata" CLA to "vrpntest" that allows one to just get
		a report of what the VRPN server is serving.

	- Consider adding a "-repunk" CLA to "vrpntest" to REPort UNKnown inputs

	- DONE: Add a screen-based (pseudo-curses) rendering option of the
		tracking data to "vrpntest" (might be good to have a means of
		rendering a quaternion representation)

	- Update the _VrpnPrintStruct() routine to report the stored input data.
		[01/22/2021: renamed to _VrpnPrintPrivateInfo()]

	- Test whether there are VRPN messages that can be sent back to the
		server! [11/06/2020: I asked Russ & Ryan about this, but
		didn't really get an answer on this one.  I may just start
		digging into the VRPN code to find out.]
		[11/27/21: For FreeVR utility, this would require that the
		VRPN input be able to handle full-duplex information.]

	- figure out what values to assign the 'version' and
		'operating parameter' fields

	- attempt to get the VRPN_LIB version working as well!

	- DONE: delete old DONOTMOVETO_INITDEV code

	- DONE: determine whether the _VrpnUserdata structure (and affiliated
		code) are required for the non VRPN_LIB version.  [Answer:
		pretty sure that it is only for the VRPN_LIB version, so now
		conditionally removed.]

	- add "phantom" type ??

	- PARTIAL: allow a FreeVR-created 6-sensor from valuator inputs to be
		created.  (ie. can either use the VRPN server options for
		creating a "tracker" from analog inputs, or can use the
		standard FreeVR method such as is available in the raw
		Magellan code).  In the X11-windows input this is the
		"sim6" input type.

	- determine whether old analog values can be saved, and compared with
		incoming analog values, reporting only those that have changed.
		NOTE: I do make use of the fact that VRPN only sends data as it
		determines they are new, so I may not need to do anything beyond
		flagging which data has recently been received -- which I now do.

	- PARTIAL: Add information about the VRPN protocol to the top-comments.
		[10/22/2020: and maybe the man page as well! (or a separate
		man page.)]
		[11/27/2021: I've recently added much of this information to the
		"vr_output.vrpn.c" source file.]

VRPN PROTOCOL NOTES:
	[See "vr_output.vrpn.c" for a more complete (but not 100%) protocol description.]

	VRPN Coordinates
		VRPN reports 6-DOF positions as X,Y,Z,Qi,Qj,Qk,Qw
		VRPN location positions are reported in meters
		VRPN translation velocities are reported in m/s
		VRPN accelerations are reported in m/s/s

	The position tracker data provided by VRPN is reported to be in
	VQS order (vector/quaternion/scale), with the quaternions calculated
	by taking the rotations in the order:
		- Q_YAW (rotation about Z)
		- Q_PITCH (rotation about Y)
		- Q_ROLL (rotation about X)

VRPN HISTORY NOTES:
	I'm not going to say much here, but based on an e-conversation with
	Russ Taylor, I understand that VRPN basically came to be in early
	1997, with some proto-work done in 1996.  Some of it's features were
	adopted from code from the UNC TrackerLib, and likewise, some of it's
	capabilities were added back to TrackerLib for the sake of existing
	applications at UNC.

	VRPN was initially in the public domain, but prior to 2013 it was put
	under the Boost licence.  Those who release the source code are
	required to give credit, but not if they only release binaries.

	There is a C-wrapped version that came out prior to 2013.  It was
	created to allow VRPN to be merged into projects that are not based
	on C++ (like FreeVR!).  However, I'd already figured out how to parse
	the VRPN protocol, so FreeVR can read VRPN without even needing to be
	linked to a C-library version of VRPN.

VRPN UNKNOWNS:

	- DONE: What is the 8th float value reported by VelQuat?
		01/26/2021: I wonder whether it might be scale -- ala VQS format
		08/27/2021: though it really makes no sense to have a scaled
			position tracking value.
		04/04/2022: Looking at the VRPN code, it's an interval delta (of
			time) -- so I'm calling it "velquat_deltat" here.

	- What does the "VRPN Control" device do?
		(and is there an example tool that generates these device events?)

		Ryan Pavlik suggested looking at his Rust implementation of the
		VRPN protocol -- See GitHub.com/vrpn/vrpn-rs

		Russ says: vrpn_Control is the sender used for notification
		messages sent to the user from the local VRPN implementation
		(got_first_connection, etc.) as opposed to messages sent from
		a named device.  These won't go over the network connection.

		There are different messages from the client to the server for
		each device type, and it is also possible for the user to define
		new types of messages.  Some of this is described in the [VRST]
		attached paper.  There used to be clock-synchronization messages
		during connection set up, but those were removed along the way
		as NTP became pervasive.

		The forwarder controller is described at
			https://github.com/vrpn/vrpn/blob/master/vrpn_ForwarderController.h
		and it is not part of a standard VRPN system.  Tom Hudson made
		these, along with Model-View-Controller and Mutex classes to
		share objects and forward connections as part of his
		dissertation work on a distributed, collaborative
		nanoManipulator.

		The Connection, BaseClass, and various device classes
		(Tracker, Button, Analog) are all that should be required to
		get a system operational.  Ryan's Rust implementation should
		show that pretty well, along with the various wrapper approaches
		(Java, Python, Android) in subdirectories.


		Russ (12/19/13) also says: The main server code starts a
		forwarder server whenever it is run.  I don't know that you'll
		need this in your client, so you should be able to ignore any
		messages to/from that device.


	- What does the "vrpn_Forwarder_Brain" device do?
		(and is there an example tool that generates these device
		events?)
		- and perhaps most importantly!  Will this always be the last
			device sent by the server?  (If so we can use this as
			a flag to know that we've received all the devices.)

		Russ (12/13/13): The vrpn_Forwarder_Brain has to do with a
		vrpn_Forwarder class, which I think basically nobody uses
		anymore.  It takes messages from one VRPN connection and sends
		them out over another one.  You can probably safely ignore all
		such messages.

		Russ (12/19/13) also says: I'm actually a bit vague on what the
		forwarder is used for; it has been a long time since we used it.
		I think it was used for collaboration, to merge two virtual
		worlds into one.


	- What interesting messages can be sent from the client to the server?
		(or to specific devices configured in a server)

	- Is it intended that a vrpn_server remember (and then regurgitate)
		requests for devices that aren't configured?

		I'm seeing this with the FreeVR "vrpntest" tool -- if one
		mistakenly requests a connection for a non-existent device,
		that device will now appear in future lists of "known" devices.

	Russ (12/19/13) also says: To capture streams from a running app, you
	can run test_vrpn (in server_src) and it will emit lots of messages
	(including text messages).  You can connect your client to it and get
	a copy of the stream.  If you run a logging server and connect to it
	with a client like vrpn_print_devices then you'll get a logfile that
	includes the ping message (incoming) and pong messages (outgoing).

**************************************************************************/
#if !defined(FREEVR) && !defined(TEST_APP) && !defined(CAVE)
/* Choose how this code should be compiled (ie. for CAVE, standalone, etc). */
#  define	FREEVR
#  undef	TEST_APP
#  undef	CAVE
#endif

#undef	VRPN_LIB	/* The #undef version does it's own parsing of the data stream, and does not require linking with the VRPN library */
#undef	DEBUG


/* Header includes needed for the both the FreeVR library, and test utility */
#include <stdio.h>		/* needed for printf & all other standard I/O */
#include <string.h>		/* needed for strcmp(), strncmp(), strncpy() & strchr() */
#include <stdint.h>		/* needed for uint16_t & uint32_t types */
#include <stdlib.h>		/* needed for abort() */
#include <errno.h>		/* for errno global and perror() */
#include <fcntl.h>		/* for fcntl() to set socket parameters */
#ifdef __APPLE__
#  include </Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/usr/include/machine/endian.h>/* set __BYTE_ORDER value */
#else
#  include <endian.h>		/* set __BYTE_ORDER value */
#endif
	#include <sys/socket.h>

#include "vr_socket.h"		/* for vrSocketCall() and vrSocketAwaitData() */
#include "vr_debug.h"
#include "vr_utils.h"		/* for vrSleep() and vrEndianSwap<N>() */
#include "vr_math.h"		/* for VR_{X,Y,Z,W,AZIM,ELEV,ROLL} */


#if defined (FREEVR)
/* Header includes needed for the FreeVR library, but not the test utility */
#  include "vr_input.h"
#  include "vr_input.opts.h"
#  include "vr_parse.h"
#  include "vr_shmem.h"
#endif


#if defined(TEST_APP) /* { */
/* Header includes needed just for the "vprntest" utility */
/* NOTE: this includes some FreeVR *source code* */
#  include <stdlib.h>		/* needed for malloc(), getenv() & atoi() */
#  include <signal.h>		/* needed for signal() code */
#  include <sys/time.h>		/* needed for gettimeofday() */
#  include "vr_socket.c"	/* needed for vrSocketCall code & vrSocketAwaitData() */
#  include "vr_utils.c"		/* needed for vrEndianSwap4() & vrEndianSwap8() */
#  define MATH_ONLY		/* exclude the FreeVR material in vr_math.c  */
#  include "vr_math.c"		/* needed for Quaternion to Euler conversion */

#  define vrShmemAlloc0 malloc
#  define vrShmemFree free

/* Declarations of functions copied from vr_system.c for TEST_APP */
vrTime vrCurrentWallTime();
vrTime vrStartTime();

/*****************************************************************************/
/* vrInputDevice is a structure type used in the test application (TEST_APP) */
/*   to mimic the type from vr_input.h, and allowing the test app to make    */
/*   use of the FreeVR input callback functions.                             */
typedef struct vrID_st {
		char		*name;		/* CONFIG: name assigned to device in config file */
		int		operating;	/* whether the device is operating */
		void		*aux_data;	/* auxiliary data, specific to the type of input */
	} vrInputDevice;
#endif /* } TEST_APP */


#if defined(CAVE) /* { */
/* Header includes needed just for interfacing with the CaveLib */
/* NOTE: It's been over a decade since this has been tested     */
#  include <math.h>
#  include "cave.h"
#  include "cave.private.h"
#  include "cave.tracker.h"
static CAVE_SENSOR_ST initial_cave_sensor = {
		0,5,0,
		0,0,0,
		0,	/*timestamp*/
		FALSE,	/*calibrated*/
		CAVE_TRACKER_FRAME
	};
#endif /* } CAVE */


#ifdef VRPN_LIB /* { */
/*** special VRPN includes ***/
#  include "vrpn/vrpn_Tracker.h"
#  include "vrpn/vrpn_Button.h"
#  include "vrpn/vrpn_Analog.h"
#endif /* } VRPN_LIB */


	/***************************************************/
	/*** definitions for interfacing with the device ***/
	/***                                             ***/

#define BUFSIZE		2048
#define HEADER_SIZE	24

#define DEFAULT_HOST	"localhost"
#define DEFAULT_PORT	3883

/* TODO: determine whether we still want these */
/* Default sensitivity values for VRPN */
#define TRANS_SENSITIVITY	0.1
#define ROT_SENSITIVITY		1.0
#define VALUATOR_SENSITIVITY	17.0	/* empirically determined */


/************************************************************************/
/* VRPNServiceFlags can be used by VRPN to specify different tolerances */
/*   for reliability (usually versus latency).                          */
/* NOTE: only the "RELIABLE" flag is presently used.                    */
/* NOTE Also: FreeVR doesn't use this at all.                           */
typedef enum {
		RELIABLE	 = (1<<0),
		FIXED_LATENCY	 = (1<<1),
		LOW_LATENCY	 = (1<<2),
		FIXED_THROUGHPUT = (1<<3),
		HIGH_THROUGHPUT	 = (1<<4)
	} VRPNServiceFlags;


/*************************************************************************/
/* VRPNMessageType provides an enumeration of the different types of     */
/*   messages that can be expected from the server.  Messages provided   */
/*   by specific devices are assigned numbers starting with 0 (zero) as  */
/*   they are chronologically encountered in the vrpn.cfg file.  Special */
/*   assignments are pre-defined for general messages and are given      */
/*   negative values.                                                    */
typedef enum {
		SENDER_DESCRIPTION	= (-1),	/* sending the "sender name" */
		TYPE_DESCRIPTION	= (-2),	/* sending the "type name" */
		UDP_DESCRIPTION		= (-3),	/* sending the host's IP address */
		LOG_DESCRIPTION		= (-4),	/* set a new Log description value */
		DISCONNECT_MESSAGE	= (-5),	/* never sent -- for logging only */
		INVALID_MESSAGE		= (-6)	/* for FreeVR communications error detection */
		/* ... -- the non-negative message-type values are mapped to  */
		/*        ASCII strings immediately after a client-server     */
		/*        connection is established, and then used to indicate*/
		/*        what type of data is arriving (like tracker data or */
		/*        button data.                                        */
	} VRPNMessageType;


/***********************************************************************/
/* VRPNLogFlags uses the lower two bits to establish which information */
/*   should be logged: incoming data, outgoing data, or both.          */
typedef enum {
		LOG_NOTHING		= (0),
		LOG_INCOMING		= (1<<0),
		LOG_OUTGOING		= (1<<1)
	} VRPNLogFlags;


/******************************************************************************/
/* VRPNHeader: data packet header for communication between client and server */
typedef struct {
		uint32_t	message_size;			/* header-length (24) + unpadded message length (4 + string_length) */
		uint32_t	time_seconds;			/* seconds component of time */
		uint32_t	time_useconds;			/* micro-seconds component of time */
		uint32_t	code;				/* numeric encoding describing message source (source code :-) */
								/*   code is "sender" when (type == SENDER_DESCRIPTION)        */
								/*   code is "message-number" when (type == TYPE_DESCRIPTION)  */
								/*   code is the device-number when sending input values       */
		int32_t		message_type;			/* type of message (from SystemMessageType) */
		int32_t		padding;			/* need 4 more bytes to be on an 8-byte boundary (NOTE: VRPN seems to fill this with a message count) */
	} VRPNHeader;

/*****************************************************************************/
/* VRPNdevice: structure containing input data from a particular VRPN device */
#define MAX_BUTTONS	16
#define MAX_VALUATORS	16
#define MAX_6SENSORS	8
typedef struct {
		char	name[2560];				/* name of this device -- according to VRPN server */

		/***********************/
		/* Text Message inputs */
		uint8_t	newly_reported_message;			/* flag indicating whether the VRPN server just sent a text message */
		uint8_t	newly_reported_pong;			/* flag indicating whether the VRPN server just sent a pong message */
		char	text_message[1024];			/* value of a recent text message */
		int32_t	text_level;				/* "level" of the recent text message */
		int32_t	text_severity;				/* "severity" of the recent text message */
		double	message_time;				/* the time of the last reported message */

		/*****************/
		/* Button inputs */
		uint8_t	ever_reported_button_any;		/* flag indicating whether the VRPN server has ever sent any button for this device */
		uint8_t	ever_reported_button[MAX_BUTTONS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_button[MAX_BUTTONS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_button[MAX_BUTTONS];		/* flag indicating whether the given button has been mapped to a FreeVR input */
		uint8_t	button[MAX_BUTTONS];			/* values of VRPN button inputs */
		double	button_time[MAX_BUTTONS];		/* time of the last button event */

		/****************************/
		/* Valuator (analog) inputs */
		uint8_t	ever_reported_analog_any;		/* flag indicating whether the VRPN server has ever sent any analog for this device */
		uint8_t	ever_reported_analog[MAX_VALUATORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_analog[MAX_VALUATORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_analog[MAX_VALUATORS];		/* flag indicating whether the given analog has been mapped to a FreeVR input */
		double	analog[MAX_VALUATORS];			/* values of VRPN valuator inputs */
		double	analog_max[MAX_VALUATORS];		/* maximum value of VRPN valuator inputs reported thus far */
		double	analog_min[MAX_VALUATORS];		/* minimum value of VRPN valuator inputs reported thus far */
		double	analog_time[MAX_VALUATORS];		/* time of the last valuator event */

		/*****************************/
		/* 6-sensor (tracker) inputs */
		uint8_t	ever_reported_posquat_any;		/* flag indicating whether the VRPN server has ever sent any posquat for this device */
		uint8_t	ever_reported_posquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_posquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		uint8_t	mapped_tracker[MAX_6SENSORS];		/* flag indicating whether the given tracker has been mapped to a FreeVR input */
		double	posquat[MAX_6SENSORS][7];		/* values of VRPN 6-sensor inputs */
		double	posquat_time[MAX_6SENSORS];		/* time of the last 6-sensor event */

		uint8_t	ever_reported_velquat_any;		/* flag indicating whether the VRPN server has ever sent any velquat for this device */
		uint8_t	ever_reported_velquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server has ever sent input numbered such */
		uint8_t	newly_reported_velquat[MAX_6SENSORS];	/* flag indicating whether the VRPN server just sent input numbered_such */
		double	velquat[MAX_6SENSORS][7];		/* values of VRPN 6-sensor input velocity values (currently unused) */
		double	velquat_time[MAX_6SENSORS];		/* time of the last 6-sensor velocity event */
	} VRPNdevice;

/**********************************************************************************/
/*** _VrpnPrivateInfo: auxiliary structure of the current data from the device. ***/
typedef struct {
		/* these are for interfacing with the daemon */
		int		vrpn_socket;			/* communication file descriptor */
		char		*vrpn_host;			/* name of daemon's host */
		int		vrpn_port;			/* port on host listening for application connections */
		int		open;				/* flag with VRPN successfully open */

#ifndef VRPN_LIB /* { */
		/* these are for internal data parsing */
		char		buf[BUFSIZE];			/* the incoming data stream */
		int		eobuf_pos;			/* the last read byte in the buffer */
#endif /* } !VRPN_LIB */

#define MAX_DEVICES 16 /* there can be <n-1> VRPN devices (device 0 reserved for the simulator) */

		/* information about the current (raw/unprocessed) values (SENDER_DESCRIPTION mappings) */
		double		vrpn_starttime;			/* time stamp of the first VRPN message (used to determine waits for playback data) */
		int		num_devices;			/* just a counter to know how many have been mentioned by the VRPN server */
		VRPNdevice	device[MAX_DEVICES];		/* array of possible VRPN input devices */

/* TODO: much of the next few lines are currently only used for the VRPN_LIB version of this code -- need to clarify which is still needed and where */
		int		num_button_devices;		/* num of button devices used */
		char		button_names[MAX_DEVICES][256];	/* name of the button devices */
#ifdef VRPN_LIB /* { */
	   vrpn_Button_Remote	*vrpn_button[MAX_DEVICES];	/* VRPN button class */
#endif /* } VRPN_LIB */

		int		num_analog_devices;		/* num of analog devices used */
		char		analog_names[MAX_DEVICES][256];	/* name of the analog devices */
#ifdef VRPN_LIB /* { */
	   vrpn_Analog_Remote	*vrpn_analog[MAX_DEVICES];	/* VRPN analog class */
#endif /* } VRPN_LIB */

		int		num_tracker_devices;		/* num of tracker devices used */
		char		tracker_names[MAX_DEVICES][256];	/* names of the tracker devices */
#ifdef VRPN_LIB /* { */
	   vrpn_Tracker_Remote	*vrpn_tracker[MAX_DEVICES];	/* VRPN tracker class */
#endif /* } VRPN_LIB */

#ifndef VRPN_LIB /* { */
		/* these are the TYPE_DESCRIPTION ASCII to integer mappings */
#  define MAX_CODES 128 /* the upper bounds on number of TYPE_DESCRIPTION codes in the array */
		int		biggest_code;			/* highest code value recorded */
		char		message_codes[MAX_CODES][256];	/* string names for each message code VRPN server is reporting */
		int		MessageSensorPos;		/* VRPN code for sensor location/rotation data */
		int		MessageSensorVel;		/* VRPN code for sensor velocity data */
		int		MessageSensorAcc;		/* VRPN code for sensor acceleration data */
		int		MessageButton;			/* VRPN code for one changed button data */
		int		MessageAllButtons;		/* VRPN code for all button data */
		int		MessageValuator;		/* VRPN code for all valuator data */
		int		MessagePing;			/* VRPN code for receiving a ping request */
		int		MessagePong;			/* VRPN code for receiving a pong response */
		int		MessageText;			/* VRPN code for text message */
		/* TODO: other VRPN classes include Text, Phantom?, ... */

		int		unanswered_ping;		/* If we get a ping, we need to response with a pong */
#endif /* } !VRPN_LIB */


		/* these are for internal data parsing */
		char		version[512];			/* self-reported version of the device */
		char		op_params[256];			/* operating parameters of the device (according to it) */
		int		server_ver_major;		/* Major version code of the VRPN server */
		int		server_ver_minor;		/* Minor version code of the VRPN server */

		/* information about how to filter the data for use with the VR library (FreeVR or CAVE) */
		float		scale_valuator;			/* scaling factor for valuators */
		float		scale_tracker_trans;		/* multiplier to scale from the VRPN location units (meters) */
		int		govern_time;			/* flag of whether VRPN data timestamp should regulate input data flow (used for data playback) */

		/* information about the inputs and mappings */
#ifdef CAVE
		/* CAVE specific fields here */
		/* ... */

#elif defined(FREEVR)
		/* FREEVR specific fields here */

		int		mapped;				/* flag to indicate when the VRPN devices have been mapped to FreeVR devices */

		/* 2switch/button mappings */
		int		num_buttons;			/* it is often wise to store the number of button inputs in addition to knowing the maximum possible */
		vr2switch	**button_inputs;		/* the maximum number of buttons is devinfo->num_2ways + devinfo->num_scontrols (all the self-controls are not necessarily buttons) */
		int		*button_map_button;		/* An array of numbers indicating the particular button */
		char		**button_map_devicename;	/* An array of VRPN device names for each button input */
		VRPNdevice	**button_map_device;		/* An array of VRPN device pointers linking to the specific device structure */

		/* Valuator/analog mappings */
		int		num_valuators;			/* it is often wise to store the number of valuator inputs in addition to knowing the maximum possible */
		vrValuator	**valuator_inputs;		/* the maximum number of valuators is devinfo->num_valuators + devinfo->num_scontrols (all the self-controls are not necessarily valuators) */
		int		*valuator_map_valuator;		/* An array of numbers indicating the particular valuator */
		float		*valuator_map_sign;		/* An array of numbers indicating the sign of each valuator input */
		char		**valuator_map_devicename;	/* An array of VRPN device names for each valuator input */
		VRPNdevice	**valuator_map_device;		/* An array of VRPN device pointers linking to the specific device structure */

		/* 6-sensor/tracker mappings */
		int		num_6sensors;			/* it is often wise to store the number of 6-sensor inputs in addition to knowing the maximum possible */
		vr6sensor	**sensor6_inputs;		/* the maximum number of 6sensors is devinfo->num_6sensors currently no self-control options) */
		int		*tracker_map_tracker;		/* An array of numbers indicating the particular tracker */
		char		**tracker_map_devicename;	/* An array of VRPN device names for each 6sensor input */
		VRPNdevice	**tracker_map_device;		/* An array of VRPN device pointers linking to the specific device structure */
		int		sim6sensor_change;		/* flag that indicates whether any of the controls to a simulated 6-sensor have changed */

#  if 0 /* Question: were these used by the VRPN_LIB version of this file? */
		vr2switch	*button_inputs[MAX_DEVICES][MAX_BUTTONS];	/* each button device can have <n> buttons */
		vrValuator	*valuator_inputs[MAX_DEVICES][MAX_VALUATORS];	/* ditto for analog */
		vr6sensor	*sensor6_inputs[MAX_DEVICES][MAX_6SENSORS];	/* ditto for trackers */
#  endif

#  if 0 /* TODO: not sure if we want/need this */
#    define MAX_CONTROLS  10
		vrControl	*control_inputs[MAX_CONTROLS];
#  endif

		/* for the FreeVR simulated 6-sensors */
		int		active_sim6sensor;		/* The simulated 6-sensor that is being actively controlled */
		vr6sensorConv	sensor6_options;		/* Structure of settings that affect how a 6-sensor is simulated */
		float		sim_values[6];			/* Array of 6 valuator values used to move the simulated 6-sensor */
#endif /* end library-specific fields */

	} _VrpnPrivateInfo;


#ifdef VRPN_LIB /* { */
/***********************************************************/
/*** structure for the userdata of the callback handles. ***/
typedef struct {
		int		  num;		/* the number in the device list of this type */
		_VrpnPrivateInfo *aux;		/* a pointer to the overall private info structure */
	} _VrpnUserdata;
#endif /* } VRPN_LIB */



	/**************************************************/
	/*** General NON public VRPN interface routines ***/
	/**************************************************/

/*************************************************************************/
/* _VrpnInitializeStruct(): sets initial values in the device specific   */
/*   internal structure to prepare for making the initial connection.    */
/*   For some devices, memory will be allocated for arrays of values.    */
/*                                                                       */
/* NOTE: Some (few) versions of this routine take a second "typename"    */
/*   argument that can be used to specify a particular device among many */
/*   that share (more or less) the same protocol.  The typename is then  */
/*   used to determine what specific features are available with this    */
/*   particular type of device.                                          */
static void _VrpnInitializeStruct(_VrpnPrivateInfo *aux, char *typename)
{
	int		count;			/* loop counter */

	bzero(aux, sizeof(*aux));		/* just in case, zero out everything first */

	aux->version[0] = '\0';
	aux->op_params[0] = '\0';

	aux->biggest_code = -1;
	aux->MessageSensorPos = -1;
	aux->MessageSensorVel = -1;
	aux->MessageSensorAcc = -1;
	aux->MessageButton = -1;
	aux->MessageAllButtons = -1;
	aux->MessageValuator = -1;
	aux->MessageText = -1;
	aux->MessagePing = -1;
	aux->MessagePong = -1;

	aux->num_button_devices = 0;
	aux->num_analog_devices = 0;
	aux->num_tracker_devices = 0;

	aux->vrpn_starttime = -1.0;			/* "-1.0" indicates that the first packet time hasn't yet been set */
	aux->num_devices = 1;				/* device 0 is reserved for a simulator device! */
	strcpy(aux->device[0].name, "simulator");	/* The simulator device has the empty name */

	aux->scale_valuator = 1.0;			/* set the default valuator scaling factor */
	aux->scale_tracker_trans = 1.0;		/* Don't scale by default */
	aux->govern_time = 0;				/* Regulate VRPN input based on data timestamp */

#ifndef TEST_APP /* {  (we have no simulated inputs for the test app) */
	/* set values for simulated 6-sensor */
	/*   start with no (zero) movement values in the valuators holders */
	aux->sim_values[0] = 0.0;
	aux->sim_values[1] = 0.0;
	aux->sim_values[2] = 0.0;
	aux->sim_values[3] = 0.0;
	aux->sim_values[4] = 0.0;
	aux->sim_values[5] = 0.0;

	aux->active_sim6sensor = 1;			/* start with the "wand" active  -- potential bug */
	aux->active_sim6sensor = 0;			/* TODO: go back to wand as default -- currently head (ie. delete this line) */

	aux->sensor6_options.azimuth_axis = VR_Y;	/* azimuth is about the Y axis */
	aux->sensor6_options.relative_axis = 1;		/* default to relative rotations */
	aux->sensor6_options.return_to_zero = 0;	/* default to free floating */
	aux->sensor6_options.ignore_all = 0;		/* default to joystick off */
	aux->sensor6_options.ignore_trans = 0;		/* default to joystick off */
	aux->sensor6_options.restrict_space = 1;	/* default restricted to working-volume */
	aux->sensor6_options.swap_yz = 0;		/* default to standard Y & Z axes */
	aux->sensor6_options.swap_transrot = 0;		/* default to unswapped translation & rotation */
	aux->sensor6_options.trans_scale = TRANS_SENSITIVITY;	/* set the default translational scaling */
	aux->sensor6_options.rot_scale = ROT_SENSITIVITY;	/* set the default rotational scaling */

	/* The default working-volume is that of the typical CAVE */
	aux->sensor6_options.working_volume_min[VR_X] = -5.0;
	aux->sensor6_options.working_volume_max[VR_X] =  5.0;
	aux->sensor6_options.working_volume_min[VR_Y] =  0.0;
	aux->sensor6_options.working_volume_max[VR_Y] = 10.0;
	aux->sensor6_options.working_volume_min[VR_Z] = -5.0;
	aux->sensor6_options.working_volume_max[VR_Z] =  5.0;
#endif /* } */

	/* everything else is zero'd by default */
}


#ifdef FREEVR /* { */
/*************************************************************************/
/* _VrpnPrint6sensorOptions(): outputs the contents of the settings of   */
/*   the simulated 6-sensors controlled by this device.                  */
/*   Output verbosity can be controlled by the "style" argument.          */
/* NOTE: simulated 6-sensors are not part of the test application.       */
static void _VrpnPrint6sensorOptions(FILE *file, _VrpnPrivateInfo *aux, vrPrintStyle style)
{
	vrFprintf(file, BOLD_TEXT "VRPN - 6-sensor settings:" NORM_TEXT "\n");
	vrFprintf(file, "\tActive 6-sensor = %d\n", aux->active_sim6sensor);
	vrFprintf(file, "\r"
		"\t\tsim_values = (%.2f, %.2f, %.2f,  %.2f, %.2f, %.2f)\n",
		aux->sim_values[VR_X],
		aux->sim_values[VR_Y],
		aux->sim_values[VR_Z],
		aux->sim_values[VR_AZIM+3],
		aux->sim_values[VR_ELEV+3],
		aux->sim_values[VR_ROLL+3]);
	vrFprintf(file, "\r"
		"\t\tsim6sensor_change = %d\n",
		aux->sim6sensor_change);

	vrFprint6sensorOptions(file, &(aux->sensor6_options), style);
}
#endif /* } FREEVR */


/*************************************************************************/
/* _VrpnPrintPrivateInfo(): outputs the contents of the device specific  */
/*   internal structure to help with configuration and system debugging. */
/*   Output verbosity can be controlled by the "style" argument.         */
static void _VrpnPrintPrivateInfo(FILE *file, _VrpnPrivateInfo *aux, vrPrintStyle style)
{
	int		devnum;		/* for looping over devices */
	int		count;		/* loop counter */
	int		count2;		/* inner loop counter */
#ifdef FREEVR /* { */
	vrGenericInput	*input;
#endif /* } FREEVR */

	vrFprintf(file, "VRPN input-device internal structure (%p):\n", aux);
	vrFprintf(file, "\r\tversion -- '%s' (%d.%d)\n", aux->version, aux->server_ver_major, aux->server_ver_minor);
	vrFprintf(file, "\r\tserver connection -- TCP@%s:%d\n", aux->vrpn_host, aux->vrpn_port);
	vrFprintf(file, "\r\toperating parameters -- '%s'\n", aux->op_params);
	vrFprintf(file, "\r\topen = %d\n", aux->open);
#ifdef FREEVR /* { */
	vrFprintf(file, "\r\tmapped = %d\n", aux->mapped);
#endif /* } FREEVR */

	vrFprintf(file, "\r\tscale_valuator = %f\n", aux->scale_valuator);
	vrFprintf(file, "\r\tscale_trans = %f\n", aux->scale_tracker_trans);
	vrFprintf(file, "\r\tgovern_time = %d\n", aux->govern_time);
	vrFprintf(file, "\r\tvrpn_starttime = %f\n", aux->vrpn_starttime);
	vrFprintf(file, "\r\tnum_devices = %d\n", aux->num_devices);

	/* NOTE: start devnum at 1 since device[0] is the simulator and is not from the VRPN server */
	vrFprintf(file, "\tVRPN Server reported devices:\n");		/* SENDER_DESCRIPTION values */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		if (aux->device[devnum].name[0] != '\0')
			vrFprintf(file, "\t\tDevice %d is '%s'\n", devnum, aux->device[devnum].name);
	}
	vrFprintf(file, "\tVRPN Server reported message types:\n");	/* TYPE_DESCRIPTION values */
	for (count = 0; count <= aux->biggest_code; count++) {
		vrFprintf(file, "\t\tMessage Code %2d is '%s'\n", count, aux->message_codes[count]);
	}

	vrFprintf(file, "\tWatch Messages:\n");				/* TYPE_DESCRIPTION values */
	vrFprintf(file, "\t\t%2d: Tracker position\n", aux->MessageSensorPos);
	vrFprintf(file, "\t\t%2d: Tracker velocity\n", aux->MessageSensorVel);
	vrFprintf(file, "\t\t%2d: Tracker acceleration\n", aux->MessageSensorAcc);
	vrFprintf(file, "\t\t%2d: Button change event\n", aux->MessageButton);
	vrFprintf(file, "\t\t%2d: All Button event\n", aux->MessageAllButtons);
	vrFprintf(file, "\t\t%2d: Valuator event\n", aux->MessageValuator);
	vrFprintf(file, "\t\t%2d: Ping message\n", aux->MessagePing);
	vrFprintf(file, "\t\t%2d: Pong message\n", aux->MessagePong);
	vrFprintf(file, "\t\t%2d: Text message\n", aux->MessageText);

#ifdef FREEVR /* { */
	/************************************/
	/** print the FreeVR input objects **/

	/* button inputs */
	vrFprintf(file, "\n\tbutton inputs:\n");
	for (count = 0; count < aux->num_buttons; count++) {
		vrFprintf(file, "\t\t2switch_input[%d] = %p (%s:%s) -- mapped to VRPN device '%s':button[%d] %s\n",
			count,
			aux->button_inputs[count],
			vrInputTypeName(aux->button_inputs[count]->input_type),
			aux->button_inputs[count]->my_object->name,
			aux->button_map_devicename[count],
			aux->button_map_button[count],
			(aux->button_map_device[count] == NULL ? RED_TEXT "(nodev)" NORM_TEXT : aux->button_map_device[count]->ever_reported_button[aux->button_map_button[count]] ? "(used)" : ""));
	}

	/* valuator inputs */
	vrFprintf(file, "\n\tvaluator inputs:\n");
	for (count = 0; count < aux->num_valuators; count++) {
		vrFprintf(file, "\t\tvaluator_input[%d] = %p (%s:%s) -- mapped to VRPN device '%s':analog[%d] %s\n",
			count,
			aux->valuator_inputs[count],
			vrInputTypeName(aux->valuator_inputs[count]->input_type),
			aux->valuator_inputs[count]->my_object->name,
			aux->valuator_map_devicename[count],
			aux->valuator_map_valuator[count],
			(aux->valuator_map_device[count] == NULL ? RED_TEXT "(nodev)" NORM_TEXT : aux->valuator_map_device[count]->ever_reported_analog[aux->valuator_map_valuator[count]] ? "(used)" : ""));
	}

	/* 6sensor inputs */
	vrFprintf(file, "\n\t6sensor inputs:\n");
	for (count = 0; count < aux->num_6sensors; count++) {
		vrFprintf(file, "\t\t6sensor_input[%d] = %p (%s:%s) -- mapped to VRPN device '%s':tracker[%d] %s\n",
			count,
			aux->sensor6_inputs[count],
			vrInputTypeName(aux->sensor6_inputs[count]->input_type),
			aux->sensor6_inputs[count]->my_object->name,
			aux->tracker_map_devicename[count],
			aux->tracker_map_tracker[count],
			(aux->tracker_map_device[count] == NULL ? RED_TEXT "(nodev)" NORM_TEXT : aux->tracker_map_device[count]->ever_reported_analog[aux->tracker_map_tracker[count]] ? "(used)" : ""));
	}

	/* simulated 6-sensor settings */
	vrFprintf(file, "\r");
	_VrpnPrint6sensorOptions(file, aux, brief);
#endif /* } FREEVR */

#ifdef FREEVRx /* { */
	/* button inputs */
	vrFprintf(file, "\r\tbutton_devices = %d:\n", aux->num_button_devices);
	for (count = 0; count < aux->num_button_devices; count++) {
#  ifdef VRPN_LIB /* { */
		vrFprintf(file, "\r\t\tbutton[%d] = '%s' (%p)\n",
			count,
			aux->button_names[count],
			aux->vrpn_button[count]);
#  endif /* } */
		/* TODO: see if the VRPN_LIB version can use the above (improved) code instead of this */
		for (count2 = 0; count2 < MAX_BUTTONS; count2++) {
			input = (vrGenericInput *)aux->button_inputs[count][count2];
			if (input != NULL) {
				vrFprintf(file, "\r\t\t\tbutton %d: input = %p (%s:%s)\n",
					count2, input, vrInputTypeName(input->input_type), input->my_object->name);
			}
		}
	}

	/* analog inputs */
	vrFprintf(file, "\r\tanalog_devices = %d:\n", aux->num_analog_devices);
	for (count = 0; count < aux->num_analog_devices; count++) {
#  ifdef VRPN_LIB /* { */
		vrFprintf(file, "\r\t\tanalog[%d] = '%s' (%p)\n",
			count,
			aux->analog_names[count],
			aux->vrpn_analog[count]);
#  endif /* } */
		for (count2 = 0; count2 < MAX_VALUATORS; count2++) {
			input = (vrGenericInput *)aux->valuator_inputs[count][count2];
			if (input != NULL) {
				vrFprintf(file, "\r\t\t\tanalog %d: input = %p (%s:%s)\n",
					count2, input, vrInputTypeName(input->input_type), input->my_object->name);
			}
		}
	}

	/* tracker inputs */
	vrFprintf(file, "\r\ttracker_devices = %d:\n", aux->num_tracker_devices);
	for (count = 0; count < aux->num_tracker_devices; count++) {
#  ifdef VRPN_LIB /* { */
		vrFprintf(file, "\r\t\ttracker[%d] = '%s' (%p)\n",
			count,
			aux->tracker_names[count],
			aux->vrpn_tracker[count]);
#  endif /* } */
		for (count2 = 0; count2 < MAX_6SENSORS; count2++) {
			input = (vrGenericInput *)aux->sensor6_inputs[count][count2];
			if (input != NULL) {
				vrFprintf(file, "\r\t\t\ttracker %d: input = %p (%s:%s)\n",
					count2, input, vrInputTypeName(input->input_type), input->my_object->name);
			}
		}
	}
#endif /* } FREEVRx */
}


/****************************************************************************/
/* _VrpnPrintHelp(): outputs the mappings from the physical device to the   */
/*    virtual inputs and controls to help the user understand the interface.*/
static void _VrpnPrintHelp(FILE *file, _VrpnPrivateInfo *aux)
{
	int		count;		/* loop counter */

#if 0 || defined(TEST_APP)
	vrFprintf(file, BOLD_TEXT "Sorry, VRPN - print_help control not yet implemented." NORM_TEXT "\n");
#else
	vrFprintf(file, BOLD_TEXT "VRPN - inputs:" NORM_TEXT "\n");
	for (count = 0; count < aux->num_buttons; count++) {
		if (aux->button_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->button_inputs[count]->my_object->desc_str,
				(aux->button_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->button_inputs[count]->my_object->name);
		}
	}
	for (count = 0; count < aux->num_valuators; count++) {
		if (aux->valuator_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->valuator_inputs[count]->my_object->desc_str,
				(aux->valuator_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->valuator_inputs[count]->my_object->name);
		}
	}
#  if 0 /* TODO: why aren't we doing 6sensors (trackers) for the help output? */
	for (count = 0; count < MAX_6SENSORS; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			vrFprintf(file, "\t%s -- %s%s\n",
				aux->sensor6_inputs[count]->my_object->desc_str,
				(aux->sensor6_inputs[count]->input_type == VRINPUT_CONTROL ? "control:" : ""),
				aux->sensor6_inputs[count]->my_object->name);
		}
	}
#  endif
#endif
}


/************************************************************************/
/* The _VrpnReadInput() function acquires incoming data from the server */
/*   source, parsing the stream of data.  The data is then placed into  */
/*   the generic portion of the "_VrpnPrivateInfo" structure.  Data is  */
/*   then transitioned into the FreeVR (or whatever) system through the */
/*   "_VrpnGetData()" function.                                         */
/* NOTE: [11/26/21] _VrpnReadInput() will parse many messages, but the  */
/*   way it is currently written, some of that data can be missed if it */
/*   comes fast enough that while the entire buffer is emptied, we      */
/*   never get out to _VrpnGetData() where the data can be used.        */
/* Returns:                                                             */
/*    -1: socket connection was closed                                  */
/*   >=0: number of packets read and parsed                             */
static int _VrpnReadInput(_VrpnPrivateInfo *aux)
{
#ifdef VRPN_LIB /* { */
	/************************************************************/
	/* This section reads the VRPN data using the VRPN library. */
	/************************************************************/
	char		return_code = 1;	/* currently not watching for any problems */
	int		count;			/* loop counter */

	for (count = 0; count < aux->num_button_devices; count++) {
		if (aux->vrpn_button[count] != NULL)
			aux->vrpn_button[count]->mainloop();
		else	vrErrPrintf(RED_TEXT "_VrpnReadInput: aux->vrpn_button[%d] shouldn't be NULL\n" NORM_TEXT, count);
	}

	for (count = 0; count < aux->num_analog_devices; count++) {
		if (aux->vrpn_analog[count] != NULL)
			aux->vrpn_analog[count]->mainloop();
		else	vrErrPrintf(RED_TEXT "_VrpnReadInput: aux->vrpn_analog[%d] shouldn't be NULL\n" NORM_TEXT, count);
	}

	for (count = 0; count < aux->num_tracker_devices; count++) {
		if (aux->vrpn_tracker[count] != NULL)
			aux->vrpn_tracker[count]->mainloop();
		else	vrErrPrintf(RED_TEXT "_VrpnReadInput: aux->vrpn_tracker[%d] shouldn't be NULL\n" NORM_TEXT, count);
	}
	return (return_code);
#else /* } <-- VRPN_LIB vs. !VRPN_LIB --> { */
	/************************************************************/
	/* This section reads the VRPN data by parsing the raw data */
	/************************************************************/
	VRPNHeader	VRPNdata;		/* storage of the incoming VRPN header */
	int		read_result;		/* return value of the call to "read()" */
	int		max_read;		/* the maximum number of bytes available in buffer */
	int		packets_decoded = 0;	/* return the number of packets that are decoded */
	int		packet_size;		/* size of the currently being parsed data packet -- adjusted as new information is determined */

	uint32_t	message_length;		/* for string-type messages */
	char		*message;		/* for string-type messages */
	int		padded_message_size;	/* the complete length of the message portion of the packet (or does it include header too?) */

	double		vrpntime;		/* time stamp of the incoming VRPN data */
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */

	int		real_data = 0;		/* have we parsed any real data yet?  (only one real-data per call) */

	/*******************************************************/
	/*** read available data into available buffer space ***/
	//do { -- changed to do the test at the top 11/26/21
	read_result = 1;	/* set so we do a read if there is any room in the buffer */
	while (read_result > 0 && aux->eobuf_pos < BUFSIZE) {
		max_read = BUFSIZE - aux->eobuf_pos;
		read_result = (ssize_t)read/*recv*/(aux->vrpn_socket, &aux->buf[aux->eobuf_pos], max_read/*BUFSIZE - aux->eobuf_pos*//*, 0*/);
//if (aux->eobuf_pos > 0) { for (int yo = 0; yo < aux->eobuf_pos; yo++) { vrPrintf("%s0x%02x(%c) ", (yo%24 ? NORM_TEXT : RED_TEXT), (unsigned char)(aux->buf[yo]), (unsigned char)(aux->buf[yo])); } vrPrintf("\n\n"); }

		/* Check whether we maxed out the buffer */
		if (read_result == max_read) {
			vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, RED_TEXT ">>>>>>> Read to the brim (%d) (read_result = %d) <<<<<<<<\n" NORM_TEXT, max_read, read_result);
		}

		/* NOTE: if read_result == 0, the socket was closed from the other side.  This condition is handled at the end of the function */
		if (read_result > 0) {
			aux->eobuf_pos += read_result;
			/* debug statement here */
		}

#  if 0
		if (aux->eobuf_pos == BUFSIZE){
			vrPrintf(RED_TEXT ">>>>>>> !!BUFFER IS FULL!! (%d) <<<<<<<<\n" NORM_TEXT, aux->eobuf_pos);
		}
#endif
#  if 1
		if (read_result < -1) {
			/* we don't care about a -1 return value */
			/* error ? */
			vrErrPrintf("read_result = %d, errno = %d\n", read_result, errno);
			perror("socket read:");
		}
#  endif
#  if 0
		if (read_result == -1) {
			/* Debugging from a -1 return value */
			/* error ? */
			vrErrPrintf("*read_result* = %d, errno = %d\n", read_result, errno);
			perror("socket read with -1:");
		}
#  endif
//vrPrintf("!!!!!!! Buffer has data (%d bytes available)! -- read_result = %d\n", BUFSIZE - aux->eobuf_pos, read_result);
	}
	//} while (read_result > 0 && aux->eobuf_pos < BUFSIZE);

#if 0 /* 11/26/21: This was a (failed) experiment to try to determine whether there's still a connection from the other side, but it never returns other than 0 for netcat'd (nc'd) data, and fails too soon */
	/******************************************************/
	/* Check whether the socket connection is still alive */
send(aux->vrpn_socket, "$", 1, 0);	/* added this because there's information that the only way to know whether a socket is connected is to try to send to it. */
struct sockaddr_in junk;
socklen_t length = sizeof(junk);
memset(&junk, 0, sizeof(junk));
int peer_result;
	peer_result = getpeername(aux->vrpn_socket, (struct sockaddr *)&junk, &length);
vrPrintf("::: peer_result = %d, addr = %x; family = %d, port = %d\n", peer_result, junk.sin_addr.s_addr, junk.sin_family, junk.sin_port);
#endif

	/*******************************************/
	/*** decode data contained in the buffer ***/

	/* decode all setup messages and the first data message contained in the buffer */
	/*   (otherwise _VRPNGetData() may skip data when parsing only the last message)*/
//vrPrintf("HEADER_SIZE = %d, sizeof(VRPNHeader) = %d, aux->eobuf_pos = %d\n", HEADER_SIZE, sizeof(VRPNHeader), aux->eobuf_pos);
	packet_size = HEADER_SIZE + 8;	/* There at least must be data for the first header */  /* NOTE: in reality, we'll always need the first number as well */
	while (aux->eobuf_pos >= packet_size && !real_data) {
		/* point the VRPN interpretation structure (aka "header") to the beginning of the buffer */
		VRPNdata = *(VRPNHeader *)aux->buf;

		/* interpret the header */
#  if __BYTE_ORDER == __LITTLE_ENDIAN
		vrEndianSwap4(&VRPNdata.message_size);
		vrEndianSwap4(&VRPNdata.time_seconds);
		vrEndianSwap4(&VRPNdata.time_useconds);
		vrEndianSwap4(&VRPNdata.code);
		vrEndianSwap4(&VRPNdata.message_type);
		vrEndianSwap4(&VRPNdata.padding);		/* For debugging -- VRPN might be sneaking a counter in here */
#  endif
/* Do some sanity tests here */
if (VRPNdata.message_size == 32024) {	/* Fix an incorrect byte */
	VRPNdata.message_size = 24;
	vrDbgPrintf("Correcting message_size to = %d, bytes in the buffer = %d, seconds = %u, code = %d\n", VRPNdata.message_size, aux->eobuf_pos, VRPNdata.time_seconds, VRPNdata.code);
}

#if 1 /* This code is only for a specific time window!  (which could be from stored data) */ /* delete me */
/* Special test for data from a specific stored run! */
if (VRPNdata.message_size == 1987211374 && VRPNdata.time_seconds == 975205989) {	/* redo the version ("cookie") comparison check */
	/* parse the version number (from the "cookie") */
vrPrintf("buf[0-4] = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n", (unsigned char)aux->buf[0], (unsigned char)aux->buf[1], (unsigned char)aux->buf[2], (unsigned char)aux->buf[3], (unsigned char)aux->buf[4]);
	sscanf(aux->buf, "vrpn: ver. %d.%d  0", &aux->server_ver_major, &aux->server_ver_minor);
	strncpy(aux->version, aux->buf, strlen(aux->buf));
	vrPrintf("VRPN-recheck: got (%d) '%s' -- major = %d, minor = %d\n", read_result, aux->buf, aux->server_ver_major, aux->server_ver_minor);
vrPrintf("VRPN: got (%d) '%s' -- major = %d, minor = %d\n", read_result, aux->buf, aux->server_ver_major, aux->server_ver_minor);
	memmove(aux->buf, &aux->buf[24], (sizeof(aux->buf)-24));	/* shift the data */
	aux->eobuf_pos -= 24;
	continue;	/* jump out of the loop if we don't have sane data */
} else

/* Special test for data from a specific stored run! */
if (VRPNdata.message_size > BUFSIZE || (VRPNdata.time_seconds < 1640000000 && VRPNdata.time_seconds != 0) || VRPNdata.code > 10000 || VRPNdata.message_type > 10000) {
	/* probably something went wrong, and we can try to correct it. */
	vrDbgPrintf("message_size = %d, bytes in the buffer = %d, seconds = %u, code = %d\n", VRPNdata.message_size, aux->eobuf_pos, VRPNdata.time_seconds, VRPNdata.code);
	vrPrintf(RED_TEXT "Trying to correct sequence misalignment -- skipping 4 bytes: 0x%02x 0x%02x 0x%02x 0x%02x (then: 0x%02x 0x%02x 0x%02x 0x%02x).\n" NORM_TEXT, (unsigned char)aux->buf[0], (unsigned char)aux->buf[1], (unsigned char)aux->buf[2], (unsigned char)aux->buf[3], (unsigned char)aux->buf[4], (unsigned char)aux->buf[5], (unsigned char)aux->buf[6], (unsigned char)aux->buf[7]);
	memmove(aux->buf, &aux->buf[4], (aux->eobuf_pos - 4));	/* shift the data */
	aux->eobuf_pos -= 4;
	continue;	/* jump out of the loop if we don't have sane data */
}
#endif

		/* Set the VRPN device for this "code" */
		vrpn_dev = &(aux->device[VRPNdata.code+1]);

		/* Keep track of the VRPN input time */
		vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
		if (aux->vrpn_starttime < 0.0) {
			aux->vrpn_starttime = vrpntime;
			vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "\nJust set VRPN Start time at %lf\n", vrCurrentWallTime());
		}
//vrPrintf("\nvrpntime is %lf; delta = %lf\n", vrpntime, vrpntime - aux->vrpn_starttime);
//vrPrintf("aux->eobuf_pos = %d\n", aux->eobuf_pos);
if (aux->eobuf_pos == BUFSIZE){
  vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, RED_TEXT ">>>>>>> BUFFER IS FULL (%d) <<<<<<<<\n" NORM_TEXT, aux->eobuf_pos);
}

		/* now make sure the entire packet is in the buffer */
		padded_message_size = VRPNdata.message_size;
		padded_message_size = (padded_message_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("padded_message_size = %d\n", padded_message_size); /* delete me */
//vrPrintf("YO: Header.message_size = %d\n", VRPNdata.message_size);
//vrPrintf("YO: Header.padding = %d\n", VRPNdata.padding);		/* For debugging -- VRPN might be sneaking a counter in here */
		if (aux->eobuf_pos < padded_message_size) {
//vrPrintf("Header.message_size = %d\n", VRPNdata.message_size);
//vrPrintf("Header.time_seconds = %d\n", VRPNdata.time_seconds);
//vrPrintf("Header.time_useconds = %d\n", VRPNdata.time_useconds);
//vrPrintf("Header.code = %d\n", VRPNdata.code);
//vrPrintf("Header.message_type = %d\n", VRPNdata.message_type);
//vrPrintf("Header.padding = %d\n", VRPNdata.padding);		/* For debugging -- VRPN might be sneaking a counter in here */
			/* debug statement here */
			vrDbgPrintfN(VRPN_DBGLVL, "Not enough data in the buffer (%d of %d needed bytes), come back for more\n", aux->eobuf_pos, padded_message_size);
			vrDbgPrintfN(VRPN_DBGLVL, "**** Thus dropping out of the while loop here! ****\n");
			vrDbgPrintf("message_size = %d, padded size = %d, bytes in the buffer = %d, code = %d\n", VRPNdata.message_size, padded_message_size, aux->eobuf_pos, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
//if (padded_message_size == 1987211376) packet_size = 24; /* abort(); [04/06/22: skip the version "cookie" string] */
//if (padded_message_size == 1987211374) packet_size = 24; /* abort(); [04/06/22: skip the version "cookie" string] */
//if (padded_message_size == 32024) packet_size = 24; /* abort(); [04/06/22: skip the version "cookie" string] */
if (padded_message_size > BUFSIZE) {
	/* probably something went wrong, and we can try to correct it. */
	vrPrintf(RED_TEXT "Trying to correct sequence misalignment -- skipping 8 bytes.\n" NORM_TEXT);
//	packet_size = 8;
	memmove(aux->buf, &aux->buf[8], (aux->eobuf_pos - 8));	/* shift the data */
	aux->eobuf_pos -= 8;
}
			continue;	/* jump out of the loop if we don't have enough data */
		}

		/* If the number is more negative than any of our messages, */
		/*   force it to be exactly the INVALID_MESSAGE value, and  */
		/*   then process it in the switch.                         */
		if (VRPNdata.message_type <= INVALID_MESSAGE) {
			VRPNdata.message_type = INVALID_MESSAGE;
		}

//vrPrintf("about to parse message of type %d\n", VRPNdata.message_type); /* delete me */
		/* now interpret the data in the message container */
		switch (VRPNdata.message_type) {
		case SENDER_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, "SENDER_DESCRIPTION\n");

			/* extract the string message -- starting with the size */
			packet_size = HEADER_SIZE + sizeof(message_length);		/* NOTE: this isn't really necessary */
			memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(&message_length);
#  endif
//vrPrintf("message_length = %d\n", message_length);	/* delete me */

			/* now calculate the overall packet size */
			packet_size = HEADER_SIZE + sizeof(message_length) + message_length;
			packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("packet_size = %d, aux->eobuf_pos = %d\n", packet_size, aux->eobuf_pos);	/* delete me */
			if (aux->eobuf_pos < packet_size) {
				/* debug statement here */
				/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
				vrErrPrintf("\nVRPN: Parsing SENDER_DESCRIPTION: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
				vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
				continue;	/* jump out of the loop if we don't have enough data */
			}

			/* store the name of the device */
			message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
			/* NOTE: subtract 1 from MAX_DEVICES because device[0] is the simulator */
			if (VRPNdata.code < MAX_DEVICES-1) {
				/* NOTE: 1 is added to "code" because of the simulator as device[0] */
				strncpy(aux->device[VRPNdata.code+1].name, message, sizeof(aux->device[0].name));
				vrDbgPrintfN(VRPN_DBGLVL, "VRPN device number (%d) has the name '%s'.\n", (int)(VRPNdata.code+1), message);
				aux->num_devices++;
				/* TODO: do a check here to see whether we've already initialized the FreeVR inputs and if so print a warning */
			} else {
#  if !defined(TEST_APP) /* { */
				vrDbgPrintfN(AALWAYS_DBGLVL, "WARNING: VRPN device number (%d) exceeds maximum (%d) -- '%s'.\n", (int)(VRPNdata.code+1), MAX_DEVICES-1, message);
#  else
				vrDbgPrintf("WARNING: VRPN device number (%d) exceeds maximum (%d) -- '%s'.\n", (int)(VRPNdata.code+1), MAX_DEVICES-1, message);
#  endif /* } TEST_APP */
			}
			vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "VRPN: Sender: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, (padding = %d), mes-len = %u, message = '%s'\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, VRPNdata.padding, message_length, message);		/* For debugging: VRPN may be sneaking a counter in here */
//vrPrintf("finished decoding SENDER_DESCRIPTION\n");	/* delete me */
			break;

		case TYPE_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, "TYPE_DESCRIPTION\n");

			/* extract the string message -- starting with the size */
			packet_size = HEADER_SIZE + sizeof(message_length);		/* NOTE: this isn't really necessary */
			memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
			//message = &aux->buf[aux->eobuf_pos];
			message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
#  if __BYTE_ORDER == __LITTLE_ENDIAN
			vrEndianSwap4(&message_length);
#  endif
//vrPrintf("message_length = %d\n", message_length);	/* delete me */

			/* now calculate the overall packet size */
			packet_size = HEADER_SIZE + sizeof(message_length) + message_length;
			packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
			if (aux->eobuf_pos < packet_size) {
				/* debug statement here */
				/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
				vrErrPrintf("\nVRPN: Parsing TYPE_DESCRIPTION: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
				vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
				continue;	/* jump out of the loop if we don't have enough data */
			}

//vrPrintf("got TYPE_DESCRIPTION string: '%s' -- code is %d\n", message, VRPNdata.code);	/* delete me */
			/* now store the TYPE_DESCRIPTION message strings    */
			/*   (and store a mapping of the important messages) */
			if (!strncmp(message, "vrpn_Tracker Pos_Quat", message_length)) {
				aux->MessageSensorPos = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL, "assigned MessageSensorPos to %d\n", aux->MessageSensorPos);
			} else if (!strncmp(message, "vrpn_Tracker Velocity", message_length)) {
				aux->MessageSensorVel = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL, "assigned MessageSensorVel to %d\n", aux->MessageSensorVel);
			} else if (!strncmp(message, "vrpn_Tracker Acceleration", message_length)) {
				aux->MessageSensorAcc = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL, "assigned MessageSensorAcc to %d\n", aux->MessageSensorAcc);
			} else if (!strncmp(message, "vrpn_Button Change", message_length)) {
				aux->MessageButton = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL, "assigned MessageButton to %d\n", aux->MessageButton);
			} else if (!strncmp(message, "vrpn_Button States", message_length)) {
				aux->MessageAllButtons = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "assigned MessageAllButtons to %d\n", aux->MessageAllButtons);
			} else if (!strncmp(message, "vrpn_Analog Channel", message_length)) {
				aux->MessageValuator = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL, "assigned MessageValuator to %d\n", aux->MessageValuator);
			} else if (!strncmp(message, "vrpn_Base text_message", message_length)) {
				aux->MessageText = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "assigned MessageText to %d\n", aux->MessageText);
			} else if (!strncmp(message, "vrpn_Base ping_message", message_length)) {
				aux->MessagePing = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "assigned MessagePing to %d\n", aux->MessagePing);
			} else if (!strncmp(message, "vrpn_Base pong_message", message_length)) {
				aux->MessagePong = VRPNdata.code;
				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, "assigned MessagePong to %d\n", aux->MessagePong);
			}
//vrPrintf("current biggest_code = %d, VRPNdata.code = %d\n", aux->biggest_code, VRPNdata.code);	/* delete me */
			if ((int32_t)(VRPNdata.code) < MAX_CODES) {
				if ((int32_t)(VRPNdata.code) > aux->biggest_code)
					aux->biggest_code = (int32_t)(VRPNdata.code);
				else	vrDbgPrintf("Warning: duplicate message code value received: %u\n", VRPNdata.code);
//vrPrintf("about to store description: code = %d, message = '%s', size = %d\n", VRPNdata.code, message, sizeof(aux->message_codes[0]));	/* delete me */
				strncpy(aux->message_codes[VRPNdata.code], message, sizeof(aux->message_codes[0]));
			}
			vrDbgPrintfN(VRPN_DBGLVL, "Type: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, mes-len = %u, message = '%s'\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, message_length, message);
//vrPrintf("finished decoding TYPE_DESCRIPTION\n");	/* delete me */
			break;

		case UDP_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "UDP_DESCRIPTION -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case LOG_DESCRIPTION:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "LOG_DESCRIPTION -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case DISCONNECT_MESSAGE:
			vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "DISCONNECT_MESSAGE -- not parsed! (code = %u)\n" NORM_TEXT, VRPNdata.code);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		case INVALID_MESSAGE:
			vrDbgPrintfN(VRPN_DBGLVL, "\nVRPN: Invalid messages type: %d -- not parsed!.\n", VRPNdata.message_type);
			packet_size = padded_message_size;	/* skip the complete packet */
			break;

		default:
			/************************************************************************/
			/* The default message is that we have some actual input data to parse. */
			/************************************************************************/
			real_data = 1;

			if (aux->govern_time) {
#if 1 /* 1=regulate/govern inputs based on time -- now a run-time flag */
				/* Don't parse data if its time has not yet come */
				if ((vrpntime - aux->vrpn_starttime) > (vrCurrentWallTime() - vrStartTime())) {	/* another option is vrCurrentSimTime() */
					/* TODO/NOTE: we could put a (very) small delay here to share the processor */
//vrPrintf("delaying input of type %d -- need %lf more time\n", VRPNdata.message_type, (vrpntime - aux->vrpn_starttime) - (vrCurrentWallTime() - vrStartTime()));
					/* By continuing here, we jump to the top of the while, and then exit */
					/*   (because the "real_data" flag is set).                           */
					/* NOTE: If we use "break", we exit the switch, but then it does the  */
					/*   processing to remove bytes from the buffer, but we want to leave */
					/*   them in the buffer for next time.                                */
					continue;
				}
			}
#endif

			/******************************/
			/*** Tracker Position Data? ***/
			if (VRPNdata.message_type == aux->MessageSensorPos) {
				/* tracking data = 64bit sensor number, 3*64 bit float location, 4*64bit float orientation (64 bytes) */
				int		count;
				int32_t	sensor;		/* the 8-byte (64bit) sensor value in the VRPN data packet */
				int32_t		int_sensor;	/* the unsigned 64 bit sensor value wreaks havoc with array access */

				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				/* redetermine the packet size now that we know the type of data */
				packet_size = HEADER_SIZE + 2*sizeof(sensor) + 7 * sizeof(vrpn_dev->posquat[0][0]);
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */

				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}


				/* now parse the data! */
				memmove(&sensor, &aux->buf[HEADER_SIZE], sizeof(sensor));
				//vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				//vrpn_dev = &(aux->device[VRPNdata.code+1]);
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&sensor);
#  endif
				int_sensor = (int32_t)sensor;		/* convert to an integer to avoid gcc array indexing bug */
				if (int_sensor < MAX_6SENSORS) {
					vrpn_dev->ever_reported_posquat_any = 1u;
					vrpn_dev->ever_reported_posquat[int_sensor] = 1u;
					vrpn_dev->newly_reported_posquat[int_sensor] = 1u;
					vrpn_dev->posquat_time[int_sensor] = vrpntime;
					for (count = 0; count < 7; count++) {
						memmove(&(vrpn_dev->posquat[int_sensor][count]), &aux->buf[HEADER_SIZE + 2*sizeof(sensor) + (count * sizeof(vrpn_dev->posquat[0][0]))], sizeof(vrpn_dev->posquat[0][0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
						vrEndianSwap8(&(vrpn_dev->posquat[int_sensor][count]));
#  endif
					}
					vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\rTime = %0.6f" NORM_TEXT " -- PosQuat data: device = %u sensor = %d, xyz = %.3f %.3f %.3f, quat = %.3f %.3f %.3f %.3f", vrpntime, VRPNdata.code, int_sensor, vrpn_dev->posquat[int_sensor][0], vrpn_dev->posquat[int_sensor][1], vrpn_dev->posquat[int_sensor][2], vrpn_dev->posquat[int_sensor][3], vrpn_dev->posquat[int_sensor][4], vrpn_dev->posquat[int_sensor][5], vrpn_dev->posquat[int_sensor][6]);
				} else {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of 6-sensor (%d) exceeds maximum (%d).\n", int_sensor+1, MAX_6SENSORS);
				}

			}
			/**********************/
			/*** Velocity Data? ***/
			else if (VRPNdata.message_type == aux->MessageSensorVel) {
				/* tracking data = 64bit sensor number, 3*64 bit float location vel, 4*64bit float orientation vel + one more unknown double! (72 bytes) */
				int		count;
				int32_t/*uint64_t*/	sensor;
				double		velquat[7];		/* Position Velocity values */
				double		velquat_deltat;		/* A Delta-time value for velquat */

				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				/* redetermine the packet size now that we know the type of data */
				packet_size = HEADER_SIZE + 2*sizeof(sensor) + 7 * sizeof(velquat[0]) + sizeof(velquat_deltat);
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */

				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}


				/* now parse the data! */
				memmove(&sensor, &aux->buf[HEADER_SIZE], sizeof(sensor));
				memmove(&velquat_deltat, &aux->buf[packet_size - sizeof(velquat_deltat)], sizeof(velquat_deltat)); /* copyt the last 8 bytes into "velquat_deltat" variable */
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&sensor);
				vrEndianSwap8(&velquat_deltat);
#  endif
				for (count = 0; count < 7; count++) {
					memmove(&velquat[count], &aux->buf[HEADER_SIZE + 2*sizeof(sensor) + (count * sizeof(velquat[0]))], sizeof(velquat[0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap8(&velquat[count]);
#  endif
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- VelQuat data: device = %u sensor = %d, xyz = %.3f %.3f %.3f, quat = %.3f %.3f %.3f %.3f, velquat_deltat = %.3f\n", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, sensor, velquat[0], velquat[1], velquat[2], velquat[3], velquat[4], velquat[5], velquat[6], velquat_deltat);

			}
			/**************************/
			/*** Acceleration Data? ***/
			else if (VRPNdata.message_type == aux->MessageSensorAcc) {
				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE + 72;
			}
			/**********************/
			/*** Valuator Data? ***/
			else if (VRPNdata.message_type == aux->MessageValuator) {
				/* valuator data = 64bit number of channels, num-channels * 64bit values */
				int		count;
				double		num_valuators;		/* Oddly this is sent as a double precision float */
				double		valuators[MAX_VALUATORS];

				memmove(&num_valuators, &aux->buf[HEADER_SIZE], sizeof(num_valuators));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap8(&num_valuators);
#  endif
				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE + sizeof(num_valuators) + sizeof(valuators[0]) * num_valuators;
				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}

				//vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				//vrpn_dev = &(aux->device[VRPNdata.code+1]);
				if (num_valuators > MAX_VALUATORS) {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of valuators available (%d) exceeds maximum (%d).\n", ((int)num_valuators), MAX_VALUATORS);
					num_valuators = MAX_VALUATORS;		/* cap number to maximum allowable */
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- (device = %u) Valuators (%d): ", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, (int)num_valuators);
				for (count = 0; count < num_valuators && count < MAX_VALUATORS; count++) {
					memmove(&valuators[count], &aux->buf[HEADER_SIZE + sizeof(num_valuators) + (count * sizeof(valuators[0]))], sizeof(valuators[0]));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap8(&valuators[count]);
#  endif
					if (!vrpn_dev->ever_reported_analog[count]) {
						/* This first time this analog/valuator input is received, set its min and max values */
						vrpn_dev->analog_max[count] = valuators[count];
						vrpn_dev->analog_min[count] = valuators[count];
					} else {
						/* otherwise adjust the min/max values as needed */
						if (valuators[count] > vrpn_dev->analog_max[count])
							vrpn_dev->analog_max[count] = valuators[count];
						if (valuators[count] < vrpn_dev->analog_min[count])
							vrpn_dev->analog_min[count] = valuators[count];
					}
					vrpn_dev->analog[count] = valuators[count];
					vrpn_dev->ever_reported_analog_any = 1u;
					vrpn_dev->ever_reported_analog[count] = 1u;
					vrpn_dev->newly_reported_analog[count] = 1u;
					vrpn_dev->analog_time[count] = vrpntime;

					vrDbgPrintfN(VRPN_DBGLVL, "% lf ", valuators[count]);
				}
				vrDbgPrintfN(VRPN_DBGLVL, "\n");
			}
			/***************************/
			/*** Single Button Data? ***/
			else if (VRPNdata.message_type == aux->MessageButton) {
				/* button change data = 32bit button number, 32bit button value (8 bytes) */
				uint32_t	button_number;
				uint32_t	button_value;

				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE + 8;
				if (aux->eobuf_pos < packet_size) {		/* NOTE: we won't need this once I assume there's at least 8 bytes after the header */
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d bytes)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}
				memmove(&button_number, &aux->buf[HEADER_SIZE], sizeof(button_number));
				memmove(&button_value, &aux->buf[HEADER_SIZE + sizeof(button_number)], sizeof(button_value));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&button_number);
				vrEndianSwap4(&button_value);
#  endif
				//vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				//vrpn_dev = &(aux->device[VRPNdata.code+1]);

				if (button_number < MAX_BUTTONS) {
					vrpn_dev->button[button_number] = button_value;
					vrpn_dev->ever_reported_button_any = 1u;
					vrpn_dev->ever_reported_button[button_number] = 1u;
					vrpn_dev->newly_reported_button[button_number] = 1u;
					vrpn_dev->button_time[button_number] = vrpntime;
				} else {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of button (%d) exceeds maximum (%d).\n", button_number+1, MAX_BUTTONS);
				}

				vrDbgPrintfN(VRPN_DBGLVL, BOLD_TEXT "\nTime = %u:%06u" NORM_TEXT " -- (device = %u) Button %d is now %d\n", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, button_number, button_value);
			}

			/************************/
			/*** All Button Data? ***/
			else if (VRPNdata.message_type == aux->MessageAllButtons) {
				/* button change data = 64bit button count, <N> 64bit button values */
				int		count;
				uint32_t	num_buttons;
				uint32_t	buttons[MAX_BUTTONS];

				memcpy(&num_buttons, &aux->buf[HEADER_SIZE], sizeof(num_buttons));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&num_buttons);
#  endif

				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE + sizeof(num_buttons) + sizeof(buttons[0]) * num_buttons;
				packet_size = (packet_size+7)&(~7);	/* use bit-wise operation to raise size to next 8-byte number */
//vrPrintf("adjusted packet_size = %d, num_buttons = %d\n", packet_size, num_buttons);	/* delete me */
				/* now determine whether there is enough data in the buffer! */
				if (aux->eobuf_pos < packet_size) {
					/* debug statement here */
					/* 01/12/13 NOTE: we should now never get here since we did a check before the switch */
					//vrErrPrintf("\nVRPN: Not enough data in the buffer (only %d bytes)\n\n\n", packet_size);
					vrErrPrintf("\nVRPN: Not enough data in the buffer (%d of %d needed bytes)\n\n\n", aux->eobuf_pos, packet_size);
					vrErrPrintf("VRPN: ************** SHOULDN'T GET HERE! need to drop out of the while loop here! *********\n");
					continue;	/* jump out of the loop if we don't have enough data */
				}

				//vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
				//vrpn_dev = &(aux->device[VRPNdata.code+1]);
				if (num_buttons > MAX_BUTTONS) {
					vrDbgPrintfN(COMMON_DBGLVL, "VRPN: Number of buttons available (%d) exceeds maximum (%d).\n", ((int)num_buttons), MAX_BUTTONS);
					num_buttons = MAX_BUTTONS;		/* cap number to maximum allowable */
				}

				vrDbgPrintfN(VRPN_DBGLVL FORCE_DBG, BOLD_TEXT "\nVRPN: Time = %u:%06u" NORM_TEXT " -- (device = %u) Buttons (%d): ", VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, (int)num_buttons);
#if 0 /* DEBUG (the Swap8 vs. Swap4 issue) */
printf("HEADER_SIZE = %d, sizeof(num_buttons) = %d, %d * sizeof(buttons[0]) = %d\n", HEADER_SIZE, sizeof(num_buttons), num_buttons, num_buttons * sizeof(buttons[0]));
for (int yo =0; yo < (HEADER_SIZE + sizeof(num_buttons) + (num_buttons * sizeof(buttons[0]))); yo++) printf("byte %3d: %x / %d\n", yo, aux->buf[yo], aux->buf[yo]);
#endif
				for (count = 0; count < num_buttons && count < MAX_BUTTONS; count++) {
					memmove(&(buttons[count]), &(aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))]), sizeof(buttons[0]));
#if 0 /* DEBUG (the Swap8 vs. Swap4 issue) */
printf("YO: moving %d bytes %0x %0x %0x %0x to %p\n", sizeof(buttons[0]),
aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))+0],
aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))+1],
aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))+2],
aux->buf[HEADER_SIZE + sizeof(num_buttons) + (count * sizeof(buttons[0]))+3], &(buttons[count])
);
#endif
#  if __BYTE_ORDER == __LITTLE_ENDIAN
					vrEndianSwap4(&buttons[count]);		/* WRS: was "Swap8" --> the source of a bug! */
#  endif
					vrpn_dev->button[count] = buttons[count];
					vrpn_dev->ever_reported_button_any = 1u;
					vrpn_dev->ever_reported_button[count] = 1u;
					vrpn_dev->newly_reported_button[count] = 1u;
					vrpn_dev->button_time[count] = vrpntime;
#if 0 /* DEBUG (the Swap8 vs. Swap4 issue) */
vrPrintf("VRPN All buttons: %d: b[count] = %d, any = %d, any[c] = %d, newly = %d\n", count, vrpn_dev->button[count], vrpn_dev->ever_reported_button_any, vrpn_dev->ever_reported_button[count], vrpn_dev->newly_reported_button[count]);
#endif

					vrDbgPrintfN(VRPN_DBGLVL, "%d ", buttons[count]);
				}
				vrDbgPrintfN(VRPN_DBGLVL, "\n");
			}

			/********************/
			/*** Ping message ***/
			else if (VRPNdata.message_type == aux->MessagePing) {
				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE;

				vrMsgPrintf("VRPN Client pinged us (<-) at %lf.\n", vrpntime);
				vrDbgPrintfN(AALWAYS_DBGLVL, "VRPN Client pinged us\n");
				aux->unanswered_ping = 1;
			}

			/********************/
			/*** Pong message ***/
			else if (VRPNdata.message_type == aux->MessagePong) {
				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE;

				//vrpn_dev = &(aux->device[VRPNdata.code+1]);
				vrpn_dev->newly_reported_pong = 1u;
				vrMsgPrintf("VRPN Client ponged us (->) at %lf.\n", vrpntime);
				vrDbgPrintfN(AALWAYS_DBGLVL, "VRPN Client ponged us\n");
			}

			/******************/
			/*** Text Data? ***/
			else if (VRPNdata.message_type == aux->MessageText) {
				int32_t		level;
				int32_t		severity;
				char		message[1024];	/* max VRPN text message size */

				/* NOTE: "packet_size" needs to be set to the expected size for this message */
				packet_size = HEADER_SIZE + sizeof(level) + sizeof(severity) + sizeof(message);

				/* now parse the data! */
				memcpy(&level, &aux->buf[HEADER_SIZE], sizeof(level));
				memcpy(&severity, &aux->buf[HEADER_SIZE + sizeof(level)], sizeof(severity));
				memcpy(message, &aux->buf[HEADER_SIZE + sizeof(level) + sizeof(severity)], sizeof(message));
				//vrpntime = VRPNdata.time_seconds + (VRPNdata.time_useconds * 0.000001);
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&level);
				vrEndianSwap4(&severity);
#  endif
				packet_size = padded_message_size;	/* skip the complete packet */

				/* now store the data */
				//vrpn_dev = &(aux->device[VRPNdata.code+1]);
				vrpn_dev->newly_reported_message = 1u;
				vrpn_dev->message_time = vrpntime;
				strncpy(vrpn_dev->text_message, message, sizeof(vrpn_dev->text_message));
				vrpn_dev->text_level = level;
				vrpn_dev->text_severity = severity;

				vrDbgPrintfN(AALWAYS_DBGLVL, "VRPN Server message (%d/%d): '%s'\n", level, severity, message);
				//vrPrintf("VRPN Server message (%d/%d): '%s'\n", level, severity, message); /* delete me */
			}

			/************************/
			/*** Unknown Datatype ***/
			else {
				memcpy(&message_length, &aux->buf[HEADER_SIZE], sizeof(message_length));
#  if __BYTE_ORDER == __LITTLE_ENDIAN
				vrEndianSwap4(&message_length);
#  endif
				message = &aux->buf[HEADER_SIZE + sizeof(message_length)];
				packet_size = padded_message_size;	/* skip the complete packet */

				//vrDbgPrintfN(VRPN_DBGLVL, "\nVRPN: Unknown data type -- %d (code = %u)\n\n", VRPNdata.message_type, VRPNdata.code); /* delete me */
				vrDbgPrintfN(VRPN_DBGLVL, RED_TEXT "\nVRPN: Unwatched data type -- '%s' %d (code = %u)\n" NORM_TEXT, aux->message_codes[VRPNdata.message_type], VRPNdata.message_type, VRPNdata.code);
				vrDbgPrintfN(VRPN_DBGLVL, "VRPN: bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d, string? = '%s'\n\n", aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type, message);

				/* print the information about the unknown message */
				int		count;
				vrPrintf(RED_TEXT "Unwatched data header:" NORM_TEXT " size = %d, time = %d:%d, code = %d ('%s'), type = %d ('%s') (padding = %d)\n\tPayload Bytes:",
					VRPNdata.message_size,
					VRPNdata.time_seconds,
					VRPNdata.time_useconds,
					VRPNdata.code, aux->device[VRPNdata.code+1].name,
					VRPNdata.message_type, aux->message_codes[VRPNdata.message_type],
					VRPNdata.padding);		/* (NOTE: VRPN hides a message count value here.) */
				for (count = HEADER_SIZE; count < VRPNdata.message_size; count++) {
					vrPrintf(" %02x", (unsigned char)aux->buf[count]);
				}
				/* TODO: also output the payload as 4-byte integers, 8-byte integers, doubles, and ASCII */
				vrPrintf("\n");
#if 0 /* Set this to "1" to exit which can make it easier to analyze the data ordering */
				exit(0);
#endif
			}
		}

		/* Do a sanity check!  (the reported message size should match the actual message size) */
		/* NOTE to developers: if you get this error, make sure that the "packet_size" variable */
		/*   is being set to the expected packet size for the given message!  (This has messed  */
		/*   me up a couple times!)                                                              */
		if (padded_message_size != packet_size) {
			vrDbgPrintf(RED_TEXT "VRPN: SANITY FAIL: padded_message_size = %d, packet_size = %d -- bytes = %d, mes-size = %u, time = %u:%06u, code = %u, mes-type = %d\n" NORM_TEXT, padded_message_size, packet_size, aux->eobuf_pos, VRPNdata.message_size, VRPNdata.time_seconds, VRPNdata.time_useconds, VRPNdata.code, VRPNdata.message_type);
				vrPrintf(RED_TEXT "Unwatched data header:" NORM_TEXT " size = %d, time = %d:%d, code = %d ('%s'), type = %d ('%s')\n\tPayload Bytes: xxx\n",
					VRPNdata.message_size,
					VRPNdata.time_seconds,
					VRPNdata.time_useconds,
					VRPNdata.code, aux->device[VRPNdata.code+1].name,
					VRPNdata.message_type, aux->message_codes[VRPNdata.message_type]);
			abort();	/* WRS: 03/28/22: When the "send_all" flag is set in _VrpnPutData(), the message sent causes this sanity check to fail. */
		}

		/* Now shift the buffer by the message size (padded to 8 byte boundaries) */
		memmove(aux->buf, &aux->buf[packet_size], (aux->eobuf_pos - packet_size));	/* shift the data */
		aux->eobuf_pos -= packet_size;
		packets_decoded++;

		/* Reset the packet size to the minimal size required for parsing */
		packet_size = HEADER_SIZE;	/* There at least must be data for the first header */  /* NOTE: in reality, we'll always need the first number as well */
	}

#  if 1 /* actually, right now [11/26/21] if read_result is 0 that's because we read 0 bytes -- and often I don't provide any room in the buffer for reading */
	/* [11/26/21: okay, so I'm trying this bit of code again, after changing the read */
	/*   at the beginning of _VrpnReadInput(), to no longer both reading when there's */
	/*   no room in the inn-put buffer.  But I guess if the VRPN server is very slow  */
	/*   that might also result in zero-byte reads.  So this is just an experiment.   */

	/* Return -1 if the socket connection is closed */
	if (read_result == 0) {
		vrDbgPrintf("\n\n_VRPNReadInput(): socket connection closed -- %d packets decoded in this pass\n", packets_decoded);
		return (-1);
	}
#  endif
	return (packets_decoded);
#endif /* } !VRPN_LIB */
}


#ifdef VRPN_LIB /* { */
/******************************************************************************/
/* _VrpnlibHandleButton(): For the VRPN_LIB version of this input interface,  */
/*    this is the callback routine to handle button ("2switch") inputs.  The  */
/*    "userdata" argument is linked to the callback for each device, and thus */
/*    contains the information needed to map the button data being handled    */
/*    to a particular FreeVR input.                                           */
/*  Used by _VrpnInitializeDevice().                                          */
static void _VrpnHandleButton(_VrpnUserdata *userdata, const vrpn_BUTTONCB but_vrpn)
{
	_VrpnPrivateInfo	*aux;
	vr2switch		*input;
	int			vrpn_device_num;

#  if 0 /* DEBUG */
	vrPrintf("_VrpnHandleButton(): Button %3d is in state: %d\n", but_vrpn.button, but_vrpn.state);
#  endif

	if (but_vrpn.button >= MAX_BUTTONS) {
		/* we've gone above the arbitrarily set maximum number of button inputs from any given device */
		vrErrPrintf("_VrpnHandleButton(): invalid button number from VRPN (%d)\n", but_vrpn.button);
		return;
	}

	aux = userdata->aux;
	vrpn_device_num = userdata->num;

	input = aux->button_inputs[vrpn_device_num][but_vrpn.button];

	vrDbgPrintfN(INPUT_DBGLVL,
		"_VrpnHandleButton(): device = %d, button = %d, state = %d, input = %p\n",
		vrpn_device_num, but_vrpn.button, but_vrpn.state, input);

	if (input != NULL) {
		switch (input->input_type) {
		case VRINPUT_BINARY:
			vrAssign2switchValue(input, but_vrpn.state);
			break;
		case VRINPUT_CONTROL:
			vrInvokeDynamicCallback(((vrControl *)(input))->callback, 1, but_vrpn.state);
			break;
		default:
			vrErrPrintf(RED_TEXT "_VrpnHandleButton: Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
			break;
		}
	}


#  ifdef CAVE
	vrpn_controls.button[but_vrpn.button] = but_vrpn.state;
#  endif
}

/******************************************************************************/
/* _VrpnlibHandleAnalog(): For the VRPN_LIB version of this input interface,  */
/*    this is the callback routine to handle analog ("valuator") inputs.  The */
/*    "userdata" argument is linked to the callback for each device, and thus */
/*    contains the information needed to map the analog data being handled    */
/*    to a particular FreeVR input.                                           */
/*  Used by _VrpnInitializeDevice().                                          */
static void _VrpnHandleAnalog(_VrpnUserdata *userdata, const vrpn_ANALOGCB analog_vrpn)
{
	_VrpnPrivateInfo	*aux;
	vrValuator		*input;
	int			vrpn_device_num;
	int			count;			/* for looping over each analog channel */

#  if 0
	vrPrintf("_VrpnHandleAnalog: Analog has %3d channels w/ values: %lf %lf %lf ...\n",
		analog_vrpn.num_channel, analog_vrpn.channel[0], analog_vrpn.channel[1], analog_vrpn.channel[2]);
#  endif

	aux = userdata->aux;
	vrpn_device_num = userdata->num;

	for (count = 0; count < analog_vrpn.num_channel && count < MAX_VALUATORS; count++) {
		input = aux->valuator_inputs[vrpn_device_num][count];

		vrDbgPrintfN(INPUT_DBGLVL,
			"_VrpnHandleAnalog(): device = %d, analog channel = %d, value = %f, input = %p\n",
			vrpn_device_num, count, analog_vrpn.channel[count], input);

		if (input != NULL) {
			switch (input->input_type) {
			case VRINPUT_VALUATOR:
				vrAssignValuatorValue(input, analog_vrpn.channel[count] * aux->scale_valuator);
				break;

			default:
				/* TODO: ... */
				break;
			}
		}
	}

#  ifdef CAVE
	vrpn_controls.valuator[0] = analog_vrpn.channel[0] * 3.5;
	vrpn_controls.valuator[1] = analog_vrpn.channel[1] * 3.5;
#  endif
}


/******************************************************************************/
/* _VrpnlibHandleTracker(): For the VRPN_LIB version of this input interface, */
/*    this is the callback routine to handle position tracker inputs.  The    */
/*    "userdata" argument is linked to the callback for each device, and thus */
/*    contains the information needed to map the tracker data being handled   */
/*    to a particular FreeVR input.                                           */
/*  Used by _VrpnInitializeDevice().                                          */
static void _VrpnlibHandleTracker(_VrpnUserdata *userdata, const vrpn_TRACKERCB tracker_vrpn)
{
	_VrpnPrivateInfo	*aux;
	vr6sensor		*input;
	vrMatrix		mat_store;
	vrQuat			tmpquat;
	int			vrpn_device_num;

#  if 0
	vrPrintf("_VrpnlibHandleTracker: Sensor %3d is now at (%11f, %11f, %11f)\n",
		tracker_vrpn.sensor, tracker_vrpn.pos[0], tracker_vrpn.pos[1], tracker_vrpn.pos[2]);
#  endif

	if (tracker_vrpn.sensor >= MAX_6SENSORS) {
		/* we've gone above the arbitrarily set maximum number of tracker inputs from any given device */
		vrErrPrintf("_VrpnlibHandleTracker(): invalid sensor number from VRPN (%d)\n", tracker_vrpn.sensor);
		return;
	}

	aux = userdata->aux;
	vrpn_device_num = userdata->num;

	input = aux->sensor6_inputs[vrpn_device_num][tracker_vrpn.sensor];

	vrDbgPrintfN(INPUT_DBGLVL,
		"_VrpnlibHandleTracker(): device = %d, sensor# = %d, loc = (%5.3f %5.3f %5.3f), input = %p\n",
		vrpn_device_num, tracker_vrpn.sensor,
		tracker_vrpn.pos[0], tracker_vrpn.pos[1], tracker_vrpn.pos[2],
		input);


	if (input != NULL) {
		switch (input->input_type) {
		case VRINPUT_6SENSOR:
			vrMatrixSetTranslation3d(&mat_store, tracker_vrpn.pos[0], tracker_vrpn.pos[1], tracker_vrpn.pos[2]);	/* this sets rot to ID */

			tmpquat.v[0] = tracker_vrpn.quat[0];
			tmpquat.v[1] = tracker_vrpn.quat[1];
			tmpquat.v[2] = tracker_vrpn.quat[2];
			tmpquat.v[3] = tracker_vrpn.quat[3];
			vrMatrixSetRotationFromQuat(&mat_store, &tmpquat);

			vrAssign6sensorValue(input, &mat_store, 0 /*, TODO: timestamp */);
			break;
		default:
			vrErrPrintf(RED_TEXT "_VrpnlibHandleTracker(): Unable to handle VRPN tracker inputs that aren't 6-sensor inputs\n" NORM_TEXT);
			break;
		}
	}

#  ifdef CAVE
	/* raise value of num_sensors when info from a larger sensor number is received */
	if (tracker_dev.sensor > num_sensors) {
		num_sensors = tracker_dev.sensor+1;
		printf("CAVE VRPN-tracker: num_sensors is now %d\n", num_sensors);
	}

	vrpn_sensor[tracker_vrpn.sensor].x = tracker_vrpn.pos[0];
	vrpn_sensor[tracker_vrpn.sensor].y = tracker_vrpn.pos[1];
	vrpn_sensor[tracker_vrpn.sensor].z = tracker_vrpn.pos[2];

	... = tracker_vrpn.quat[0]; (qx)
	... = tracker_vrpn.quat[1]; (qy)
	... = tracker_vrpn.quat[2]; (qz)
	... = tracker_vrpn.quat[3]; (qw)
	vrpn_sensor[tracker_vrpn.sensor].azim = 0.0;
	vrpn_sensor[tracker_vrpn.sensor].elev = 0.0;
	vrpn_sensor[tracker_vrpn.sensor].roll = 0.0;

	CAVEGetTimestamp(&vrpn_sensor[tracker_vrpn.sensor].timestamp);
#  endif
}

#endif /* } VRPN_LIB */

/**********************************************************/
/* _VrpnInitializeDevice() is called in the OPEN phase    */
/*   of input interface -- after the types of inputs have */
/*   been determined (during the CREATE phase).           */
/* For VRPN, after the socket connection has been made,   */
/*   and the version number obtained.                     */
static int _VrpnInitializeDevice(_VrpnPrivateInfo *aux)
{
	int		device_count;		/* for looping through list of each device type */
	int		count;			/* for looping over inputs of each type */
#ifdef VRPN_LIB /* { */
	_VrpnUserdata	*userdata;		/* struct for passing info to the VRPN library callback */

	if (aux == NULL)
		return -1;

	/* TODO: verify that the VRPN functions exist before calling them */
	/*   if they don't exist, print a warning and return (-1).        */

	/************************************************/
	/* make the connection(s) to the VRPN server(s) */
	for (device_count = 0; device_count < aux->num_button_devices; device_count++) {
		/* create the userdata for the callback */
		userdata = (_VrpnUserdata *)malloc(sizeof(*userdata));
		userdata->num = device_count;
		userdata->aux = aux;

		/* open the VRPN Button connection and set the incoming data callback */
#  if 0 /* attempt to test for the existence of the function */
		vrPrintf("About to create new vrpn Button device '%s' (first check for function existence)\n", aux->button_names[device_count]);
		if ((void *)vrpn_Button_Remote == NULL)
			vrPrintf(RED_TEXT "cannot find function 'vrpn_Button_Remote'\n" NORM_TEXT);
#  endif
		aux->vrpn_button[device_count] = new vrpn_Button_Remote(aux->button_names[device_count]);
		aux->vrpn_button[device_count]->register_change_handler((void *)userdata, (vrpn_BUTTONCHANGEHANDLER)_VrpnHandleButton);
	}

	for (device_count = 0; device_count < aux->num_analog_devices; device_count++) {
		/* create the userdata for the callback */
		userdata = (_VrpnUserdata *)malloc(sizeof(*userdata));
		userdata->num = device_count;
		userdata->aux = aux;

		/* open the VRPN Analog connection and set the incoming data callback */
		aux->vrpn_analog[device_count] = new vrpn_Analog_Remote(aux->analog_names[device_count]);
		aux->vrpn_analog[device_count]->register_change_handler((void *)userdata, (vrpn_ANALOGCHANGEHANDLER)_VrpnHandleAnalog);
	}

	for (device_count = 0; device_count < aux->num_tracker_devices; device_count++) {
		/* create the userdata for the callback */
		userdata = (_VrpnUserdata *)malloc(sizeof(*userdata));
		userdata->num = device_count;
		userdata->aux = aux;

		/* open the VRPN Tracker connection and set the incoming data callback */
		aux->vrpn_tracker[device_count] = new vrpn_Tracker_Remote(aux->tracker_names[device_count]);
		aux->vrpn_tracker[device_count]->register_change_handler((void *)userdata, (vrpn_TRACKERCHANGEHANDLER)_VrpnlibHandleTracker);
	}
#else /* } <-- VRPN_LIB vs. not VRPN_LIB --> { */
	int	retval;		/* value returned by _VrpnReadInput() */

	/********************/
	/* clear the palate */
	retval = _VrpnReadInput(aux);		/* NOTE: the first reading of VRPN data will often contain all the setup information */
	/* if we didn't get any data, try again for 10 times */
	for (count = 0; count < 100/*10*/ && retval <= 0; count++) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnInitializeDevice(): " RED_TEXT "VRPN Server is not yet sending enough data.  Trying again...\n" NORM_TEXT);
		vrSleep(100000);
		retval = _VrpnReadInput(aux);
		if (retval > 0)
			/* What we know at this point is that we've successfully read */
			/*   *some* data -- there may be more handshake data to come. */
			vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnInitializeDevice(): " RED_TEXT "... Success!\n" NORM_TEXT);
	}

#  ifdef FREEVR /* { */
	/* set each VRPN input's device pointer to device matching device named in string */
	if (aux->num_devices > 0) {
		/* buttons */
		for (count = 0; count < aux->num_buttons; count++) {
			/* find an existing device with the matching name */
			for (device_count = 0; device_count < aux->num_devices; device_count++) {
				if (!strcmp(aux->button_map_devicename[count], aux->device[device_count].name)) {
					aux->button_map_device[count] = &(aux->device[device_count]);
					aux->button_map_device[count]->mapped_button[aux->button_map_button[count]] = 1;
				}
			}
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeDevice(): Just mapped button[%d](%s) to %p\n", count, aux->button_map_devicename[count], aux->button_map_device[count]);
			if (aux->button_map_device[count] == NULL) {
				vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnInitializeDevice(): " RED_TEXT BOLD_TEXT "Alert! VRPN Server is not sending the requested device '%s' for button input %d.\n" NORM_TEXT, aux->button_map_devicename[count], count);
			}
		}

		/* valuators */
		for (count = 0; count < aux->num_valuators; count++) {
			/* find an existing device with the matching name */
			for (device_count = 0; device_count < aux->num_devices; device_count++) {
				if (!strcmp(aux->valuator_map_devicename[count], aux->device[device_count].name)) {
					aux->valuator_map_device[count] = &(aux->device[device_count]);
					aux->valuator_map_device[count]->mapped_analog[aux->valuator_map_valuator[count]] = 1;
				}
			}
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeDevice(): Just mapped valuator[%d](%s) to %p\n", count, aux->valuator_map_devicename[count], aux->valuator_map_device[count]);
			if (aux->valuator_map_device[count] == NULL) {
				vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnInitializeDevice(): " RED_TEXT BOLD_TEXT "Alert! VRPN Server is not sending the requested device '%s' for valuator input %d.\n" NORM_TEXT, aux->valuator_map_devicename[count], count);
			}
		}

		/* 6sensors */
		for (count = 0; count < aux->num_6sensors; count++) {
			/* find an existing device with the matching name */
			for (device_count = 0; device_count < aux->num_devices; device_count++) {
				if (!strcmp(aux->tracker_map_devicename[count], aux->device[device_count].name)) {
					aux->tracker_map_device[count] = &(aux->device[device_count]);
					aux->tracker_map_device[count]->mapped_tracker[aux->tracker_map_tracker[count]] = 1;
				}
			}
			vrDbgPrintfN(VRPN_DBGLVL, "_VrpnInitializeDevice(): Just mapped tracker[%d](%s) to %p\n", count, aux->tracker_map_devicename[count], aux->tracker_map_device[count]);
			if (aux->tracker_map_device[count] == NULL) {
				vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnInitializeDevice(): " RED_TEXT BOLD_TEXT "Alert! VRPN Server is not sending the requested device '%s' for tracker input %d.\n" NORM_TEXT, aux->tracker_map_devicename[count], count);
			}
		}

		aux->mapped = 1;
	}

	/* find the first simulator-sensor */
	for (aux->active_sim6sensor = 0
		; /* conditional */							/* keep looping if: */
		   (aux->active_sim6sensor < aux->num_6sensors)				/*   still within the list bounds */
		&& (aux->sensor6_inputs[aux->active_sim6sensor] != NULL)			/*   tracker does not exist! */
		&& (aux->tracker_map_device[aux->active_sim6sensor] != &(aux->device[0]))	/*   tracker is not from device[0] (ie. simulator) */
		; /* increment */
		aux->active_sim6sensor++)
#    if 0	/* delete me */
  vrPrintf("!!!!!! for: active_sim6sensor=%d, num_6=%d, aux->sensor6_inputs[%d]=%p, aux->tracker_map_device[aux->active_sim6sensor]->name='%s'\n", aux->active_sim6sensor, aux->num_6sensors, aux->active_sim6sensor, aux->sensor6_inputs[aux->active_sim6sensor], aux->tracker_map_device[aux->active_sim6sensor]->name)
#    endif
		; /* no-op -- the looping is the reward */
	if (aux->active_sim6sensor >= aux->num_6sensors) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR (%.4f): " RED_TEXT "VRPN Warning, there is no simulated 6-sensor (not a problem if not planning to use one).\n" NORM_TEXT, vrCurrentSimTime());
	} else {
		vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sim6sensor]), 1);
		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: 6sensor[%d] now active.\n", aux->active_sim6sensor);
	}
#  endif /* } FREEVR */

	if (vrDbgDo(VRPN_DBGLVL)) {
		_VrpnPrintPrivateInfo(stdout, aux, verbose);
	}

	/**********************************/
	/* request version and param info */
	/* TODO: this */

	/********************************/
	/* send some device setup codes */
	/* TODO: this (if necessary) */
#  if 0
	/* At this point for VRPN servers compiled with the ??? option, we  */
	/*   would need to inform the server what tracker/input data we are */
	/*   interested in.  But typically VRPN servers are not configured  */
	/*   that way and just send all available data.                     */
	_VRPNSendSenderMessage(fd_socket, 0, "head", RELIABLE);
	_VRPNSendSenderMessage(fd_socket, 1, "wand", RELIABLE);
#  endif

#  if 0
	/************************/
	/* read some input data */
	_VrpnReadInput(aux);
#  endif
#endif /* } */

	return 0;
}

/***********************************************************/
/* NOTE: This input system does not require the functions: */
/*    * _DEVICECloseDevice()                               */
/*    * _DEVICEResetDevice()                               */
/*    * _DEVICEButtonValue()                               */
/*    * _DEVICEValuatorValue()                             */
/***********************************************************/




	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/


#if defined(FREEVR) /* { */
/*****************************************************************/
/***   Functions for FreeVR access of VRPN devices for input   ***/
/*****************************************************************/


	/************************************/
	/***  FreeVR NON public routines  ***/
	/************************************/


/************************************************************************/
/* _VrpnParseArgs(): this function is used to read input device (or     */
/*    stream) configuration options from the FreeVR configuration file, */
/*    command-line or environment variable string.  The meaning and     */
/*    syntax for each option is described at the top of this source     */
/*    file (and will often be repeated in the man page documentation    */
/*    at the end.                                                       */
static void _VrpnParseArgs(_VrpnPrivateInfo *aux, char *args)
{
	float	volume_values[6];		/* for reading the working volume array */
	float	scale_value = -1.0;		/* for reading one of the scaling factor values */

	/* In the case of no arguments, just return */
	if (args == NULL)
		return;

	/******************************************/
	/** Argument format: "host" "=" hostname **/
	/******************************************/
	vrArgParseString(args, "host", &(aux->vrpn_host));

	/**********************************************/
	/** Argument format: "hostname" "=" hostname **/
	/**********************************************/
	vrArgParseString(args, "hostname", &(aux->vrpn_host));

	/**************************************/
	/** Argument format: "port" "=" file **/
	/**************************************/
	vrArgParseInteger(args, "port", &(aux->vrpn_port));

#  if 1 /* not sure whether we'll want this */
	/********************************************/
	/** Argument format: "valScale" "=" number **/
	/********************************************/
	vrArgParseFloat(args, "valscale", &(aux->scale_valuator));
#  endif

	/**********************************************/
	/** Argument format: "trackScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "trackscale", &scale_value)) {
		aux->scale_tracker_trans = scale_value;
	}

	/*****************************************/
	/** Argument format: "scale" "=" number **/ /* (redundant with "trackscale") */
	/*****************************************/
	if (vrArgParseFloat(args, "scale", &scale_value)) {
		aux->scale_tracker_trans = scale_value;
	}

	/******************************************************************/
	/** Argument format: "govern_time" "=" { "on" | "off" | number } **/
	/******************************************************************/
	(vrArgParseBool(args, "govern_time", &(aux->govern_time)));

		/*************************************/
		/* arguments for simulated 6-sensors */

	/***************************************************************/
	/** Argument format: "restrict" "=" { "on" | "off" | number } **/
	/***************************************************************/
	vrArgParseBool(args, "restrict", &(aux->sensor6_options.restrict_space));

	/***********************************************************************/
	/** Argument format: "valuatorOverride" "=" { "on" | "off" | number } **/
	/***********************************************************************/
	vrArgParseBool(args, "valuatoroverride", &(aux->sensor6_options.ignore_trans));

	/*******************************************************************/
	/** Argument format: "returnToZero" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "returntozero", &(aux->sensor6_options.return_to_zero));

	/******************************************************************/
	/** Argument format: "relativeRot" "=" { "on" | "off" | number } **/
	/******************************************************************/
	vrArgParseBool(args, "relativerot", &(aux->sensor6_options.relative_axis));

	/*******************************************************************/
	/** Argument format: "swapTransRot" "=" { "on" | "off" | number } **/
	/*******************************************************************/
	vrArgParseBool(args, "swaptransrot", &(aux->sensor6_options.swap_transrot));

	/*********************************************************************/
	/** Argument format: "sensor_swap_yz" "=" { "on" | "off" | number } **/
	/*********************************************************************/
	vrArgParseBool(args, "sensor_swap_yz", &(aux->sensor6_options.swap_yz));
	vrArgParseBool(args, "sensor_swap_upin", &(aux->sensor6_options.swap_yz));

	/******************************************************************************************/
	/** Argument format: "workingVolume" "=" <float> <float> <float> <float> <float> <float> **/
	/******************************************************************************************/
	if (vrArgParseFloatList(args, "workingvolume", volume_values, 6)) {
		aux->sensor6_options.working_volume_min[VR_X] = volume_values[0];
		aux->sensor6_options.working_volume_max[VR_X] = volume_values[1];
		aux->sensor6_options.working_volume_min[VR_Y] = volume_values[2];
		aux->sensor6_options.working_volume_max[VR_Y] = volume_values[3];
		aux->sensor6_options.working_volume_min[VR_Z] = volume_values[4];
		aux->sensor6_options.working_volume_max[VR_Z] = volume_values[5];
	}

	/**********************************************/
	/** Argument format: "transScale" "=" number **/
	/**********************************************/
	if (vrArgParseFloat(args, "transscale", &(aux->sensor6_options.trans_scale))) {
		aux->sensor6_options.trans_scale *= TRANS_SENSITIVITY;
	}

	/********************************************/
	/** Argument format: "rotScale" "=" number **/
	/********************************************/
	if (vrArgParseFloat(args, "rotscale", &(aux->sensor6_options.rot_scale))) {
		aux->sensor6_options.rot_scale *= ROT_SENSITIVITY;
	}


	/** TODO: other arguments to parse go here **/
}


/***************************************************************************/
/* The _VrpnGetData(): function calls the _VruiDDReadInput() function to   */
/*   get the latest data into the generic portion of the "_VrpnGetData"    */
/*   structure.  Then this routine puts that data into the FreeVR data     */
/*   structures.  I.e. this is the FreeVR-specific portion of the data     */
/*   parsing.                                                              */
static void _VrpnGetData(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			count;			/* for looping over things */
	VRPNdevice		*vrpn_dev;		/* pointer to a device in the VRPN device list */
	int			input;
	int			vrpn_map;		/* local storage of the mapping code from FreeVR input number to VRPN input number */
	float			valuator_value;		/* holder for compositing a value together from components (e.g. sign & value) */
	int			retval;			/* value returned by _VrpnReadInput() */

	/*******************************************/
	/* Confirm the device is reported as ready */
	/* TODO: perhaps check to see whether the input structure has been initialized yet, and if not, call it AGAIN here */
	/* If the input mapping hasn't taken place yet, try again */
	if (!aux->mapped) {
		vrErrPrintf("_VrpnGetData(): trying again to initialize the input mapping!\n");
		vrSleep(10000);				/* allow a short amount of time to pass */
		_VrpnInitializeDevice(aux);
	}

	/* If still not mapped, return so other inputs can do some work */
	if (!aux->mapped) {
		vrDbgPrintfN(1, "_VrpnGetData(): VRPN inputs not yet mapped, skipping!\n");
		return;
	}

	/*******************/
	/* gather the data */
	retval = _VrpnReadInput(aux);
	if (retval < 1)
		return;

	/****************************************/
	/* Handle button (2-switch) information */
	for (count = 0; count < aux->num_buttons; count++) {
		vrpn_map = aux->button_map_button[count];
		vrpn_dev = aux->button_map_device[count];

		/* verify we have a good button input */
		if ((aux->button_inputs[count] != NULL) && (vrpn_dev != NULL)) {

			/* only handle buttons with a change of state (ie. new data from the VRPN server) */
			if (vrpn_dev->newly_reported_button[vrpn_map]) {
#  if 0
printf("button change on device '%s', num %d\n", vrpn_dev->name, vrpn_map);
#  endif
				switch (aux->button_inputs[count]->input_type) {
				case VRINPUT_BINARY:
					vrAssign2switchValue((vr2switch *)(aux->button_inputs[count]), (vrpn_dev->button[vrpn_map] != 0));
					vrpn_dev->newly_reported_button[vrpn_map] = 0u;
#  if 0	/* delete me */
vrPrintf("Assignment: VRPN button %d maps to FreeVR button %d, max = %d, val = %d, prev = %d\n", vrpn_map, count, aux->num_dd_buttons, aux->incoming_buttons[vrpn_map], aux->incoming_buttons_prev[vrpn_map]);
#  endif
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(aux->button_inputs[count]))->callback, 1, (vrpn_dev->button[vrpn_map] != 0));
					vrpn_dev->newly_reported_button[vrpn_map] = 0u;
					break;
				default:
					vrErrPrintf(RED_TEXT "_VrpnGetData(): Unable to handle button inputs that aren't Binary or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	}

	/****************************************/
	/* Handle valuator (analog) information */
	for (count = 0; count < aux->num_valuators; count++) {
		vrpn_map = aux->valuator_map_valuator[count];
		vrpn_dev = aux->valuator_map_device[count];

		/* verify we have a good valuator input */
		if ((aux->valuator_inputs[count] != NULL) && (vrpn_dev != NULL)) {

			/* only handle valuators with a change of state (ie. new data from the VRPN server) */
			if (vrpn_dev->newly_reported_analog[vrpn_map]) {
#  if 0	/* delete me */
vrPrintf("valuator change on device '%s', num %d\n", vrpn_dev->name, vrpn_map);
#  endif
				/* TODO: add aux->scale_valuator feature */
				valuator_value = vrpn_dev->analog[vrpn_map] * aux->scale_valuator * aux->valuator_map_sign[count];
				switch (aux->valuator_inputs[count]->input_type) {
				case VRINPUT_VALUATOR:
					vrAssignValuatorValue((vrValuator *)(aux->valuator_inputs[count]), valuator_value);
					vrpn_dev->newly_reported_analog[vrpn_map] = 0u;
#  if 0	/* delete me */
vrPrintf("Assignment: VRPN valuator %d maps to FreeVR valuator %d, max = %d, val = %d, prev = %d\n", vrpn_map, count, aux->num_dd_valuators, aux->incoming_valuators[vrpn_map], aux->incoming_valuators_prev[vrpn_map]);
#  endif
					break;
				case VRINPUT_CONTROL:
					vrCallbackInvokeDynamic(((vrControl *)(aux->valuator_inputs[count]))->callback, 1, &valuator_value);
					vrpn_dev->newly_reported_analog[vrpn_map] = 0u;
					break;
				default:
					vrErrPrintf(RED_TEXT "_VrpnGetData(): Unable to handle valuator inputs that aren't Binary or Control inputs\n" NORM_TEXT);
					break;
				}
			}
		}
	}

	/****************************************/
	/* Handle 6sensor (tracker) information */
	for (count = 0; count < aux->num_6sensors; count++) {
		vrMatrix	new_position;
		vrVector	new_location;
		vrQuat		new_orientation;
		float		scale_tracker_trans = aux->scale_tracker_trans;

		vrpn_map = aux->tracker_map_tracker[count];
		vrpn_dev = aux->tracker_map_device[count];

		/* verify we have a good tracker and that there is new data! */
		if ((aux->sensor6_inputs[count] != NULL) && (vrpn_dev != NULL) && vrpn_dev->newly_reported_posquat[vrpn_map]) {
			/* Copy the incoming tracker data into the FreeVR structure */
			new_location.v[VR_X] = vrpn_dev->posquat[vrpn_map][VR_X] * scale_tracker_trans;
			new_location.v[VR_Y] = vrpn_dev->posquat[vrpn_map][VR_Y] * scale_tracker_trans;
			new_location.v[VR_Z] = vrpn_dev->posquat[vrpn_map][VR_Z] * scale_tracker_trans;

			new_orientation.v[VR_X] = vrpn_dev->posquat[vrpn_map][VR_QX];
			new_orientation.v[VR_Y] = vrpn_dev->posquat[vrpn_map][VR_QY];
			new_orientation.v[VR_Z] = vrpn_dev->posquat[vrpn_map][VR_QZ];
			new_orientation.v[VR_W] = vrpn_dev->posquat[vrpn_map][VR_QW];

			vrMatrixSetIdentity(&new_position);
			vrMatrixSetTransFromVector(&new_position, &new_location);
			vrMatrixSetRotationFromQuat(&new_position, &new_orientation);	// look here

			/* Since 6-sensors generally change rapidly, there  */
			/*   is little point and often moderate expense in  */
			/*   comparing for changes from the current value   */
			/*   to the new value.  (Although perhaps watching  */
			/*   the timestamp might be a useful way to do this.*/
			/* TODO: perhaps verify that this is indeed a VRINPUT_6SENSOR */
			if (vrpn_dev == &(aux->device[0])) {
				/* device[0] is the simulated 6-sensor device */
				if (!aux->sensor6_options.ignore_all && aux->sim6sensor_change) {
					vrAssign6sensorValueFromValuators(aux->sensor6_inputs[count], aux->sim_values, &(aux->sensor6_options), -1 /* (ie. no change to oob flag) */);
					if ((aux->sim_values[VR_X] == 0)
					   && (aux->sim_values[VR_Y] == 0)
					   && (aux->sim_values[VR_Z] == 0)
					   && (aux->sim_values[VR_AZIM+3] == 0)
					   && (aux->sim_values[VR_ELEV+3] == 0)
					   && (aux->sim_values[VR_ROLL+3] == 0)) {
						aux->sim6sensor_change = 0;
						vrpn_dev->newly_reported_posquat[vrpn_map] = 0u;
					}
				}
			} else {
				/* all other devices are actually from the VRPN server */
				vrAssign6sensorValue(aux->sensor6_inputs[count], &new_position, 0 /* TODO: timestamp: , vrpn_dev->posquat_time[vrpn_map] */);
				vrpn_dev->newly_reported_posquat[vrpn_map] = 0u;
			}

		}
	}

#  if 0 /* this just prints all data -- for code development only { */
	/****************************************/
	/* Handle button (2-switch) information */
	for (count = 0; count < MAX_DEVICES; count++) {
		vrpn_dev = &(aux->device[count]);
		for (input = 0; input < MAX_BUTTONS; input++) {
			if (vrpn_dev->newly_reported_button[input]) {
				printf("dev[%s]->button[%d] = %d\n", vrpn_dev->name, input, vrpn_dev->button[input]);
				vrpn_dev->newly_reported_button[input] = 0u;
			}
		}
	}

	/****************************************/
	/* Handle analog (valuator) information */
	for (count = 0; count < MAX_DEVICES; count++) {
		int	print_flag = 0;		/* temporary */
		vrpn_dev = &(aux->device[count]);
		for (input = 0; input < MAX_VALUATORS; input++) {
			if (vrpn_dev->newly_reported_analog[input]) {
				printf("dev[%s]->val[%d] = % 6.3f  ", vrpn_dev->name, input, vrpn_dev->analog[input]);
				vrpn_dev->newly_reported_analog[input] = 0u;
				print_flag = 1;	/* temporary */
			}
		}
		if (print_flag)		/* temporary */
			printf("\n");	/* temporary */
	}

	/*****************************************/
	/* Handle tracker (6-sensor) information */
	for (count = 0; count < MAX_DEVICES; count++) {
		vrpn_dev = &(aux->device[count]);
		for (input = 0; input < MAX_6SENSORS; input++) {
			if (vrpn_dev->newly_reported_posquat[input]) {
				printf("dev[%s]->s%d x:%.3f y:%.3f z:%.3f  qx:%.3f qy:%.3f qz:%.3f qw:%.3f\n",
					vrpn_dev->name, input,
					vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z],
					vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
				vrpn_dev->newly_reported_posquat[input] = 0u;
			}
		}
	}
#  endif /* } */
}


	/****************************************************************/
	/*      Function(s) for parsing VRPN "input" declarations.      */
	/*                                                              */
	/*  These _VRPN<type>Input() functions are called during the    */
	/*  CREATE phase of the input interface.                        */

/* ----------------------------------------------------------------------- */
/* The arguments for each of these "input" declaration parsers follow the  */
/*   same form, no matter what type of input or device:                    */
/*                                                                         */
/* devinfo -- a pointer to the overall structure that defines this device. */
/*            It is used to get access to the auxiliary data of the device.*/
/*                                                                         */
/* input -- a pointer to the vrGenericInput structure that is a specific   */
/*          FreeVR input, such as a single vr2switch, that is part of the  */
/*          specified input device.                                        */
/*          NOTE: the "input_type" can be determined by accessing the      */
/*          "input->input_type" field.  This can be used to have a single  */
/*          mapping function handle requests for multiple types of inputs. */
/*          See _XwindowsKeyboardInput() for an example of this.           */
/*                                                                         */
/* The remaining argument contains three strings that are from the portion */
/*   of the input config statement contained within the parentheses.  This */
/*   is pre-parsed as the "DTI" -- {<sub-Device>:}<Type>[<Instance>].  Any */
/*   of these subfields can be an empty string.  Usually:                  */
/*                                                                         */
/* device -- a string that may be used to indicate a sub-device of the     */
/*           provided hardware.  For example, an X11 input device may      */
/*           have "keyboard", "mouse" and "dial+buttons" subdevices.       */
/*           In most cases the "device" value is the empty string, and     */
/*           probably could always be included as part of the "type" value.*/
/*                                                                         */
/* type -- a string that indicates the class of input interface on the     */
/*         particular device/subdevice.  For example, an X11 input device  */
/*         may have "key", "mousebutton" and "pointerlocation" types.      */
/*         Or, a Magellan may have "button" and "6dof" types.              */
/*                                                                         */
/*         NOTE that the "type" value is also used in determining which of */
/*         these functions should be called (by matching the first element */
/*         of the vrInputFunctions list below), so it has already played a */
/*         roll in how this input is handled.  Therefore it's value is     */
/*         often ignored in the parsing function itself.  However, it is   */
/*         possible for more than one <type> string to point to the same   */
/*         device parsing function.  Thus in those cases it may be         */
/*         important to make use of this information within the parsing    */
/*         function.                                                       */
/*                                                                         */
/*         NOTE also that this string is different than the "input_type"   */
/*         data mentioned above as part of the "input" in that this is a   */
/*         string name that might describe the physical form of the input, */
/*         versus the "input_type" which describes a logical FreeVR input  */
/*         type such as a BINARY, VALUATOR or 6SENSOR input.               */
/*                                                                         */
/* instance -- a string usually used to indicate a specific instance of    */
/*             the type of data.  For example button "2" on a Magellan,    */
/*             unit "1" of a Flock of Birds, or key "Esc" of an X11        */
/*             keyboard.  Other data may also be included for altering how */
/*             this input is handled (See the this device for an example). */
/*                                                                         */
/*    NOTE: since the "device" subfield is often optional, and the "type"  */
/*    subfield is used to determine which function to call, many of these  */
/*    functions will only make use of the "instance" subfield.             */
/*                                                                         */
/* The return values specify whether the initial match is confirmed, or    */
/*   rejected.  For some devices (such as the Xwindows-keyboard), there    */
/*   may be an overall match for the keyboard, but then the specific key   */
/*   instance may not match a known key for the keyboard.  In such cases   */
/*   a VRINPUT_MATCH_UNABLE is returned, indicating that the search failed,*/
/*   but that the search for a matching device should cease.  If the match */
/*   is successful, then a VRINPUT_MATCH_ABLE is returned.  If the match   */
/*   altogether fails, then a VRINPUT_NOMATCH is returned, and the search  */
/*   continues.                                                            */
/*                                                                         */
/* NOTE: these functions are called before making the physical connection  */
/*   to the device, so it is not possible to query features of the device  */
/*   within these functions.                                               */
/*                                                                         */
/* NOTE: the vrInputDTI parsing (done in vr_parse.c) is for the common     */
/*   form of how inputs are specified in the FreeVR configuration (rc)     */
/*   files.  Thus the following input specification:                       */
/*       input "2switch[left]" = "2switch(Joystick0:button[left])";        */
/*   would parse the devices input as:                                     */
/*       device (d) = "Joystick0"    ("sub-device" is probably more exact) */
/*       type (t) = "button"                                               */
/*       instance (i) = "left"                                             */
/* ----------------------------------------------------------------------- */


/***************************************************************************/
/* _VrpnButtonInput(): this function takes a "device,type,instance"        */
/*    (DTI) configuration triplet for a "button" class of input, parses    */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "button(<sub-device>:<type>[<instance>, <args>])"   */
/*    NOTE: for the <Vrpn> device, there may not be sub-device options.    */
/* See the "List of possible inputs" in _VrpnCreateFunction() below        */
/*    for the codes that map to this function.                             */
/*                                                                         */
/* Here is a sample definition for a Vrpn button:                          */
/*     input "2switch[left]" = "2switch(Joystick0:button[0])";             */
static vrInputMatch _VrpnButtonInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			button_num;		/* the current count of FreeVR button inputs */
	int			button_choice;		/* the chosen number for this button */

#if 0 /* old style configuration { */
	char			vrpn_device[256];
	char			*device_end;
	int			vrpn_device_num;

	/* A comma separates the VRPN device name, and the number */
	/*   of the desired input from that device.               */
	device_end = strchr(dti->instance, ',');

	/* If no comma, then report an error in the input specification */
	/*   otherwise add this VRPN field as an input.                 */
	if (device_end != NULL) {
		/* copy just the name of the device into vrpn_device */
		strncpy(vrpn_device, dti->instance, device_end - dti->instance);
		vrpn_device[device_end - dti->instance] = '\0';

		/* the rest of the string should just be a number of the particular sensor */
		button_num = vrAtoI(dti->instance);

		/* now see if the given device name is already on */
		/*   the list of specified button devices.       */
		vrpn_device_num = vrSearch256StringList(vrpn_device, aux->num_button_devices, aux->button_names);

		/* if not, then make a VRPN connection to this button device and add to the list */
		if (vrpn_device_num < 0) {
			vrpn_device_num = aux->num_button_devices;
			vrDbgPrintfN(1 /* TODO: delete this or make INPUT_DBGLVL */, "adding new button device to list: number %d\n", vrpn_device_num);

			/* add device name to the list of VRPN button devices */
			strcpy(aux->button_names[vrpn_device_num], vrpn_device);
			aux->num_button_devices++;
			vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "aux->button_names[%d] is now '%s'\n", vrpn_device_num, aux->button_names[vrpn_device_num]);
		}

		vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "VRPN: doing button-sensor device '%s' (button %d), %d\n", vrpn_device, vrpn_device_num, button_num);

		/* now assign given input to the existing or new button device */
		aux->button_inputs[vrpn_device_num][button_num] = (vr2switch *)input;

		/* TODO: make initial button assignment */
#  if 0 /* still need to make sure the instance stuff is okay */
		vrAssign2switch(aux->button_inputs[button_num], strchr(dti->instance, ','));
#  endif
		vrDbgPrintfN(INPUT_DBGLVL, "VRPN: assigned 2switch event of value 0x%02x to input pointer = %p)\n",
			button_num, aux->button_inputs[vrpn_device_num][button_num]);

	} else {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VrpnButtonInput: Warning, invalid input specification (%s, %s).\n" NORM_TEXT,
			dti->type, dti->instance);
	}
#else /* } <-- old configuration vs. new --> { */
	/*******************/
	/* select a button */
	button_num = aux->num_buttons;
	aux->num_buttons++;
	button_choice = vrAtoI(dti->instance);

	/*****************************/
	/* check the selected button */
	if (button_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_VrpnButtonInput(): Warning, button['%s'] did not match any known button\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->button_inputs[button_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VrpnButtonInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->button_inputs[button_num] = (vr2switch *)input;
	aux->button_map_button[button_num] = button_choice;
	aux->button_map_devicename[button_num] = vrShmemStrDup(dti->device);

	vrDbgPrintfN(INPUT_DBGLVL, "_VrpnButtonInput(): assigned button event of value %d (VRPN button %d) to input pointer = %p)\n",
		button_num, button_choice, aux->button_inputs[button_num]);
#endif /* } (new configuration) */

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}


/***************************************************************************/
/* _VrpnValuatorInput(): this function takes a "device,type,instance"      */
/*    (DTI) configuration triplet for a "valuator" class of input, parses  */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "valuator(<sub-device>:<type>[<instance>, <args>])" */
/*    NOTE: for the <Vrpn> device, there may not be sub-device options.    */
/* _VrpnValuatorInput() was copied from _VruiDDValuatorInput().            */
/*                                                                         */
/* Here are sample definitions for a Vrpn valuator:                        */
/*    input "2switch[A]" = "2switch(Joystick0:button[1])";                 */
/*    input "2switch[B]" = "2switch(WiiMote0:button[2])";                  */
/* Indices are 1-based.                                                    */
static vrInputMatch _VrpnValuatorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			valuator_num;		/* the current count of FreeVR valuator inputs */
	int			valuator_choice;	/* the chosen number for this valuator */

	/* The Vrpn method of handling input arrays was taken from how the */
	/*   VruiDD handler works.                                         */

	/********************************/
	/* select a valuator (and sign) */
	valuator_num = aux->num_valuators;
	aux->num_valuators++;
	if (dti->instance[0] == '-') {
		aux->valuator_map_sign[valuator_num] = -1.0;
		valuator_choice = vrAtoI(&(dti->instance[1]));	/* skip the negation sign */
	} else {
		aux->valuator_map_sign[valuator_num] =  1.0;
		valuator_choice = vrAtoI(dti->instance);
	}

	/*******************************/
	/* check the selected valuator */
	if (valuator_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_VrpnValuatorInput(): Warning, valuator['%s'] did not match any known valuator\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->valuator_inputs[valuator_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VrpnValuatorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->valuator_inputs[valuator_num] = (vrValuator *)input;
	aux->valuator_map_valuator[valuator_num] = valuator_choice;
	aux->valuator_map_devicename[valuator_num] = vrShmemStrDup(dti->device);

	vrDbgPrintfN(INPUT_DBGLVL, "_VrpnValuatorInput(): assigned valuator event of value %d (Vrpn valuator %d) to input pointer = %p)\n",
		valuator_num, valuator_choice, aux->valuator_inputs[valuator_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}
#if 0 /* { future delete me */
/**************************************************************************/
/* NOTE: _VrpnAnalogInput() is old and deprecated */
static int _VrpnAnalogInput(vrInputDevice *devinfo, vrGenericInput *input, char *device, char *type, char *instance)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			valuator_num;
	char			vrpn_device[256];
	char			*device_end;
	int			vrpn_device_num;

#  if 0 /* old style configuration { */
	/* A comma separates the VRPN device name, and the number */
	/*   of the desired input from that device.               */
	device_end = strchr(instance, ',');

	/* If no comma, then report an error in the input specification */
	/*   otherwise add this VRPN field as an input.                 */
	if (device_end != NULL) {
		/* copy just the name of the device into vrpn_device */
		strncpy(vrpn_device, instance, device_end - instance);
		vrpn_device[device_end - instance] = '\0';

		/* the rest of the string should just be a number of the particular sensor */
		valuator_num = vrAtoI(device_end+1);

		/* now see if the given device name is already on */
		/*   the list of specified analog devices.       */
		vrpn_device_num = vrSearch256StringList(vrpn_device, aux->num_analog_devices, aux->analog_names);

		/* if not, then make a VRPN connection to this analog device and add to the list */
		if (vrpn_device_num < 0) {
			vrpn_device_num = aux->num_analog_devices;
			vrDbgPrintfN(1 /* TODO: delete this or make INPUT_DBGLVL */, "adding new analog device to list: number %d\n", vrpn_device_num);

			/* add device name to the list of VRPN analog devices */
			strcpy(aux->analog_names[vrpn_device_num], vrpn_device);
			aux->num_analog_devices++;
			vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "aux->analog_names[%d] is now '%s'\n", vrpn_device_num, aux->analog_names[vrpn_device_num]);

		}

		vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "doing valuator device '%s', %d\n", vrpn_device, valuator_num);

		/* now assign given input to the existing or new analog device */
		aux->valuator_inputs[vrpn_device_num][valuator_num] = (vrValuator *)input;

		/* TODO: make initial valuator assignment */
#  if 0 /* still need to make sure the instance stuff is okay */
		vrAssignValuator(aux->valuator_inputs[valuator_num], strchr(instance, ','));
#  endif
		vrDbgPrintfN(INPUT_DBGLVL, "assigned Valuator event of value 0x%02x to input pointer = %p)\n",
			valuator_num, aux->valuator_inputs[vrpn_device_num][valuator_num]);

	} else {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VrpnAnalogInput: Warning, invalid input specification (%s, %s).\n" NORM_TEXT,
			type, instance);
	}
#endif /* } old style configuration */

	return 1;	/* input declaration match */
}
#endif /* } */


/***************************************************************************/
/* _Vrpn6sensorInput(): this function takes a "device,type,instance"       */
/*    (DTI) configuration triplet for a "6sensor" class of input, parses   */
/*    the "DTI" triplet and maps it to a FreeVR input system virtual input */
/*    in the syntax of "6sensor(<sub-device>:<type>[<instance>, <args>])"  */
/*    NOTE: for the <Vrpn> device, there may not be sub-device options.    */
/*                                                                         */
/* Here are sample definitions for a Vrpn position sensor:                 */
/*    input "head" = "6sensor(Tracker0:tracker[0])";                       */
/*    input "sim-wand" = "6sensor(sim6[1])";                               */
/*                                                                         */
/* NOTE: this is an exact copy of the VruiDD version (except the name).    */
/*    (Except apparently I then changed "sensor6" to "tracker".)           */
static vrInputMatch _Vrpn6sensorInput(vrInputDevice *devinfo, vrGenericInput *input, vrInputDTI *dti)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	vrMatrix		mat_store;		/* temporary matrix storage */
	vr6sensor		*sensor6_input;		/* pointer to this new 6-sensor input */
	int			tracker_num;		/* the current count of FreeVR tracker inputs */
	int			tracker_choice;		/* the chosen number for this tracker */

	/* The Vrpn method of handling input arrays may be a combination of the way */
	/*   suggested by vr_input.skeleton.c as the FoB, MS & Fastrak input drivers. */

	/********************/
	/* select a 6sensor */
	tracker_num = aux->num_6sensors;
	aux->num_6sensors++;
	tracker_choice = vrAtoI(dti->instance);

	/******************************/
	/* check the selected 6sensor */
	if (tracker_choice == -1) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_Vrpn6sensorInput(): Warning, tracker['%s'] did not match any known tracker\n", dti->instance);
		return VRINPUT_MATCH_UNABLE;	/* device match, but still unable to handle request */
	} else if (aux->sensor6_inputs[tracker_num] != NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_Vrpn6sensorInput(): Warning, new input from %s[%s] overwriting old value.\n" NORM_TEXT,
			dti->type, dti->instance);
	}

	/********************/
	/* make the mapping */
	aux->sensor6_inputs[tracker_num] = (vr6sensor *)input;
	aux->tracker_map_tracker[tracker_num] = tracker_choice;
	sensor6_input = aux->sensor6_inputs[tracker_num];

	if (!strcasecmp(dti->type, "sim6")) {
		/* force this to use the simulator device if the type implies it */
		aux->tracker_map_devicename[tracker_num] = vrShmemStrDup("simulator");
	} else {
		aux->tracker_map_devicename[tracker_num] = vrShmemStrDup(dti->device);
	}

	/* vrAssign6sensorR2Exform() parses the rest of the instance setting */
	vrAssign6sensorR2Exform(sensor6_input, strchr(dti->instance, ','));

	/* Set the initial values */
	vrMatrixSetIdentity(&mat_store);
	vrMatrixPostMult(&mat_store, sensor6_input->ipos_xform);
	vrAssign6sensorValue(sensor6_input, &mat_store, -1 /* , vrCurrentWallTime() */);
	vrAssign6sensorActiveValue(sensor6_input, 0);
	vrAssign6sensorErrorValue(sensor6_input, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "_Vrpn6sensorInput(): assigned tracker event of value %d (Vrpn tracker %d) to input pointer = %p)\n",
		tracker_num, tracker_choice, aux->sensor6_inputs[tracker_num]);

	return VRINPUT_MATCH_ABLE;	/* input declaration match */
}

#if 0 /* old style configuration (deprecated?) { */ /* future delete me */
/**************************************************************************/
static int _VrpnTrackerInput(vrInputDevice *devinfo, vrGenericInput *input, char *device, char *type, char *instance)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			sensor_num;
	char			vrpn_device[256];
	char			*device_end;
	int			vrpn_device_num;

#  if 0 /* old style configuration { */
	/* A comma separates the VRPN device name, and the number */
	/*   of the desired input from that device.               */
	device_end = strchr(instance, ',');

	/* If no comma, then report an error in the input specification */
	/*   otherwise add this VRPN field as an input.                 */
	if (device_end != NULL) {
		/* copy just the name of the device into vrpn_device */
		strncpy(vrpn_device, instance, device_end - instance);
		vrpn_device[device_end - instance] = '\0';

		/* the rest of the string should just be a number of the particular sensor */
		sensor_num = vrAtoI(device_end+1);

		/* now see if the given device name is already on */
		/*   the list of specified tracker devices.       */
		vrpn_device_num = vrSearch256StringList(vrpn_device, aux->num_tracker_devices, aux->tracker_names);

		/* if not, then make a VRPN connection to this tracker device and add to the list */
		if (vrpn_device_num < 0) {
			vrpn_device_num = aux->num_tracker_devices;
			vrDbgPrintfN(1 /* TODO: delete this or make INPUT_DBGLVL */, "adding new tracker device to list: number %d\n", vrpn_device_num);

			/* add device name to the list of VRPN tracker devices */
			strcpy(aux->tracker_names[vrpn_device_num], vrpn_device);
			aux->num_tracker_devices++;
			vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "aux->tracker_names[%d] is now '%s'\n", vrpn_device_num, aux->tracker_names[vrpn_device_num]);

		}

		vrDbgPrintfN(1 /* TODO: make INPUT_DBGLVL */, "doing tracker-sensor device '%s' (tracker %d), %d\n", vrpn_device, vrpn_device_num, sensor_num);

		/* now assign given input to the existing or new tracker device */
		aux->sensor6_inputs[vrpn_device_num][sensor_num] = (vr6sensor *)input;

		/* the vrAssign6sensorR2Exform() function does the parsing of the next token */
		vrAssign6sensorR2Exform((vr6sensor *)input, strchr(device_end+1, ','));

		vrDbgPrintfN(INPUT_DBGLVL, "assigned 6sensor event of value 0x%02x to input pointer = %p)\n",
			sensor_num, aux->sensor6_inputs[vrpn_device_num][sensor_num]);

	} else {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "_VrpnTrackerInput: Warning, invalid input specification (%s, %s).\n" NORM_TEXT,
			type, instance);
	}
#  endif /* } */

	return 1;	/* input declaration match */
}
#endif /* } */


	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*      Callbacks for controlling the VRPN features.        */
	/*                                                          */
	/*  These _Vrpn<operation>Callback() functions are          */
	/*  called during the CREATE phase of the input interface.  */

/*************************************************************************/
/* This collection of callbacks (for controlling the device features) is */
/*   designed to allow the FreeVR system configuration to map physical   */
/*   inputs of this device to control the operation of the VR system     */
/*   as opposed to the norm, where inputs are used to affect the         */
/*   simulated virtual world.                                            */
/*                                                                       */
/* Each command callback takes exactly two arguments:                    */
/*   * devinfo -- this is information about this particular instance of  */
/*                  connecting with an input device or system.  This     */
/*                  provides:                                            */
/*                  1) the ability to alter the settings of the device   */
/*                     itself using the "aux" data field (for example    */
/*                     the SpaceTec devices can do a "reset" with        */
/*                     _SpacetecSensorResetCallback(), or the X11 input  */
/*                     can loop through all the simulated sensors with   */
/*                     _XwindowsSensorNextCallback();                    */
/*               or 2) these callbacks can be used to affect the overall */
/*                     VR system, such as toggling the "paused" flag.    */
/*   * value -- an integer flag that can affect the callback.            */
/*       NOTE: this assumes that the callback is from a button input, or */
/*       from a valuator passing a threshold.  For buttons, the value    */
/*       will be 1 (one) when the button is depressed.                   */
/*************************************************************************/

#if 0 /* delete me */
/***************************************************************************/
/* _VrpnSystemPauseToggleCallback(): this callback toggles the "paused"    */
/*   field of the overall FreeVR "context" structure.  When the system is  */
/*   "paused", rendering still functions, but the inputs are frozen.  This */
/*   can be used for debugging, or for still photography of a CAVE.        */
static void _VrpnSystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(SELFCTRL_DBGLVL, "Vrpn Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/*****************************************************************************/
/* _VrpnPrintContextStructCallback(): this callback generates output of      */
/*   the overall system "context" data.  It is generally used when debugging */
/*   the library or the configuration.                                       */
static void _VrpnPrintContextStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintContext(stdout, devinfo->context, verbose);		/* NOTE: this effectively requires the inclusion of "vr_system.h" */
}

/************************************************************************/
/* _VrpnPrintConfigStructCallback(): this callback outputs the entire   */
/*    VR system configuration to the text console.                      */
static void _VrpnPrintConfigStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintConfig(stdout, devinfo->context->config, verbose);	/* NOTE: this effectively requires the inclusion of "vr_config.h" */
}

/************************************************************************/
/* _VrpnPrintInputStructCallback(): this callback outputs the current   */
/*    state of the entire input structure of the running VR system.     */
static void _VrpnPrintInputStructCallback(vrInputDevice *devinfo, int value)
{
	if (value == 0)
		return;

	vrFprintInput(stdout, devinfo->context->input, verbose);
}
#endif

/***************************************************************************/
/* _VrpnPrintPrivateInfoCallback(): this callback outputs the contents     */
/*    of the device specific internal structure to help with configuration */
/*    and system debugging.                                                */
static void _VrpnPrintPrivateInfoCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrintPrivateInfo(stdout, aux, verbose);
}

/***********************************************************************/
/* _VrpnPrintHelpCallback(): this callback outputs the full system's */
/*     help information string.                                        */
static void _VrpnPrintHelpCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrintHelp(stdout, aux);
}

/************************************************************************/
/* _VrpnPrint6sensorOptionsCallback(): the callback outputs the options */
/*   used to control the simulated 6sensor virtual input.               */
static void _VrpnPrint6sensorOptionsCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo  *aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	_VrpnPrint6sensorOptions(stdout, aux, verbose);
}

/**************************************************************************/
/* _VrpnSensorNextCallback(): this callback affects the simulated 6sensor */
/*    controls by changing which of the available 6sensors to manipulated */
/*    from the current one to the next one in the (linked) list.          */
static void _VrpnSensorNextCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == 0)
		return;

	if (devinfo->num_6sensors == 0) {
		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: next sensor -- no sensors available.\n",
			aux->active_sim6sensor);
		return;
	}
#if 0	/* delete me */
vrPrintf("sim6 mapped? %d %d %d %d %d\n", aux->device[0].mapped_tracker[0], aux->device[0].mapped_tracker[1], aux->device[0].mapped_tracker[2], aux->device[0].mapped_tracker[3], aux->device[0].mapped_tracker[4]);
#endif

	/* set the current sensor as non-active */
	if (aux->active_sim6sensor >= aux->num_6sensors) {
		/* there is no active sensor */
		vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnSensorNextCallback(): " RED_TEXT "Warning, attempting to change simulated 6-sensor, but none available.\n" NORM_TEXT);
	} else {
		vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sim6sensor]), 0);
	}

	/* search from the next possible sensor to the end of the list */
	/* NOTE: VRPN, unlike other devices, has a select set of trackers that are */
	/*   part of the simulation group, so only search those.  (which in practice */
	/*   means keep searching the list if a near-match isn't from the simulation */
	/*   group.)                                                                 */
	do {
		aux->active_sim6sensor++;
#if 0	/* delete me */
vrPrintf("@@@@@@ do-while: active_sim6sensor=%d, num_6=%d, aux->sensor6_inputs[%d]=%p, aux->tracker_map_device[aux->active_sim6sensor]->name='%s'\n", aux->active_sim6sensor, aux->num_6sensors, aux->active_sim6sensor, aux->sensor6_inputs[aux->active_sim6sensor], (aux->active_sim6sensor < aux->num_6sensors ? aux->tracker_map_device[aux->active_sim6sensor]->name : "<void>"));
#endif
												/* keep looping if: */
	} while (	   (aux->active_sim6sensor < aux->num_6sensors)				/*   we haven't gone off the end of the list */
			&& (aux->sensor6_inputs[aux->active_sim6sensor] != NULL)			/*   tracker does not exist! */
			&& (aux->tracker_map_device[aux->active_sim6sensor] != &(aux->device[0])));	/*   tracker is not from device[0] (ie. simulator) */

	/* if none found (ie. we were basically already at the end of the list), then search from the beginning */
#if 0	/* delete me */
vrPrintf("@@@@@@ pre-for: active_sim6sensor=%d, num_6=%d, aux->sensor6_inputs[%d]=%p, aux->tracker_map_device[aux->active_sim6sensor]->name='%s'\n", aux->active_sim6sensor, aux->num_6sensors, aux->active_sim6sensor, aux->sensor6_inputs[aux->active_sim6sensor], (aux->active_sim6sensor < aux->num_6sensors ? aux->tracker_map_device[aux->active_sim6sensor]->name : "<void>"));
#endif
	if (aux->active_sim6sensor >= aux->num_6sensors
			|| (aux->sensor6_inputs[aux->active_sim6sensor] == NULL)
			|| (aux->tracker_map_device[aux->active_sim6sensor] != &(aux->device[0]))) {
		for (aux->active_sim6sensor = 0
			; /* conditional */							/* keep looping if: */
			   (aux->active_sim6sensor < aux->num_6sensors)				/*   still within the list bounds */
			&& (aux->sensor6_inputs[aux->active_sim6sensor] != NULL)			/*   tracker does not exist! */
			&& (aux->tracker_map_device[aux->active_sim6sensor] != &(aux->device[0]))	/*   tracker is not from device[0] (ie. simulator) */
			; /* increment */
			aux->active_sim6sensor++)
#if 0	/* delete me */
vrPrintf("@@@@@@ for: active_sim6sensor=%d, num_6=%d, aux->sensor6_inputs[%d]=%p, aux->tracker_map_device[aux->active_sim6sensor]->name='%s'\n", aux->active_sim6sensor, aux->num_6sensors, aux->active_sim6sensor, aux->sensor6_inputs[aux->active_sim6sensor], aux->tracker_map_device[aux->active_sim6sensor]->name)
#endif
			; /* no-op -- the looping is the reward */
	}
#if 0	/* delete me */
vrPrintf(BOLD_TEXT "New active sensor is %d\n" NORM_TEXT, aux->active_sim6sensor);
#endif

	/* set the newly found sensor as active */
	if (aux->active_sim6sensor >= aux->num_6sensors) {
		/* again, no active sensor */
		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: No simulated 6-sensor available to make active.\n", aux->active_sim6sensor);
	} else {
		vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sim6sensor]), 1);
		aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */

		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: 6sensor[%d] now active.\n", aux->active_sim6sensor);
	}
}

/* TODO: see if there is a way to call this as an N-switch */
/*****************************************************************/
/* _VrpnSensorSetCallback(): this callback selects a particular  */
/*    simulated 6sensor from the enumerated available list given */
/*    by the value of the "value" argument.                      */
static void _VrpnSensorSetCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	if (value == aux->active_sim6sensor)
		return;

	if (value < 0 || value >= MAX_6SENSORS) {
		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set sensor (%d) -- out of range.\n", value);
	}

	if (aux->sensor6_inputs[value] == NULL) {
		vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set sensor (%d) -- no such sensor available.\n", value);
		return;
	}

	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sim6sensor]), 0);
	aux->active_sim6sensor = value;
	vrAssign6sensorActiveValue(((vr6sensor *)aux->sensor6_inputs[aux->active_sim6sensor]), 1);

	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: 6sensor[%d] now active.\n", aux->active_sim6sensor);
}

/********************************************************************/
/* _VrpnSensorSet<N>Callback(): the following 10 callbacks select   */
/*   the Nth simulated 6sensor from the list of available simulated */
/*   6sensors.  (Nothing happens is there is no such 6sensor.)  The */
/*   value of "value" is ignored.                                   */
static void _VrpnSensorSet0Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 0); }

static void _VrpnSensorSet1Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 1); }

static void _VrpnSensorSet2Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 2); }

static void _VrpnSensorSet3Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 3); }

static void _VrpnSensorSet4Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 4); }

static void _VrpnSensorSet5Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 5); }

static void _VrpnSensorSet6Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 6); }

static void _VrpnSensorSet7Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 7); }

static void _VrpnSensorSet8Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 8); }

static void _VrpnSensorSet9Callback(vrInputDevice *devinfo, int value)
{	_VrpnSensorSetCallback(devinfo, 9); }

/******************************************************************/
/* _VrpnSensorResetCallback(): this callback resets the location  */
/*   and orientation of the currently active simulated 6sensor    */
/*   to the original (identity) position.                         */
static void _VrpnSensorResetCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;

	if (value == 0)
		return;

	sensor = &(devinfo->sensor6[aux->active_sim6sensor]);
	vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
	vrAssign6sensorActiveValue(sensor, -1);
	vrAssign6sensorErrorValue(sensor, 0);
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: reset 6sensor[%d].\n", aux->active_sim6sensor);
}

/*********************************************************************/
/* _VrpnSensorResetAllCallback(): this callback resets the location  */
/*   and orientation of all the available simulated 6sensors to their*/
/*   original (identity) position.                                   */
static void _VrpnSensorResetAllCallback(vrInputDevice *devinfo, int value)
{
static	vrMatrix		idmat = { { 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0 } };
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	vr6sensor		*sensor;
	int			count;

	if (value == 0)
		return;

	for (count = 0; count < MAX_6SENSORS; count++) {
		if (aux->sensor6_inputs[count] != NULL) {
			sensor = &(devinfo->sensor6[count]);
			vrAssign6sensorValue(sensor, &idmat, 0 /* , timestamp */);
			vrAssign6sensorActiveValue(sensor, (count == aux->active_sim6sensor));
			vrAssign6sensorErrorValue(sensor, 0);
			vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: reset 6sensor[%d].\n", count);
		}
	}
}

/****************************************************************/
/* _VrpnTempValuatorOnlyCallback(): this callback sets the      */
/*   "ignore_all" value of the simulated 6sensor system, the    */
/*   effect of which is to both temporarily shift the input     */
/*   (perhaps a 6-DOF input device) from any manipulation of    */
/*   the active simulated 6sensor to instead manipulate the     */
/*   virtual joystick.                                          */
/* In this instance, the "ignore_all" is set to the value of    */
/*   the "value" argument, which has the effect of making the   */
/*   alteration only while the key/button is being held down,   */
/*   returning to normal when it is released.                   */
static void _VrpnTempValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_all = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/******************************************************************/
/* _VrpnToggleValuatorOnlyCallback(): this callback toggles       */
/*   the "ignore_all" value of the simulated 6sensor system,      */
/*   the effect of which is to "permanently" shift the input      */
/*   (perhaps a 6-DOF input device) from any manipulation of the  */
/*   active simulated 6sensor to instead manipulate the virtual   */
/*   joystick.                                                    */
/* In this instance, the "ignore_all" is toggled when the value   */
/*   of the "value" argument is one (i.e the key/button is first  */
/*   pressed.  There is no effect when "value" is zero.           */
static void _VrpnToggleValuatorOnlyCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_all ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: ignore_all = %d.\n",
		aux->sensor6_options.ignore_all);
}

/****************************************************************/
/* _VrpnTempValuatorOverrideCallback(): this callback sets the  */
/*   "ignore_trans" value of the simulated 6sensor system, the  */
/*   effect of which is to temporarily shift the input (usually */
/*   a physical joystick) from translating the active simulated */
/*   6sensor to instead manipulate the virtual joystick.        */
/* In this instance, the "ignore_trans" is set to the value of  */
/*   the "value" argument, which has the effect of making the   */
/*   shift only while the key/button is being held down,        */
/*   returning to normal when it is released.                   */
static void _VrpnTempValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* set the field to the current state of value (ie. 1 when depressed, 0 when released) */
	aux->sensor6_options.ignore_trans = value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/******************************************************************/
/* _VrpnToggleValuatorOverrideCallback(): this callback toggles   */
/*   the "ignore_trans" value of the simulated 6sensor system,    */
/*   the effect of which is to "permanently" shift the input      */
/*   (usually a physical joystick) from translating the active    */
/*   simulated 6sensor to instead manipulate the virtual joystick.*/
/* In this instance, the "ignore_trans" is toggled when the value */
/*   of the "value" argument is one (i.e the key/button is first  */
/*   pressed.  There is no effect when "value" is zero.           */
static void _VrpnToggleValuatorOverrideCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.ignore_trans ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: ignore_trans = %d.\n",
		aux->sensor6_options.ignore_trans);
}

/**********************************************************************/
/* _DEVICEToggleRestrictSpaceCallback(): this callback toggles the    */
/*   "restrict_space" value of the simulated 6sensor system, the      */
/*   effect of which is to limit the range of translational movement  */
/*   of the 6sensor to a specified working volume.  This is typically */
/*   used to enforce the limits that a physical CAVE or workspace     */
/*   would cause, enabling the user to test whether a user interface  */
/*   element would be in reach of a user in a CAVE.                   */
/*                                                                    */
/* TODO: this should probably also go through all the 6sensors        */
/*   and move them to be within the allowed workspace when space      */
/*   restriction is turned on.                                        */
static void _VrpnToggleRestrictSpaceCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.restrict_space ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: restrict_space = %d.\n",
		aux->sensor6_options.restrict_space);
}

/**********************************************************************/
/* _VrpnToggleReturnToZeroCallback(): this callback toggles the       */
/*   "return_to_zero" value of the simulated 6sensor system, the      */
/*   effect of which sets the current translational and rotational    */
/*   movements of the 6sensor to exactly the (possibly scaled) values */
/*   of the physical input device versus accumulating movements and   */
/*   rotations allowing the user to "fly" the simulated input around. */
static void _VrpnToggleReturnToZeroCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.return_to_zero ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: return_to_zero = %d.\n",
		aux->sensor6_options.return_to_zero);
}

/********************************************************************/
/* _VrpnToggleRelativeAxisCallback(): this callback toggles the     */
/*   "relative_axis" value of the simulated 6sensor system, the     */
/*   effect of which is to change the coordinate system that is     */
/*   used when moving and rotating.  When set to "1", movement is   */
/*   around the 6sensor's own coordinate system (aka "local CS");   */
/*   and when set to "0", movement is based on the world coordinate */
/*   system (aka "global CS").                                      */
static void _VrpnToggleRelativeAxisCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.relative_axis ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: relative_axis = %d.\n",
		aux->sensor6_options.relative_axis);
}

/*************************************************************/
/* _VrpnToggleSwapYZCallback(): this callback toggles the    */
/*   "swap_yz" value of the simulated 6sensor system, the    */
/*   effect of which is to change between Y-UP and Z-UP      */
/*   coordinate systems -- something that is often required  */
/*   when using different position trackers and/or graphics  */
/*   outputs.  It can also be used simply as a means of      */
/*   manipulating a simulated 6sensor in 3 dimensions by     */
/*   changing which pair of dimensions are being controlled  */
/*   at any given time.                                      */
static void _VrpnToggleSwapYZCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.swap_yz ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: swap_yz = %d.\n",
		aux->sensor6_options.swap_yz);
}

/*********************************************************************/
/* _VrpnToggleSwapTransRotCallback(): this callback toggles the      */
/*   "swap_transrot" value of the simulated 6sensor system, the      */
/*   effect of which is to change the values feeding the translation */
/*   with those feeding the rotation.  The purpose of this is to be  */
/*   able to use a single joystick to control both, and then at the  */
/*   press of a button toggle which of the two is being controlled.  */
/* NOTE: in practice, there are some sign-changes required to make   */
/*   this work as the user would expect.  Thus these sign-changes    */
/*   are hard-coded into the vrAssign6sensorValueFromValuators()     */
/*   function which evaluates the simulated 6sensor inputs.          */
/*   The swapping is between these three value pairs:                */
/*      * VR_X & -VR_AZIM                                            */
/*      * VR_Z &  VR_ELEV                                            */
/*      * VR_Y & -VR_ROLL                                            */
static void _VrpnToggleSwapTransRotCallback(vrInputDevice *devinfo, int value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* take no action when the button is released */
	if (value == 0)
		return;

	/* toggle the field (NOTE: we only get here when button is depressed) */
	aux->sensor6_options.swap_transrot ^= 1;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: swap_transrot = %d.\n",
		aux->sensor6_options.swap_transrot);
}

/************************************************************/
/* _VrpnSetTransXCallback(): TODO: describe this callback */
static void _VrpnSetTransXCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_X] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set trans_x = %f.\n", *value);
}

/************************************************************/
/* _VrpnSetTransYCallback(): TODO: describe this callback */
static void _VrpnSetTransYCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	int			simulated_tracker_num;

	aux->sim_values[VR_Y] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	simulated_tracker_num = aux->tracker_map_tracker[aux->active_sim6sensor];
	aux->device[0].newly_reported_posquat[simulated_tracker_num] = 1u;				/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set trans_y = %f.\n", *value);
}

/************************************************************/
/* _VrpnSetTransZCallback(): TODO: describe this callback */
static void _VrpnSetTransZCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_Z] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set trans_z = %f.\n", *value);
}

/************************************************************/
/* _VrpnSetAzimuthCallback(): TODO: describe this callback */
static void _VrpnSetAzimuthCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_AZIM+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set azimuth = %f.\n", *value);
}

/************************************************************/
/* _VrpnSetElevationCallback(): TODO: describe this callback */
static void _VrpnSetElevationCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ELEV+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "VRPN Control: set elevation = %f.\n", *value);
}

/************************************************************/
/* _VrpnSetRollCallback(): TODO: describe this callback */
static void _VrpnSetRollCallback(vrInputDevice *devinfo, float *value)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	aux->sim_values[VR_ROLL+3] = *value;
	aux->sim6sensor_change = 1;		/* indicate new data for simulated 6-sensor */
	aux->device[0].newly_reported_posquat[aux->tracker_map_tracker[aux->active_sim6sensor]] = 1u;	/* VRPN new data flag */
	vrDbgPrintfN(SELFCTRL_DBGLVL, "Vrpn Control: set roll = %f.\n", *value);
}


	/*************************************************/
	/*   Callbacks for interfacing with the device.  */
	/*                                               */

/**********************************************************************/
/* This collection of callbacks (for interfacing with the device) are */
/*   the standard ways in which the FreeVR system creates an instance */
/*   of access to a particular device.  It consists of the six (6)    */
/*   basic operational functions:                                     */
/*     1) Creating the device instance                                */
/*     2) Open a communications channel with the device               */
/*     3) Close the communications channel with the device            */
/*     4) Reset the device                                            */
/*     5) Poll input data from the device                             */
/* and 6) Print information about the state of the device             */
/**********************************************************************/


/***********************************************************************/
/* _VrpnCreateFunction(): this callback maps the configuration input   */
/*    options and control options to the above callback routines that  */
/*    will then execute the code to handle each type of input/control. */
/*                                                                     */
/* There are N primary tasks of this function:                         */
/*     1) Allocate all the memory for the device instance data.        */
/*     2) Parse the options for operating this device instance.        */
/*     3) Allocate the memory for all the inputs and self-controls.    */
/*     4) Actually map the callbacks for all inputs and self-controls. */
/* and 5) Set the flag that the device is configured and ready to go.  */
static void _VrpnCreateFunction(vrInputDevice *devinfo)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_VrpnInputs[] = {
				{ "button", VRINPUT_2WAY, _VrpnButtonInput },
				{ "analog", VRINPUT_VALUATOR, _VrpnValuatorInput },
				{ "tracker", VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ "sim6", VRINPUT_6SENSOR, _Vrpn6sensorInput },
				{ NULL, VRINPUT_UNKNOWN, NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_VrpnControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", vrInputSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", vrInputPrintContextStructCallback },
				{ "print_config", vrInputPrintConfigStructCallback },
				{ "print_input", vrInputPrintInputStructCallback },
				{ "print_devinfo", _VrpnPrintPrivateInfoCallback },
				{ "print_sim6opts", _VrpnPrint6sensorOptionsCallback },
				{ "print_help", _VrpnPrintHelpCallback },

				/* simulated 6-sensor selection controls */
				{ "sensor_next", _VrpnSensorNextCallback },
				{ "setsensor", _VrpnSensorSetCallback },	/* NOTE: this is non-boolean */
				{ "setsensor(0)", _VrpnSensorSet0Callback },
				{ "setsensor(1)", _VrpnSensorSet1Callback },
				{ "setsensor(2)", _VrpnSensorSet2Callback },
				{ "setsensor(3)", _VrpnSensorSet3Callback },
				{ "setsensor(4)", _VrpnSensorSet4Callback },
				{ "setsensor(5)", _VrpnSensorSet5Callback },
				{ "setsensor(6)", _VrpnSensorSet6Callback },
				{ "setsensor(7)", _VrpnSensorSet7Callback },
				{ "setsensor(8)", _VrpnSensorSet8Callback },
				{ "setsensor(9)", _VrpnSensorSet9Callback },
				{ "sensor_reset", _VrpnSensorResetCallback },
				{ "sensor_resetall", _VrpnSensorResetAllCallback },

				/* simulated 6-sensor manipulation controls */
				{ "temp_valuator", _VrpnTempValuatorOverrideCallback },
				{ "toggle_valuator", _VrpnToggleValuatorOverrideCallback },
				{ "temp_valuator_only", _VrpnTempValuatorOnlyCallback },
				{ "toggle_valuator_only", _VrpnToggleValuatorOnlyCallback },
				{ "toggle_relative", _VrpnToggleRelativeAxisCallback },
				{ "toggle_space_limit", _VrpnToggleRestrictSpaceCallback },
				{ "toggle_return_to_zero", _VrpnToggleReturnToZeroCallback },
				{ "toggle_swap_transrot", _VrpnToggleSwapTransRotCallback },
				{ "toggle_swap_yz", _VrpnToggleSwapYZCallback },
				{ "set_transx", _VrpnSetTransXCallback },
				{ "set_transy", _VrpnSetTransYCallback },
				{ "set_transz", _VrpnSetTransZCallback },
				{ "set_azim", _VrpnSetAzimuthCallback },
				{ "set_elev", _VrpnSetElevationCallback },
				{ "set_roll", _VrpnSetRollCallback },

		/** TODO: other callback control functions go here **/
				/* other controls */

				/* end of the list */
				{ NULL, NULL } };

	_VrpnPrivateInfo	*aux = NULL;

	/******************************************/
	/* allocate and initialize auxiliary data */
	devinfo->aux_data = (void *)vrShmemAlloc0(sizeof(*aux));
	aux = (_VrpnPrivateInfo *)devinfo->aux_data;
	_VrpnInitializeStruct(aux, devinfo->type);

	/******************/
	/* handle options */
	aux->vrpn_host = vrShmemStrDup(DEFAULT_HOST);	/* initialize to the default host value */
	aux->vrpn_port = DEFAULT_PORT;			/* initialize to the default port value */
	_VrpnParseArgs(aux, devinfo->args);


	/***************************************/
	/* create the inputs and self-controls */
	vrInputCountDataContainers(devinfo);

	/* Allocate memory for specific input types */
	/*   NOTE: something to be aware of -- when a 0 byte allocation is requested, vrShmemAlloc0() will return NULL */
	aux->button_inputs = (vr2switch **)vrShmemAlloc((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(vr2switch *));
	aux->button_map_button = (int *)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(int));
	aux->button_map_devicename = (char **)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(char *));
	aux->button_map_device = (VRPNdevice **)vrShmemAlloc0((devinfo->num_2ways + devinfo->num_scontrols) * sizeof(VRPNdevice *));

	aux->valuator_inputs = (vrValuator **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(vrValuator *));
	aux->valuator_map_valuator = (int *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(int));
	aux->valuator_map_sign = (float *)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(float));
	aux->valuator_map_devicename = (char **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(char *));
	aux->valuator_map_device = (VRPNdevice **)vrShmemAlloc0((devinfo->num_valuators + devinfo->num_scontrols) * sizeof(VRPNdevice *));

	aux->sensor6_inputs = (vr6sensor **)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(vr6sensor *));
	aux->tracker_map_tracker = (int *)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(int));
	aux->tracker_map_devicename = (char **)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(char *));
	aux->tracker_map_device = (VRPNdevice **)vrShmemAlloc0((devinfo->num_6sensors) * sizeof(VRPNdevice *));

	/* Create the inputs within the FreeVR system */
	/* These two functions do the real work of this function, by creating */
	/*   the mappings as they are laid out in the configuration file.     */
	vrInputCreateDataContainers(devinfo, _VrpnInputs);
	vrInputCreateSelfControlContainers(devinfo, _VrpnInputs, _VrpnControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	vrDbgPrintf("_VrpnCreateFunction(): Done creating VRPN inputs for '%s'\n", devinfo->name);
	devinfo->created = 1;

	return;
}

#endif /* } FREEVR */

/***************************************************************************/
/* _VrpnOpenFunction(): this callback opens the communications channel     */
/*    with the physical device or data streaming utility.  This is         */
/*    generally unique for each type of device or stream.                  */
/*                                                                         */
/* There are no return values, but there are two input-device flags that   */
/*    are set to indicate the state of the device:                         */
/*      * open -- indicates when the communications channel is working     */
/*      * operating -- indicates the device is actually ready to send data */
static void _VrpnOpenFunction(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;
static	char	*version = "vrpn: ver. 07.28  0\0\0\0\0\0\0\0\0";	/* Using the current version number */
	int	read_result;			/* return value of the call to "read()" */
	int	rval;				/* return value */

	vrTrace("_VrpnOpenFunction", devinfo->name);

#ifdef VRPN_LIB /* { */
	/* TODO: consider putting a check for VRPN functions here (ie. is the DSO loaded). */
#endif /* } VRPN_LIB */

	/*******************/
	/* open the device */

	/* open the "incoming" port (ie. port connected to the device) */
	aux->vrpn_socket = vrSocketCall(aux->vrpn_host, aux->vrpn_port);
	if (aux->vrpn_socket < 0) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "_VrpnOpenFunction(): " RED_TEXT "Error, unable to open connection to socket '%s:%d'!\n" NORM_TEXT, aux->vrpn_host, aux->vrpn_port);
		return;		/* return w/o setting the open flag */
	}

	if (fcntl(aux->vrpn_socket, F_SETFL, O_RDWR | O_NONBLOCK | O_NOCTTY) < 0) {
		vrDbgPrintfN(AALWAYS_DBGLVL, "_VrpnOpenFunction(): " RED_TEXT "An error occurred while trying to set the file controls.\n" NORM_TEXT);
		return;		/* return w/o setting the open flag */
	}

//	aux->open = 1;		/* communication line has been established */

	/* establish connection with the protocol */
	rval = write(aux->vrpn_socket, version, 24);	/* NOTE: not using vrSocketSendMsg() because the length is fixed, including several NULL terminator characters */
	if (rval != 24) {
		vrDbgPrintfN(AALWAYS_DBGLVL, "_VrpnOpenFunction(): " RED_TEXT "An error occurred while trying to send VRPN hail message - %d bytes sent instead of %d.\n" NORM_TEXT, rval, 24);
		if (rval < 0)
			perror("_VrpnOpenFunction(): write()");
	}
	vrSocketAwaitData(aux->vrpn_socket);		/* NOTE: currently hard-coded to wait up to 0.15 second */
	read_result = (ssize_t)recv(aux->vrpn_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf)-aux->eobuf_pos, 0);
	if (read_result <= 0) {
/* Try a few more times [04/06/22: Experiment] */ /* Probably delete me */
for (int count = 0; count < 10 && read_result <= 0; count++) {
  sleep(2);
  read_result = (ssize_t)recv(aux->vrpn_socket, &aux->buf[aux->eobuf_pos], sizeof(aux->buf)-aux->eobuf_pos, 0);
}
		vrPrintf(RED_TEXT "_VrpnOpenFunction(): returning unopened!\n" NORM_TEXT);
		return;		/* didn't get a quick response -- return w/o setting the open flag */
	}
	aux->open = 1;		/* communication line has been established */
	aux->eobuf_pos = read_result;

	/* parse the version number (from the "cookie") */
	sscanf(aux->buf, "vrpn: ver. %d.%d  0", &aux->server_ver_major, &aux->server_ver_minor);
	strncpy(aux->version, aux->buf, strlen(aux->buf));
#if !defined(TEST_APP) /* { */
	vrDbgPrintfN(VRPN_DBGLVL, "FreeVR (%.4f): VRPN: got (%d) '%s' -- major = %d, minor = %d\n",
		vrCurrentSimTime(), read_result, aux->buf, aux->server_ver_major, aux->server_ver_minor);
#else
	vrPrintf("VRPN: got (%d) '%s' -- major = %d, minor = %d\n",
		read_result, aux->buf, aux->server_ver_major, aux->server_ver_minor);
#endif
	memmove(aux->buf, &aux->buf[24], (sizeof(aux->buf)-24));	/* shift the data */
	aux->eobuf_pos -= 24;

	/* initialize the device */
	if (_VrpnInitializeDevice(aux) < 0) {
		devinfo->operating = 0;
		vrErrPrintf("_VrpnOpenFunction(): "
			RED_TEXT "Warning, unable to initialize '%s' DEVICE.\n" NORM_TEXT,
			devinfo->name);
	} else {
		devinfo->operating = 1;
		vrDbgPrintf("_VrpnOpenFunction(): Done opening and initializing VRPN input device '%s'.\n", devinfo->name);
	}

	vrTrace("_VrpnOpenFunction", "returning");

	return;
}


/***************************************************************************/
/* _VrpnCloseFunction(): this callback closes the communications channel   */
/*   with the VRPN device.                                                 */
static void _VrpnCloseFunction(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* TODO: add VRPN shut down operations */

	/*******************/
	/* close the socket */
	close(aux->vrpn_socket);

	if (aux != NULL) {
		vrShmemFree(aux);	/* aka devinfo->aux_data */
		aux = NULL;
	}

	return;
}


#if defined(FREEVR) /* { */
/*************************************************************************/
/* _VrpnResetFunction(): this callback provides the means by which the   */
/*    device or input stream can be reset to a good state.               */
static void _VrpnResetFunction(vrInputDevice *devinfo)
{
	_VrpnPrivateInfo	*aux = (_VrpnPrivateInfo *)devinfo->aux_data;

	/* TODO: Reset the VRPN DEVICE */

	return;
}


/****************************************************************************/
/* _VrpnPollFunction(): this callback actually reads the data from the      */
/*    input device or stream using a polling (request) method of operation. */
/*    A check is done to confirm that the device is in "operating" state    */
/*    before attempting to read the data.                                   */
static void _VrpnPollFunction(vrInputDevice *devinfo)
{
	if (devinfo->operating) {
		_VrpnGetData(devinfo);
	} else {
		/* TODO: try to open the device again */
	}

	return;
}



	/*******************************/
	/***  FreeVR public routine  ***/
	/*******************************/

/****************************************************************************/
/* vrVrpnInitInfo(): this function serves as the "public" means by which    */
/*   a device (or input stream) is interfaced with from the FreeVR library. */
/*   This function creates the shared memory for the device version string, */
/*   and then instantiates the new instance with all the callbacks needed   */
/*   for FreeVR to operate the input.                                       */
void vrVRPNInitInfo(vrInputDevice *devinfo)
{
	devinfo->version = (char *)vrShmemStrDup("-get from VRPN device-");
	devinfo->Create = vrCallbackCreateNamed("Vrpn:Create-Def", _VrpnCreateFunction, 1, devinfo);
	devinfo->Open = vrCallbackCreateNamed("Vrpn:Open-Def", _VrpnOpenFunction, 1, devinfo);
	devinfo->Close = vrCallbackCreateNamed("Vrpn:Close-Def", _VrpnCloseFunction, 1, devinfo);
	devinfo->Reset = vrCallbackCreateNamed("Vrpn:Reset-Def", _VrpnResetFunction, 1, devinfo);
	devinfo->PollData = vrCallbackCreateNamed("Vrpn:PollData-Def", _VrpnPollFunction, 1, devinfo);
	devinfo->PrintAux = vrCallbackCreateNamed("Vrpn:PrintAux-Def", _VrpnPrintPrivateInfo, 0);

	vrDbgPrintfN(INPUT_DBGLVL, "vrVrpnInitInfo: callbacks created.\n");
}
#endif /* } FREEVR */


	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



/****************************************************************************/
/****************************************************************************/
/* The following section is the independent "test" utility that can be used */
/*   to test communications and output from the input device (or stream).   */
#if defined(TEST_APP) /* { */

/*********************************************************************/
/* Ugh, I hate globals, but I don't know a better way to get the     */
/*   "done" value into the interrupt signal function exit_testapp(). */
static	int	done = -1;	/* -1 indicates that communication isn't open yet */
vrTime	time_immemorial = -1;	/* -1 indicates that the start time hasn't been set yet */

#define	RENDERLINES	0
#define	RENDERCSV	1
#define	RENDERSCREEN	2


/*********************************************************************/
/* Copied from "vr_system.c" to avoid linking multiple files.        */
vrTime vrStartTime()
{
	return time_immemorial;
}


/*********************************************************************/
/* Copied from "vr_system.c" to avoid linking multiple files.        */
vrTime vrCurrentWallTime()
{
	vrTime		time_now;
	struct timeval	now;

	gettimeofday(&now, NULL);
	time_now = ((vrTime)now.tv_sec + (vrTime)now.tv_usec / 1000000.0);

	return (time_now);
}


/*******************************************************************/
/* _RenderValuesLine(): Render the Tracking values in a streaming  */
/*   line style format -- with a carriage return, so will overwrite*/
/*   previous lines (unless the screen is insufficiently wide.     */
void _RenderValuesLine(_VrpnPrivateInfo *aux)
{
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */
	int		devnum;			/* loop counter for looping on devices */
	int		input;			/* loop counter for each device's inputs */
	int		devname_flag;		/* flag of whether a device's name has been printed yet */
	int		button1_value = -1;	/* value of the first actual button in the system (used for chorded-quit) */
	int		button2_value = -1;	/* value of the second actual button in the system (used for chorded-quit) */

	/* for calculating the Euler angles of a position tracker input */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
	vrEuler		euler_input;		/* The Euler+translation representation */

	/* Display any newly arrived text messages */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		devname_flag = 0;
		if (vrpn_dev->newly_reported_message) {
			char	*replace;		/* used for replacing newlines with a space */
			if (!devname_flag) {
				printf("Device %d(%s):", devnum, vrpn_dev->name);
				devname_flag = 1;
			}
			while ((replace = strchr(vrpn_dev->text_message, '\n')) != NULL) *replace = ' ';	/* replace newlines with a space */
			printf("message = %s  ", vrpn_dev->text_message);
			vrpn_dev->newly_reported_message = 0u;
		}
		if (devname_flag)
			printf("\n");
	}

	/* Display tracker (6-sensor) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		devname_flag = 0;
		for (input = 0; input < MAX_6SENSORS; input++) {
			if (vrpn_dev->newly_reported_posquat[input]) {
				if (!devname_flag) {
					printf("Device %d(%s):", devnum, vrpn_dev->name);
					devname_flag = 1;
				}
				printf("s%d x:% .3f y:% .3f z:% .3f  qx:% .3f qy:% .3f qz:% .3f qw:% .3f  ",
					input,
					vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z],
					vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
				vrpn_dev->newly_reported_posquat[input] = 0u;

				/* Store the given <input> in FreeVR types (and calculate Eulers) */
				vrPointSet3d(&location_input, vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z]);
				vrQuatSet4d(&quat_input, vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
				vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
			}
		}
		if (devname_flag)
			printf("\n");
	}
	/* NOTE: tracker (6-sensor) velocity and acceleration values are not (yet) output */


	/* Display analog (valuator) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		devname_flag = 0;
		for (input = 0; input < MAX_VALUATORS; input++) {
			if (vrpn_dev->newly_reported_analog[input]) {
#if 0 /* old, more verbose method */
				if (!devname_flag) {
					printf("Device %d(%s):", devnum, vrpn_dev->name);
					devname_flag = 1;
				}
				printf("val[%d] = % 6.3f  ", input, vrpn_dev->analog[input]);
#else /* print more concisely so we can see more values */
				if (!devname_flag) {
					printf("Device %d(%s):valuators: ", devnum, vrpn_dev->name);
					devname_flag = 1;
				}
				printf("%d:%6.3lf ", input, vrpn_dev->analog[input]);
#endif
				vrpn_dev->newly_reported_analog[input] = 0u;
			}
		}
		if (devname_flag)
			printf("\n");
	}

	/* Display button (2-switch) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		devname_flag = 0;
		button1_value = -1;	/* reset back to -1 for each device */
		button2_value = -1;	/* reset back to -1 for each device */
		for (input = 0; input < MAX_BUTTONS; input++) {
			if (vrpn_dev->newly_reported_button[input]) {
				if (!devname_flag) {
					printf("Device %d(%s):", devnum, vrpn_dev->name);
					devname_flag = 1;
				}
				printf("button[%d] = %d  ", input, vrpn_dev->button[input]);
				vrpn_dev->newly_reported_button[input] = 0u;
			}

			/* Alternate mode of quitting is to press buttons 1 & 2 of any VRPN device */
			if (button1_value < 0)
				button1_value = vrpn_dev->button[input];
			else if (button2_value < 0)
				button2_value = vrpn_dev->button[input];
			if ((button1_value > 0) && (button2_value> 0))
				done = 1;
		}
		if (devname_flag)
			printf("\n");
	}

	/* Flush all the output */
	fflush(stdout);
}


/********************************************************************/
/* _RenderValuesCSV(): Render the Tracking values in a streaming    */
/*   format with terse comma-separated-value information.  All new  */
/*   data reports are placed on their own line (record), though the */
/*   columns will vary depending on the type of data.               */
/* NOTE: unlike the line-rendering style, each line will have one   */
/*   input type, so there is no need to keep track of whether a     */
/*   particular device is already being reported.                   */
void _RenderValuesCSV(_VrpnPrivateInfo *aux)
{
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */
	int		devnum;			/* loop counter for looping on devices */
	int		input;			/* loop counter for each device's inputs */
	int		button1_value = -1;	/* value of the first actual button in the system (used for chorded-quit) */
	int		button2_value = -1;	/* value of the second actual button in the system (used for chorded-quit) */

	/* Display any newly arrived text messages */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		char	*replace;		/* used for replacing newlines with a space */
		vrpn_dev = &(aux->device[devnum]);
		if (vrpn_dev->newly_reported_message) {
			while ((replace = strchr(vrpn_dev->text_message, '\n')) != NULL) *replace = ' ';	/* replace newlines with a space */
			printf("%lf,", vrpn_dev->message_time);
			printf("%d,%s,", devnum, vrpn_dev->name);
			printf("%d,%d,%d,%s\n", aux->MessageText, vrpn_dev->text_level, vrpn_dev->text_severity, vrpn_dev->text_message);
			vrpn_dev->newly_reported_message = 0u;
		}
	}

	/* Display tracker (6-sensor) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		for (input = 0; input < MAX_6SENSORS; input++) {
			if (vrpn_dev->newly_reported_posquat[input]) {
				printf("%lf,", vrpn_dev->posquat_time[input]);
				printf("%d,%s,", devnum, vrpn_dev->name);
				//printf("%d,%d,% .3f,% .3f,% .3f,% .3f,% .3f,% .3f,% .3f\n",
				printf("%d,%d,% .6f,% .6f,% .6f,% .6f,% .6f,% .6f,% .6f\n",	/* I want higher res data for my project */
					aux->MessageSensorPos, input,
					vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z],
					vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
				vrpn_dev->newly_reported_posquat[input] = 0u;

#if 0 /* not doing Eulers for CSV output */
				/* Store the given <input> in FreeVR types (and calculate Eulers) */
				vrPointSet3d(&location_input, vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z]);
				vrQuatSet4d(&quat_input, vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
				vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
#endif
			}
		}
	}

	/* Display analog (valuator) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		for (input = 0; input < MAX_VALUATORS; input++) {
			if (vrpn_dev->newly_reported_analog[input]) {
				printf("%lf,", vrpn_dev->analog_time[input]);
				printf("%d,%s,", devnum, vrpn_dev->name);
				printf("%d,%d,%6.3lf\n", aux->MessageValuator, input, vrpn_dev->analog[input]);
				vrpn_dev->newly_reported_analog[input] = 0u;
			}
		}
	}

	/* Display button (2-switch) information */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		vrpn_dev = &(aux->device[devnum]);
		button1_value = -1;	/* reset back to -1 for each device */
		button2_value = -1;	/* reset back to -1 for each device */
		for (input = 0; input < MAX_BUTTONS; input++) {
			if (vrpn_dev->newly_reported_button[input]) {
				printf("%lf,", vrpn_dev->button_time[input]);
				printf("%d,%s,", devnum, vrpn_dev->name);
				printf("%d,%d,%d\n", aux->MessageButton, input, vrpn_dev->button[input]);
				vrpn_dev->newly_reported_button[input] = 0u;
			}

			/* Alternate mode of quitting is to press buttons 1 & 2 of any VRPN device */
			if (button1_value < 0)
				button1_value = vrpn_dev->button[input];
			else if (button2_value < 0)
				button2_value = vrpn_dev->button[input];
			if ((button1_value > 0) && (button2_value> 0))
				done = 1;
		}
	}

	/* Flush all the output */
	fflush(stdout);
}


#if 1 /* Newer, cleaner, version */
/*******************************************************************/
/* _RenderValuesScreen(): Render the Tracking values in a text-gui */
/*   style format  (Do screenrender style).                        */
/* NOTE: this routine relies on using the ANSI terminal character  */
/*   codes.  Since this is just a test application, I decided to   */
/*   forgo the requirement of the n/curses library, and instead    */
/*   limit the terminal on which this version would work.          */
/*   On Linux do "man console_codes" to get the sequence codes     */
/*   for other operations.                                         */
void _RenderValuesScreen(_VrpnPrivateInfo *aux)
{
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */
	int		devnum;			/* for looping through devices */
	int		input;			/* loop counter for each device's inputs */
	int		button1_value = -1;	/* value of the first actual button in the system (used for chorded-quit) */
	int		button2_value = -1;	/* value of the second actual button in the system (used for chorded-quit) */
	int		line, column = 0;	/* line/column on the screen to place data */
	int		startcol = 40;		/* the first X value to use for column renderings */
	float		percent;		/* used for simple valuator visualization */

	int		num_lines = 40;		/* number of lines for showing chart */
	int		header_lines = 6;
	int		center = num_lines / 2.0 + (header_lines - 2);	/* shift for header */
	//int		center = (num_lines-header_lines) / 2.0 + header_lines + 1;	/* shift for header */ /* vruidd version */
	int		col_width = 13;	/* space allocated to each column of input data */
	int		vertical;		/* the vertical line to render on -- used to prevent underflow */

	float		max_trans = 5.0;	/* 5 feet in each direction */
	float		max_rot = 180.0;	/* 180 degrees max in each direction */

	float		trans_scale = (num_lines-header_lines) * -0.5 / max_trans;
	float		rot_scale = (num_lines-header_lines) * -0.5 / max_rot;
	float		quat_scale = (num_lines-header_lines) * -0.5 / 1.0; /* (1.0 = max_quat) */

	/* for calculating the Euler angles of a position tracker input */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
	vrEuler		euler_input;		/* The Euler+translation representation */


	/* move to upper left corner of screen */
	vrPrintf(ANSI_MOVETOUPPERLEFT);

#if 0 /* clear some or all of the screen -- choose */
	/* clear to end of screen */
	vrPrintf(ANSI_CLEARTOEOS);
#else
	/* Clear just the part of the screen we're using to avoid deleting messages at the bottom */
	for (line = 0; line <= num_lines; line++) {
		vrPrintf(ANSI_MOVETOROWCOL, line, 0);
		vrPrintf(TTY_CLEARTORIGHT);
	}
#endif

	/* draw a data vs. debug separating line */
	vrPrintf(ANSI_MOVETOROWCOL, num_lines+1, 0);
	vrPrintf("=" TTY_REPEATCHAR, 80);
	vrPrintf(ANSI_MOVETOROWCOL, 1000, 0);

	/* save current cursor position */
	//vrPrintf("\n1\n2\n3\n4\n5\n" ANSI_CURSORLOC_SAVE);	delete me
	vrPrintf(ANSI_CURSORLOC_SAVE);

	/******************************/
	/* Display header information */
	vrPrintf(ANSI_MOVETOROWCOL "VRPN Server reported devices:", 2, 2);
	/* NOTE: device[0] is the "simulator" device that I add, so skip it. */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		if (aux->device[devnum].name[0] != '\0')
			vrPrintf(ANSI_MOVETOROWCOL "%2d: '%s'", devnum+2, 4, devnum, aux->device[devnum].name);
	}

	/* Set normal text output to be within the given margin      */
	/*   (This pushes all debug output into the "margin" region, */
	/*   and thus doesn't obscure the screen rendering.)         */
	/*   NOTE: we exaggerate the bottom line to allow the screen */
	/*   to be expanded and still work properly.                 */
	vrPrintf(ANSI_SETMARGIN_TOPBOT, num_lines+2, num_lines+200);

	/******************************************/
	/* Display tracker (6-sensor) information */
	/* start counting at '1' to skip the simulator placeholder */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);

		if (vrpn_dev->ever_reported_posquat_any) {
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, column*col_width + startcol, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "trackers" ANSI_TEXTNOUNDER, line + 5, startcol + column*col_width);

			for (input = 0; input < MAX_6SENSORS; input++) {
#if 0
vrPrintf("\e[%d;30H input = %d -- scale = %f, X=%f, Y=%f, Z=%f\n", 42+input, input, trans_scale, vrpn_dev->posquat[input][VR_X]*trans_scale, vrpn_dev->posquat[input][VR_Y]*trans_scale, vrpn_dev->posquat[input][VR_Z]*trans_scale);
#endif
			/* MARKER-START -- unindented by 3 */
				if (vrpn_dev->ever_reported_posquat[input]) {
					/* Store the given <input> in FreeVR types (inc. Eulers) */
					vrPrintf(ANSI_CURSORLOC_RETURN);	/* jump to output margin for possible debugging output */
					vrPointSet3d(&location_input, vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z]);
					vrQuatSet4d(&quat_input, vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
					vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
					vrPrintf(ANSI_CURSORLOC_SAVE);		/* save cursor for next time */

					/* now print info headers */
					vrPrintf(ANSI_MOVETOROWCOL "%2d:XYZAERijkW",  line +  6, startcol + column*col_width, input);
#if 1 /* [optional] show the XYZ values under the header */
					vrPrintf(ANSI_MOVETOROWCOL "X=% 9.5f", line +  7, startcol + column*col_width, vrpn_dev->posquat[input][VR_X]);
					vrPrintf(ANSI_MOVETOROWCOL "Y=% 9.5f", line +  8, startcol + column*col_width, vrpn_dev->posquat[input][VR_Y]);
					vrPrintf(ANSI_MOVETOROWCOL "Z=% 9.5f", line +  9, startcol + column*col_width, vrpn_dev->posquat[input][VR_Z]);
#endif

					/* print a *zero* line */
					vrPrintf(ANSI_MOVETOROWCOL RED_TEXT ">----------<" NORM_TEXT, center, startcol + column*col_width + 2);

#if 1 /* [optional] switch to reverse mode for moving values */
					vrPrintf(REVERSE_TEXT);	/* switch to reverse */
#endif
					/* print the X value */
					vertical = (int)(vrpn_dev->posquat[input][VR_X]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, startcol + column*col_width + 3);

					/* print the Y value */
					vertical = (int)(vrpn_dev->posquat[input][VR_Y]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "Y" ANSI_TEXTFORENORM, vertical, startcol + column*col_width + 4);

					/* print the Z value */
					vertical = (int)(vrpn_dev->posquat[input][VR_Z]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "Z" ANSI_TEXTFORENORM, vertical, startcol + column*col_width + 5);

#if 1 /* [optional] Output the converted Euler rotations */
					/* print the A value (Azimuth) */
					vrPrintf(ANSI_MOVETOROWCOL "A", (int)(euler_input.r[VR_AZIM]*rot_scale)+center, startcol + column*col_width + 6);

					/* print the E value (Elevation) */
					vrPrintf(ANSI_MOVETOROWCOL "E", (int)(euler_input.r[VR_ELEV]*rot_scale)+center, startcol + column*col_width + 7);

					/* print the R value (Roll) */
					vrPrintf(ANSI_MOVETOROWCOL "R", (int)(euler_input.r[VR_ROLL]*rot_scale)+center, startcol + column*col_width + 8);

#endif
#if 1 /* [optional] Output the Quaternion values */
					/* Output the Quaternion values */
					vrPrintf(ANSI_MOVETOROWCOL "i", (int)(vrpn_dev->posquat[input][VR_QX]*quat_scale)+center, startcol + column*col_width + 9);
					vrPrintf(ANSI_MOVETOROWCOL "j", (int)(vrpn_dev->posquat[input][VR_QY]*quat_scale)+center, startcol + column*col_width + 10);
					vrPrintf(ANSI_MOVETOROWCOL "k", (int)(vrpn_dev->posquat[input][VR_QZ]*quat_scale)+center, startcol + column*col_width + 11);
					vrPrintf(ANSI_MOVETOROWCOL "W", (int)(vrpn_dev->posquat[input][VR_QW]*quat_scale)+center, startcol + column*col_width + 12);
#endif


					vrPrintf(NORM_TEXT);	/* switch to normal mode (non-reverse) */
					column++;
#if 0 /* [optional] Print un-responsive trackers */
				} else {
					vrPrintf(ANSI_MOVETOROWCOL "%2d:*******", line + 6, startcol + column*col_width, input);
					column++;
#endif
				}
			}
			/* MARKER-STOP -- unindented by 3 */
		}
	}

	/*****************************************/
	/* Display analog (valuator) information */
	/* start counting at '1' to skip the simulator placeholder */
	/* NOTE: The Analog diagram takes two columns */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);

		if (vrpn_dev->ever_reported_analog_any) {
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, startcol + column*col_width + 2, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "analogs" ANSI_TEXTNOUNDER, line + 5, startcol + column*col_width + 2);
			for (input = 0; input < MAX_VALUATORS; input++) {
				if (vrpn_dev->ever_reported_analog[input]) {
					vrPrintf(ANSI_MOVETOROWCOL "%2d:%7.3f", line + 6, startcol + column*col_width + 2, input, vrpn_dev->analog[input]);
#if 0  /* experimental small valuator representation */
					percent = (vrpn_dev->analog[input]-vrpn_dev->analog_min[input]) / (vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]);
					vrPrintf(" %7s",
						(percent < 0.2 ? "|-    |" :
						(percent < 0.4 ? "| -   |" :
						(percent < 0.6 ? "|  +  |" :
						(percent < 0.8 ? "|   - |" :
						                 "|    -|")))));
#else
					if ((vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]) < 0.0001) {
						/* When we don't have min/max values yet, then fill with dashes */
						vrPrintf(" %11s", "|---------|");
					} else {
						percent = (vrpn_dev->analog[input]-vrpn_dev->analog_min[input]) / (vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]);
						vrPrintf(" %11s",
							(percent < 0.11 ? "|-        |" :
							(percent < 0.22 ? "| -       |" :
							(percent < 0.33 ? "|  -      |" :
							(percent < 0.44 ? "|   -     |" :
							(percent < 0.56 ? "|    +    |" :
							(percent < 0.67 ? "|     -   |" :
							(percent < 0.78 ? "|      -  |" :
							(percent < 0.89 ? "|       - |" :
									  "|        -|")))))))));
#endif
#if 0 /* show the range of values received thus far */
						vrPrintf(" [%.2f,%.2f]", vrpn_dev->analog_min[input], vrpn_dev->analog_max[input]);
#endif
						line++;
					}
				}
			}
			column++;	/* take 2 columns */
			column++;
		}
	}

	/*****************************************/
	/* Display button (2-switch) information */
	/* start counting at '1' to skip the simulator placeholder */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);
		button1_value = -1;	/* reset back to -1 for each device */
		button2_value = -1;	/* reset back to -1 for each device */

		if (vrpn_dev->ever_reported_button_any) {
//printf("\e[10;20Hcolumn = %d\n", column);
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, startcol + column*col_width + 2, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "buttons" ANSI_TEXTNOUNDER, line + 5, startcol + column*col_width + 2);
			for (input = 0; input < MAX_BUTTONS; input++) {
				if (vrpn_dev->ever_reported_button[input]) {
					vrPrintf(ANSI_MOVETOROWCOL "%2d: %d", line + 6, startcol + column*col_width + 2, input, vrpn_dev->button[input]);
					line++;
				}

#if 0
				/* Alternate mode of quitting is to press buttons 1 & 2 of any VRPN device */
				if (button1_value < 0)
					button1_value = vrpn_dev->button[input];
				else if (button2_value < 0)
					button2_value = vrpn_dev->button[input];
				if ((button1_value > 0) && (button2_value> 0))
					done = 1;
#endif
			}
			column++;
		}
	}

	/* return to the original cursor position */
	vrPrintf(ANSI_CURSORLOC_RETURN);

	/*******************************************/
	/* Display text & pong message information */
	/*   (put them below the render frame)     */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		char	*replace;		/* used for replacing newlines with a space */
		vrpn_dev = &(aux->device[devnum]);
		if (vrpn_dev->newly_reported_message) {
			while ((replace = strchr(vrpn_dev->text_message, '\n')) != NULL) *replace = ' ';	/* replace newlines with a space */
			printf("VRPN Message: %s\n", vrpn_dev->text_message);
			vrpn_dev->newly_reported_message = 0u;
		}
#if 0
		if (vrpn_dev->newly_reported_pong) {
			printf("VRPN Pong recieved\n");
			vrpn_dev->newly_reported_pong = 0u;
		}
#endif
	}

	fflush(stdout);
}


#else
/*******************************************************************/
/* _RenderValuesScreen(): Render the Tracking values in a text-gui */
/*   style format  (Do screenrender style).                        */
/* NOTE: this routine relies on using the ANSI terminal character  */
/*   codes.  Since this is just a test application, I decided to   */
/*   forgo the requirement of the n/curses library, and instead    */
/*   limit the terminal on which this version would work.          */
/*   On Linux do "man console_codes" to get the sequence codes     */
/*   for other operations.                                         */
void _RenderValuesScreen(_VrpnPrivateInfo *aux)
{
	VRPNdevice	*vrpn_dev;		/* pointer to a device in the VRPN device list */
	int		devnum;			/* loop counter for looping on devices */
	int		input;			/* loop counter for each device's inputs */
	int		button1_value = -1;	/* value of the first actual button in the system (used for chorded-quit) */
	int		button2_value = -1;	/* value of the second actual button in the system (used for chorded-quit) */
	int		line, column = 0;	/* line/column on the screen to place data */
	int		startcol = 40;		/* the first X value to use for column renderings */
	int		count;			/* loop counter */
	float		percent;		/* used for simple valuator visualization */

	int		num_lines = 40;		/* number of lines for showing chart */
	float		max_trans = 5.0;	/* 5 feet in each direction */
	float		max_rot = 180.0;	/* max for each direction */

	int		column_width = 13;	/* space allocated to each column of input data */

	int		header_lines = 6;
	int		center = num_lines / 2.0 + (header_lines - 2);	/* shift for header */
	int		vertical;		/* the vertical line to render on -- used to prevent underflow */
	float		trans_scale = (num_lines-header_lines) * -0.5 / max_trans;
//trans_scale *= 100.0;
	float		rot_scale = (num_lines-header_lines) * -0.5 / max_rot;
	float		quat_scale = (num_lines-header_lines) * -0.5 / 1.0; /* (1.0 = max_quat) */

	/* for calculating the Euler angles of a position tracker input */
	vrPoint		location_input;		/* The location/translation portion */
	vrQuat		quat_input;		/* The quaternion portion */
	vrEuler		euler_input;		/* The Euler+translation representation */


	/* move to upper left corner of screen */
	vrPrintf(ANSI_MOVETOUPPERLEFT);

#if 0 /* clear some or all of the screen -- choose */
	/* clear to end of screen */
	vrPrintf(ANSI_CLEARTOEOS);
#else
	/* Clear just the part of the screen we're using to avoid deleting messages at the bottom */
	for (line = 0; line <= num_lines; line++) {
		vrPrintf(ANSI_MOVETOROWCOL, line, 0);
		vrPrintf(TTY_CLEARTORIGHT);
	}
#endif

	/* draw a data vs. debug separating line */
	vrPrintf(ANSI_MOVETOROWCOL, num_lines+1, 0);
	vrPrintf("=" TTY_REPEATCHAR, 80);
	vrPrintf(ANSI_MOVETOROWCOL, 1000, 0);

	/* save current cursor position */
	//vrPrintf("\n1\n2\n3\n4\n5\n" ANSI_CURSORLOC_SAVE);
	vrPrintf(ANSI_CURSORLOC_SAVE);

	/* Display device information */
	vrPrintf(ANSI_MOVETOROWCOL "VRPN Server reported devices:", 2, 2);
	/* NOTE: device[0] is the "simulator" device that I add, so skip it. */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		if (aux->device[devnum].name[0] != '\0')
			vrPrintf(ANSI_MOVETOROWCOL "%2d: '%s'", devnum+2, 4, devnum, aux->device[devnum].name);
	}

	/* Set normal text output to be within the given margin      */
	/*   (This pushes all debug output into the "margin" region, */
	/*   and thus doesn't obscure the screen rendering.)         */
	/*   NOTE: we exaggerate the bottom line to allow the screen */
	/*   to be expanded and still work properly.                 */
	vrPrintf(ANSI_SETMARGIN_TOPBOT, num_lines+2, num_lines+200);


	/******************************************/
	/* Display tracker (6-sensor) information */
	/* start counting at '1' to skip the simulator placeholder */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);

		if (vrpn_dev->ever_reported_posquat_any) {
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, column*column_width + startcol, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "trackers" ANSI_TEXTNOUNDER, line + 5, startcol + column*column_width);

			for (input = 0; input < MAX_6SENSORS; input++) {
#if 0
vrPrintf("\e[%d;30H input = %d -- scale = %f, X=%f, Y=%f, Z=%f\n", 42+input, input, trans_scale, vrpn_dev->posquat[input][VR_X]*trans_scale, vrpn_dev->posquat[input][VR_Y]*trans_scale, vrpn_dev->posquat[input][VR_Z]*trans_scale);
#endif
				if (vrpn_dev->ever_reported_posquat[input]) {
					/* Store the given <input> in FreeVR types (inc. Eulers) */
					vrPrintf(ANSI_CURSORLOC_RETURN);	/* jump to output margin for possible debugging output */
					vrPointSet3d(&location_input, vrpn_dev->posquat[input][VR_X], vrpn_dev->posquat[input][VR_Y], vrpn_dev->posquat[input][VR_Z]);
					vrQuatSet4d(&quat_input, vrpn_dev->posquat[input][VR_QX], vrpn_dev->posquat[input][VR_QY], vrpn_dev->posquat[input][VR_QZ], vrpn_dev->posquat[input][VR_QW]);
					vrEulerSetFromQuatPoint(&euler_input, &quat_input, &location_input);
					vrPrintf(ANSI_CURSORLOC_SAVE);		/* save cursor for next time */

					/* now print info headers */
					vrPrintf(ANSI_MOVETOROWCOL "%2d:XYZAERijkW",  line +  6, startcol + column*column_width, input);
#if 1 /* [optional] show the XYZ values under the header */
					vrPrintf(ANSI_MOVETOROWCOL "X=% 9.5f", line +  7, startcol + column*column_width, vrpn_dev->posquat[input][VR_X]);
					vrPrintf(ANSI_MOVETOROWCOL "Y=% 9.5f", line +  8, startcol + column*column_width, vrpn_dev->posquat[input][VR_Y]);
					vrPrintf(ANSI_MOVETOROWCOL "Z=% 9.5f", line +  9, startcol + column*column_width, vrpn_dev->posquat[input][VR_Z]);
#endif

					/* print a *zero* line */
					vrPrintf(ANSI_MOVETOROWCOL RED_TEXT ">----------<" NORM_TEXT, center, startcol + column*column_width + 2);

#if 1 /* [optional] switch to reverse mode for moving values */
					vrPrintf(REVERSE_TEXT);	/* switch to reverse */
#endif
					/* print the X value */
					vertical = (int)(vrpn_dev->posquat[input][VR_X]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "X" ANSI_TEXTFORENORM, vertical, startcol + column*column_width + 3);

					/* print the Y value */
					vertical = (int)(vrpn_dev->posquat[input][VR_Y]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "Y" ANSI_TEXTFORENORM, vertical, startcol + column*column_width + 4);

					/* print the Z value */
					vertical = (int)(vrpn_dev->posquat[input][VR_Z]*trans_scale)+center;
					if (vertical < 0) {
						/* if we go off the top, stop and make it red */
						vertical = 0;
						vrPrintf(RED_TEXT);
					}
					if (vertical > num_lines+1) {
						/* if we go off the bottom, stop and make it red */
						vertical = num_lines+1;
						vrPrintf(RED_TEXT);
					}
					vrPrintf(ANSI_MOVETOROWCOL "Z" ANSI_TEXTFORENORM, vertical, startcol + column*column_width + 5);

#if 1 /* Output the converted Euler rotations */
					/* print the A value (Azimuth) */
					vrPrintf(ANSI_MOVETOROWCOL "A", (int)(euler_input.r[VR_AZIM]*rot_scale)+center, startcol + column*column_width + 6);

					/* print the E value (Elevation) */
					vrPrintf(ANSI_MOVETOROWCOL "E", (int)(euler_input.r[VR_ELEV]*rot_scale)+center, startcol + column*column_width + 7);

					/* print the R value (Roll) */
					vrPrintf(ANSI_MOVETOROWCOL "R", (int)(euler_input.r[VR_ROLL]*rot_scale)+center, startcol + column*column_width + 8);

#endif
#if 1 /* Output the Quaternion values */
					vrPrintf(ANSI_MOVETOROWCOL "i", (int)(vrpn_dev->posquat[input][VR_QX]*quat_scale)+center, startcol + column*column_width + 9);
					vrPrintf(ANSI_MOVETOROWCOL "j", (int)(vrpn_dev->posquat[input][VR_QY]*quat_scale)+center, startcol + column*column_width + 10);
					vrPrintf(ANSI_MOVETOROWCOL "k", (int)(vrpn_dev->posquat[input][VR_QZ]*quat_scale)+center, startcol + column*column_width + 11);
					vrPrintf(ANSI_MOVETOROWCOL "W", (int)(vrpn_dev->posquat[input][VR_QW]*quat_scale)+center, startcol + column*column_width + 12);
#endif


					vrPrintf(NORM_TEXT);	/* switch to normal mode */
					column++;
#if 0 /* [optional] Print un-responsive trackers */
				} else {
					vrPrintf(ANSI_MOVETOROWCOL "%2d:*******", line + 6, startcol + column*column_width, input);
					column++;
#endif
				}
			}
		}
	}

	/*****************************************/
	/* Display analog (valuator) information */
	/* start counting at '1' to skip the simulator placeholder */
	/* NOTE: The Analog diagram takes two columns */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);

		if (vrpn_dev->ever_reported_analog_any) {
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, startcol + column*column_width + 2, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "analogs" ANSI_TEXTNOUNDER, line + 5, startcol + column*column_width + 2);
			for (input = 0; input < MAX_VALUATORS; input++) {
				if (vrpn_dev->ever_reported_analog[input]) {
					vrPrintf(ANSI_MOVETOROWCOL "%2d:%7.3f", line + 6, startcol + column*column_width + 2, input, vrpn_dev->analog[input]);
#if 0  /* experimental small valuator representation */
					percent = (vrpn_dev->analog[input]-vrpn_dev->analog_min[input]) / (vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]);
					vrPrintf(" %7s",
						(percent < 0.2 ? "|-    |" :
						(percent < 0.4 ? "| -   |" :
						(percent < 0.6 ? "|  +  |" :
						(percent < 0.8 ? "|   - |" :
						                 "|    -|")))));
#else
					if ((vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]) < 0.0001) {
						/* When we don't have min/max values yet, then fill with dashes */
						vrPrintf(" %11s", "|---------|");
					} else {
						percent = (vrpn_dev->analog[input]-vrpn_dev->analog_min[input]) / (vrpn_dev->analog_max[input]-vrpn_dev->analog_min[input]);
						vrPrintf(" %11s",
							(percent < 0.11 ? "|-        |" :
							(percent < 0.22 ? "| -       |" :
							(percent < 0.33 ? "|  -      |" :
							(percent < 0.44 ? "|   -     |" :
							(percent < 0.56 ? "|    +    |" :
							(percent < 0.67 ? "|     -   |" :
							(percent < 0.78 ? "|      -  |" :
							(percent < 0.89 ? "|       - |" :
									  "|        -|")))))))));
#endif
#if 0 /* show the range of values received thus far */
						vrPrintf(" [%.2f,%.2f]", vrpn_dev->analog_min[input], vrpn_dev->analog_max[input]);
#endif
						line++;
					}
				}
			}
			column++;	/* take 2 columns */
			column++;
		}
	}

	/*****************************************/
	/* Display button (2-switch) information */
	/* start counting at '1' to skip the simulator placeholder */
	for (devnum = 1; devnum < MAX_DEVICES; devnum++) {
		line = 0;
		vrpn_dev = &(aux->device[devnum]);
		button1_value = -1;	/* reset back to -1 for each device */
		button2_value = -1;	/* reset back to -1 for each device */

		if (vrpn_dev->ever_reported_button_any) {
//printf("\e[10;20Hcolumn = %d\n", column);
			vrPrintf(ANSI_MOVETOROWCOL "%-12s", line + 4, startcol + column*column_width + 2, vrpn_dev->name);
			vrPrintf(ANSI_MOVETOROWCOL ANSI_TEXTUNDER "buttons" ANSI_TEXTNOUNDER, line + 5, startcol + column*column_width + 2);
			for (input = 0; input < MAX_BUTTONS; input++) {
				if (vrpn_dev->ever_reported_button[input]) {
					vrPrintf(ANSI_MOVETOROWCOL "%2d: %d", line + 6, startcol + column*column_width + 2, input, vrpn_dev->button[input]);
					line++;
				}

#if 0
				/* Alternate mode of quitting is to press buttons 1 & 2 of any VRPN device */
				if (button1_value < 0)
					button1_value = vrpn_dev->button[input];
				else if (button2_value < 0)
					button2_value = vrpn_dev->button[input];
				if ((button1_value > 0) && (button2_value> 0))
					done = 1;
#endif
			}
			column++;
		}
	}

	/* return to the original cursor position */
	vrPrintf(ANSI_CURSORLOC_RETURN);

	/*******************************************/
	/* Display text & pong message information */
	/*   (put them below the render frame)     */
	for (devnum = 0; devnum < MAX_DEVICES; devnum++) {
		char	*replace;		/* used for replacing newlines with a space */
		vrpn_dev = &(aux->device[devnum]);
		if (vrpn_dev->newly_reported_message) {
			while ((replace = strchr(vrpn_dev->text_message, '\n')) != NULL) *replace = ' ';	/* replace newlines with a space */
			printf("VRPN Message: %s\n", vrpn_dev->text_message);
			vrpn_dev->newly_reported_message = 0u;
		}
#if 0
		if (vrpn_dev->newly_reported_pong) {
			printf("VRPN Pong recieved\n");
			vrpn_dev->newly_reported_pong = 0u;
		}
#endif
	}

	fflush(stdout);
}
#endif


/**************************************************************************/
/* exit_testapp(): this function is used as a signal callback, such that  */
/*    when called, it simply sets the "done" flag, so the main while loop */
/*    in main() while know to terminate at the end of the current pass.   */
/* NOTE: The interruption messages are preceded with a hash symbol ('#')  */
/*    for the sake of making them comments in CSV output.                 */
void exit_testapp()
{
	if (done < 0) {
		/* if this is the case, we haven't even begun the reading loop, so just exit */
		vrPrintf("\nVRPN: Interrupted prior to entering the read-loop.\n");
		exit(1);
	}

	if (done == 1) {
		/* if we've already been here, this is a sign to force quit */
		printf("\n#Persistent interruption... terminating.\n");

		exit(0);
	}

	done = 1;
	printf("\n#Interruption -- if program doesn't close cleanly, try interrupting again.\n");
}


/*****************************************************************************/
/* A test program to communicate with a VRPN device and print the results.   */
int main(int argc, char *argv[])
{
static	char			*err_missing_arg = "%s: missing argument to '%s'.\n";
	_VrpnPrivateInfo	*aux;
	char			*progname;			/* name of the program executable */
	int			count;				/* for looping over things */
	vrInputDevice		device;				/* using FreeVR input device struct */
	int			read_return;			/* return value from the read function */
	int			retval;				/* value returned by _VrpnOpenFunction() */

	/* command line options */
	int			rendermode = RENDERLINES;	/* default: how to render the data */
	int			justonce = 0;			/* default: quit after just one output */
	int			nodata_flag = 0;		/* default: whether or not to actual show input data */
#ifdef VRPN_LIB /* { */
	char			*hostname = "149.166.144.248";	/* default */
	char			*tracker_name = "Tracker0";	/* default */
	char			tracker_dev[256];
#endif /* } VRPN_LIB */

	done = -1;				/* not yet initialized (redundant) */
	signal(SIGINT, exit_testapp);


	/******************************/
	/* setup the device structure */
	aux = (_VrpnPrivateInfo *)malloc(sizeof(*aux));
	memset(aux, 0, sizeof(*aux));
	_VrpnInitializeStruct(aux, "default");

	device.name = "VRPN test device";
	device.operating = 0;
	device.aux_data = aux;

	time_immemorial = vrCurrentWallTime();

	/*********************************************************/
	/* set default parameters based on environment variables */

	if (getenv("VRPN_HOST") != NULL)
		aux->vrpn_host = getenv("VRPN_HOST");
	else	aux->vrpn_host = DEFAULT_HOST;			/* default, if no host given */

	aux->vrpn_port = ((getenv("VRPN_PORT")==NULL) ? DEFAULT_PORT : atoi(getenv("VRPN_PORT")));

#ifdef VRPN_LIB /* { */
	if (getenv("VRPN_TRACKER") != NULL)
		tracker_name = getenv("VRPN_TRACKER");
#endif /* } VRPN_LIB */


	/*****************************************************/
	/* adjust parameters based on command-line-arguments */
	progname = argv[0];
	while ((argc > 1) && (argv[1][0] == '-')) {

		/****************/
		/* Set the port */
		if (!strcmp(argv[1], "-p") || !strcmp(argv[1], "-port")) {
			argv++; argc--;
			if (argc > 1) {
				aux->vrpn_port = atoi(argv[1]);
				argv++; argc--;
			} else {
				fprintf(stderr, err_missing_arg, progname, argv[0]);
				fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-screen|-csv] [-list|-nodata|-once|-1] [-p <port>(%d)] [<VRPN host> (default = '%s')]\n",
					progname, aux->vrpn_port, aux->vrpn_host);
				exit(1);
			}
			continue;		/* return to the while */
		}

		/*****************************************************/
		/* Report information about the device and then quit */
		/* NOTE: the "-nodata" option already basically does a "list" of what's coming from the vrpn_server */
		if (!strcmp(argv[1], "-nodata") || !strcmp(argv[1], "-list")) {
			nodata_flag = 1;
			argv++; argc--;
			continue;		/* return to the while */
		}

		/***************************/
		/* set the "justonce" flag */
		if (!strcmp(argv[1], "-1") || !strcmp(argv[1], "-once")) {
			argv++; argc--;
			justonce = 1;
			continue;		/* return to the while */
		}

		/***************************/
		/* set the "govern_time" flag */
		if (!strcmp(argv[1], "-gt") || !strcmp(argv[1], "-govern_time")) {
			argv++; argc--;
			aux->govern_time = 1;
			continue;		/* return to the while */
		}

		/***************************************/
		/* set the "rendermode" flag to screen */
		if (!strcmp(argv[1], "-screen")) {
			argv++; argc--;
			rendermode = RENDERSCREEN;
			continue;		/* return to the while */
		}

		/************************************/
		/* set the "rendermode" flag to CSV */
		if (!strcmp(argv[1], "-csv")) {
			argv++; argc--;
			rendermode = RENDERCSV;
			continue;		/* return to the while */
		}

		/******************/
		/* Unknown option */
		else {
			/* There are currently no other "-" options, so this is an error */
			fprintf(stderr, RED_TEXT "Usage:" NORM_TEXT " %s [-screen|-csv] [-list|-nodata|-once|-1] [-p <port>(%d)] [<VRPN host> (default = '%s')]\n",
				progname, aux->vrpn_port, aux->vrpn_host); /* NOTE: reported defaults might be effected by environment variables */
			exit(1);
		}
	}

	/* if there are any arguments left, use the first as the daemon hostname */
	if (argc > 1) {
		aux->vrpn_host = strdup(argv[1]);
	}


	/**********************************/
	/* open and initialize the device */
	aux->open = 0;
#ifdef VRPN_LIB /* { */
	snprintf(tracker_dev, sizeof(tracker_dev), "%s@%s", tracker_name, hostname);
	printf("Opening device: %s\n", tracker_dev);
	aux->vrpn_tracker = new vrpn_Tracker_Remote(tracker_dev);
	if (aux->vrpn_tracker == NULL) {
		fprintf(stderr, RED_TEXT "couldn't open VRPN tracker port %s\n" NORM_TEXT, tracker_dev);
		snprintf(aux->version, sizeof(aux->version), "- unconnected VRPN -");
	} else {
		aux->vrpn_tracker->register_change_handler(NULL, _VrpnlibHandleTracker);
		aux->open = 1;
		done = 0;
	}
#else /* Do it the brute-force way */
	_VrpnOpenFunction(&device);
#  if 0 /* I think this is the wrong way to handle this (probably delete me) */
	for (count = 0; count < 10 && aux->open <= 0; count++) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "main(): " RED_TEXT "VRPN Server is not yet communicating.  Trying again...\n" NORM_TEXT);
		vrSleep(100000);
		_VrpnOpenFunction(&device);
	}
#  endif
	done = 0;
#endif /* } VRPN_LIB */

	if (_VrpnInitializeDevice(aux) < 0) {
		vrErrPrintf("main: " RED_TEXT "Warning, unable to initialize VRPN.\n" NORM_TEXT);
		aux->open = 0;
	}

	/**********************/
	/* display the output */

	_VrpnPrintPrivateInfo(stdout, aux, verbose);

	/* quit if flagged to print the device info but not data */
	if (nodata_flag) {
		_VrpnCloseFunction(&device);
		exit(0);
	}

	/* for CSV-rendering, output a simple header */
	if (rendermode == RENDERCSV && aux->open) {
		vrPrintf("# ---------- VRPN Output in CSV format ----------\n");
		vrPrintf("#Meta, %d, vrpn_Base ping_message\n", aux->MessagePing);
		vrPrintf("#Meta, %d, vrpn_Base pong_message\n", aux->MessagePong);
		vrPrintf("#Meta, %d, vrpn_Base text_message\n", aux->MessageText);
		vrPrintf("#Meta, %d, vrpn_Button Change\n", aux->MessageButton);
		vrPrintf("#Meta, %d, vrpn_Analog Channel\n", aux->MessageValuator);
		vrPrintf("#Meta, %d, vrpn_Tracker Pos_Quat\n", aux->MessageSensorPos);
		vrPrintf("#Meta, %d, vrpn_Tracker Velocity\n", aux->MessageSensorVel);
		vrPrintf("#Meta, %d, vrpn_Tracker Acceleration\n", aux->MessageSensorAcc);
		vrPrintf("#Meta,begin\n");
		vrPrintf("#Header, input-time, device-number, device-name, message-type, input-number, input-value(s)\n");
	}

	/* for screen-rendering, scroll everything up so it won't be erased */
	if (rendermode == RENDERSCREEN && aux->open) {
		for (count = 0; count < 60; count++)
			vrPrintf("\n");
		_RenderValuesScreen(aux);	/* This will print the device names at a minimum */
	}

	/****************************************/
	/* read the data and display the output */
	while(aux->open && !done) {

		/* if any new data was received then display it */
//vrPrintf(">>>>>>>>>>about to call _VrpnReadInput()\n");
		read_return = _VrpnReadInput(aux);
		usleep(1000);
//vrPrintf("read_return = %d\n", read_return);
		if (read_return > 0) {
			switch(rendermode) {
			default:
			case RENDERLINES:
				_RenderValuesLine(aux);
				break;;
			case RENDERCSV:
				_RenderValuesCSV(aux);
				break;;
			case RENDERSCREEN:
				_RenderValuesScreen(aux);
				break;;
			}

			/* we're done already if just one printout */
			if (justonce)
				done = 1;
		} else if (read_return == -1) {
			/* there was a communication error during the read, so end */
			done = 1;
		} else {
			/* read_return must be 0 */
			//vrPrintf("No VRPN data read this time\n");
		}
	}

	/*****************/
	/* close up shop */
	if (aux != NULL) {
		/* for CSV-rendering, output closing Meta-data tag */
		if (rendermode == RENDERCSV && aux->open) {
			vrPrintf("#Meta,end\n");
		}

		/* for screen-rendering, correct the terminal mode */
		if (rendermode == RENDERSCREEN && aux->open) {
			/* clean up the screen rendering */
			vrPrintf(ANSI_SETMARGIN_TOPBOT_OFF);
			vrPrintf(ANSI_CURSORLOC_RETURN);
		}
#ifdef VRPN_LIB /* { */
		/* TODO: close the VRPN library connections */
#else /* Do it the brute-force way */
		_VrpnCloseFunction(&device);
#endif /* } VRPN_LIB */
	}

	vrPrintf(BOLD_TEXT "\nVRPN device closed\n" NORM_TEXT);
}

#endif /* } TEST_APP */



	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/
	/*===========================================================*/



#if defined(CAVE) /* { */


	/* ... CAVE stuff here if to also work with CAVElib */


#endif /* } CAVE */

